//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;

/**
 * Basic class for model entities.
 *
 * <p>Agents in the model are representing entities, that can change their state
 * without interaction with other entities. Currently there are 3 types of Agents:
 * LandscapeUnits, Animals and Nests that inherit after this class.</p>
 * 
 * <p> Each Agent is characterized by its unique id number and its
 * AgentState. Each Agent owns its inbox and outbox together with methods that
 * enable using them for message exchange. Agents keep their own trace of time,
 * that once a day is synchronized with the global time. </p>
 *
 * @author Katarzyna Kułakowska
 * @author Tomasz Kułakowski
 * @since 2013-01-08
 */
public abstract class Agent implements Observable {

	/**
	 * Implements or specifies daily activities of an Agent and implements the
	 * methods necessary for using inbox and outbox. <p> The implementation
	 * changes with life stage. </p>
	 *
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @version 1.0
	 */
	public abstract class AgentState implements Observable {
		/*
		 * Methods:
		 */

		/**
		 * Returns true if the Agent is dead, but false if its greenhorn = true.
		 * <p> If the greenhorn = false, changes the value of alive to false and
		 * sends messages to all the Agents noted on contactList that this Agent
		 * is dead. Greenhorns are not tested dying either because they have
		 * been checked in their previous AgentState or because they just
		 * appeared in the program and have not conducted their first daily
		 * activities yet. </p> 
		 * 
		 * <p> Should be overwritten and called from
		 * derived classes only in case of the death of the Agent. </p>
		 *
		 * @param cause 
		 * @return true of the Agent dies, false if it is a greenhorn.
		 * @since 2013-01-7
		 * @verification TEST 1 Warning if the agent is dead already (alive
		 * = false), TEST 2 Warning if the agent died without specifying the cause
		 * (cause = null)
		 */
		public boolean checkDeath(String cause) {
			// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 1
				if (alive == false) {
					try {
						throw new SkyPigException("Agent dying twice: "
								+ String.valueOf(getId()));
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
			}
			// *****VERIFICATION MODE END*****//
			if (greenhorn) {
				return false;
			} else {
				if (cause != null) {
					alive = false;
					causeOfDeath = cause;
					for (long i : contactList) {
						sendMessage(new Message(id, i, "dead_agent"));
					}
					return true;
				} else {
					// *****VERIFICATION MODE START*****//
					try {
						throw new SkyPigException(
								"Agent died without a cause: "
								+ String.valueOf(getId()));
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
					// *****VERIFICATION MODE END*****//
					return false;
				}
			}
		}

		/**
		 * In derived classes should specify how the food is searched for and
		 * collected. Should result in collecting all the food for the current
		 * day and spending the energy.
		 *
		 * <p> Must be overwritten. </p>
		 */
		public abstract void findFood();

		/**
		 * Resets the values of AgentState fields that should be reset daily.
		 *
		 * <p> Should be overwritten if there are fields to be reset in derived
		 * class. </p>
		 */
		public void reset() {
			readyForForaging = true;
		}

		/**
		 * In derived classes should specify how energy from the gathered food
		 * is distributed for necessary processes. Should result in allocating
		 * the energy.
		 *
		 * <p> Must be overwritten. </p>
		 */
		public abstract void manageEnergy();

		/**
		 * In derived classes should specify the activities that are not
		 * conducted daily, but change with season or other variables. Should
		 * result in conducting the proper seasonal activity
		 *
		 * <p> Must be overwritten. </p>
		 */
		public void conductSeasonalActivity() {
			return;
		}

		/**
		 * In derived classes should specify conditions for state change. Should
		 * result in changing the state for a new AgentState if necessary.h
		 *
		 * <p> Must be overwritten. </p>
		 *
		 * @return a new AgentState to be set as state or a current one if there
		 * should be no change
		 */
		public abstract AgentState changeState();

		/**
		 * In derived classes should specify and create the Signature with all
		 * the information that should be stored in database. Should result in
		 * conducting the proper seasonal activity
		 *
		 * <p> May be overwritten. </p>
		 *
		 * @return Signature to be collected and stored in the database.
		 */
		@Override
		public Signature getSignature() {
			Signature mySignature = new Signature(this);
			mySignature.addPrimitiveField("id", "bigint",
					Long.toString(getId()));
			mySignature.addPrimaryKey("id");
			return mySignature;
		}

		/**
		 * Reads messages and takes the proper action in response. <p> Handles
		 * the following types of messages: with topic = "dead_agent", by removing the
		 * sender id from the contactList and returning the answer with the
		 * default topic = "re:dead_agent" and with topic = "remove_agent", also
		 * by removing the sender id from the contactList and returning the
		 * answer with the default topic = "re:remove_agent </p> <p> Should be
		 * overwritten to handle other topics or to specify additional actions to be
		 * taken after receiving "dead_agent" or "remove_agent" message. </p>
		 *
		 * @param m the message that is to be read.
		 * @return Message that is the rely to the read message. 
		 * @since 2012-01-07
		 * @verification TEST 1 Warning if the topic of the message is not recognised.
		 */
		public Message readMessage(Message m) {
			if (m.topic.equalsIgnoreCase("re:dead_agent")
					|| m.topic.equalsIgnoreCase("re:not_existing")
					|| m.topic.equalsIgnoreCase("re:remove_agent")) {
				removeFromContactList(m.senderId);
				return null;
			} else if (m.topic.equalsIgnoreCase("re:dead")) {
				return null;
			} else if (m.topic.equalsIgnoreCase("dead_agent")) {
				removeFromContactList(m.senderId);
				return new Message(getId(), m.senderId, "re:dead_agent", m);
			} else if (m.topic.equalsIgnoreCase("remove_agent")) {
				removeFromContactList(m.senderId);
				return new Message(getId(), m.senderId, "re:remove_agent", m);
			} else {
				// *****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					// ***TEST 1
					try {
						throw new SkyPigException("Unknown message topic.");
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
				// *****VERIFICATION MODE END*****//
				return new Message(getId(), m.senderId,
						"Unknown message topic", m);
			}
		}

		/**
		 * While the inbox is not empty retrieves, reads and replies to
		 * messages.
		 *
		 * <p> Does not need to be overwriten. <p>
		 */
		public void handleMessages() {
			while (!inbox.isEmpty()) {
				Message m = retrieveMessage();
				Message replyToM = null;
				if (!alive && !m.topic.startsWith("re:")) {
					replyToM = new Message(getId(), m.senderId, "re:dead", m,
							null);
				} else {
					replyToM = readMessage(m);
				}
				if (m.inReplyTo == null) {
					outbox.sendMessage(replyToM);
				}
			}
		}

		/**
		 * Sends the given message to the specified recipient.
		 *
		 * @param m the Message to be sent.
		 */
		public final void sendMessage(Message m) {
			outbox.sendMessage(m);
			return;
		}

		/**
		 * Retrieves and removes one message from the inbox.
		 *
		 * <p> If the retrieved Message is a reply to a previously sent Message
		 * (contains it in the field inReplyTo) the Message replied to is
		 * removed from the Outbox. </p>
		 *
		 * @return the retrieved Message or null if the inbox is empty.
		 * @single 2013-01-07
		 * @verification TEST 1 Warning if the reply Message (with topic starting
		 * with "re:") does not contain the Message it is replying to, TEST 2 
		 * checks if the Message that is not a reply (with topic not
		 * starting with "re:")contains any message intReplyTo.
		 */
		public final Message retrieveMessage() {
			Message newMessage = inbox.getMessage();
			// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 1
				if (newMessage.topic.startsWith("re:")
						&& newMessage.inReplyTo == null) {
					try {
						throw new SkyPigException("Message inReplyTo not "
								+ "added to reply message: " + newMessage.topic);
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
				// ***TEST 2
				if (!newMessage.topic.startsWith("re:")
						&& newMessage.inReplyTo != null) {
					try {
						throw new SkyPigException("Wrong message topic for "
								+ "a reply message" + newMessage.topic);
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);

					}
				}
			}
			// *****VERIFICATION MODE END*****//
			outbox.removeMessage(newMessage.inReplyTo);
			return newMessage;
		}

		/**
		 * Checks it the inbox contains any Messages from a specified sender.
		 *
		 * @param sender id of the Messages to be found and returned
		 * @return the Set of Mesages with the sender specified or an empty Set
		 * if they are not found in the inbox
		 */
		public final Set<Message> findMessage(Long sender) {
			return inbox.findMessage(sender);
		}

		/**
		 * 
		 * @param topic
		 * @return
		 */
		public final boolean checkIfMessagePending(String topic) {
			return outbox.findMessage(topic);
		}

		/**
		 * Adds the specified Agent to the contactList.
		 *
		 * @param agentId of the Agent to be added to contactList
		 */
		public void addToContactList(long agentId) {
			contactList.add(agentId);
		}

		/**
		 * 
		 * @param agentId
		 */
		public void removeFromContactList(long agentId) {
			contactList.remove(agentId);
		}

		/**
		 * 
		 * @return
		 * @throws model.core.Agent.AgentException
		 */
		public boolean getReadyForForaging() throws AgentException {
			if (alive) {
				if (readyForForaging) {
					readyForForaging = false;
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		}

		/**
		 * 
		 * @return
		 */
		public boolean isMigrating() {
			return false;
		}
	}

	/*
	 * Fields:
	 */
	/**
	 * Value of the highest Agent id used in the model.
	 */
	private static long lastId = 1;
	/**
	 * Internal date of an Agent. The comparison of this date with a global date
	 * allows to assess, if this Agent has finished its daily activities. It is
	 * synchronized with the global date once a day, after the Agent has
	 * finished.
	 */
	private Calendar currentDate;
	/**
	 * Informs if the Agent is alive.
	 */
	private boolean alive;
	/**
	 * Informs of the cause of the Agent's death. Null if alive.
	 */
	private String causeOfDeath;
	/**
	 * Unique value characterizing the Agent.
	 */
	private long id;
	/**
	 * Stores Inbox for receiving messages.
	 */
	private Inbox inbox;
	/**
	 * Stores Outbox for sending messages.
	 */
	private Outbox outbox;
	/**
	 * Stores the AgentState of the Agent. Agent State specifies the behaviour
	 * if the Agent.
	 */
	private AgentState state;
	/**
	 * If true, indicates that this Agent has not been launched yet or has been
	 * launched within the last day.
	 *
	 * <p> This field enables keeping track of an Agent activities immediately
	 * after launching. </p>
	 */
	private boolean greenhorn;
	/**
	 * Set of ids of Agents that should be notified of this Agent death to
	 * remove the references to it.
	 */
	private Set<Long> contactList;
	private boolean finished;
	private boolean readyForForaging;
	private Calendar ready;

	/*
	 * Methods:
	 */
	/**
	 * Constructs a new Agent.
	 */
	public Agent() {
		alive = true;
		causeOfDeath = null;
		id = lastId++;
		currentDate = World.getDate();
		inbox = new Inbox();
		outbox = new Outbox();
		greenhorn = false;
		contactList = new HashSet<Long>();
		finished = true;
		readyForForaging = false;
		ready = World.getDate();
	}

	/**
	 * Schedules the behaviour of an Agent.
	 *
	 * <p> If the currentDate is before the Apocalypse the Agent checks if the
	 * new day has started. If the new day has not started yet, it proceeds to
	 * end-of-day activities that will be described later. </p> <p> If the new
	 * day has started (the global date is before the Agents' currentDate) the
	 * Agent chcecks the value of alive. If alive = false (meaning the Agent
	 * died yesterday) the method returns and ends the thread. Otherwise it
	 * continues daily activities: </p> <p> It executed the states' method
	 * checkDeath to check if it dies today according to the state-specific
	 * rules. If yes, it proceeds to end-of-day activities. If not, it is
	 * setting the greenhorn to false and resets the state. Then it changes the
	 * state if necessery according to the state-specific rules. </p> <p>
	 * Afterwards the Agent starts executing daily bbehaviour characteristic for
	 * Animals and Nests but in accordance with state-specific rules implemented
	 * also in LandscapeUnits. It manages its energy budget, finds food and
	 * conduct seasonal activities specific for the state. </p> <p> After the
	 * above activities are executed, the Agent proceeds to end-of-day
	 * activities . It handles all previously unhandled messages. If its outbox
	 * is not empty (meaning that some sent messages have not been answered) it
	 * falls asleep and waits for notification in the incoming message. While
	 * notified, it handles messages again and if the outbox is empty,
	 * synchronizes its current date with the global date to mark the end of the
	 * day and falls asleep again. </p> <p> If the current date is on the day of
	 * Apocalypse, it returns. </p>
	 *
	 * @return 
	 * @verification TEST 1 Warning if the Agent starts the day with status 
	 * finished = true, as it means scheduling error, TEST 2 Warning if the 
	 * inherent Agent date is ahead of the World date, as this means scheduling error.
	 * 
	 *
	 */
	public final boolean run() {
		////// VERIFICATION MODE START //////
		// ***TEST 1
		if (World.verificationMode) {
			if (finished) {
				try {
					throw new SkyPigException(
							"Finished Agent entered run()");
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
		////// VERIFICATION MODE END //////
		finished = false;
		if (state.isMigrating()) {
			if (currentDate.before(World.getDate())) {
				currentDate.add(Calendar.DATE, 1);
			}
			finished = true;
		} else {
			if (currentDate.before(World.getDate())) {
				currentDate.add(Calendar.DATE, 1);
				if (state.checkDeath(null) == false) {
					greenhorn = false;
					resetState();
					state = state.changeState();
				}
			}
			try {
				boolean readyforaging = state.getReadyForForaging();
				if (readyforaging) {
					state.manageEnergy();
					state.findFood();
					state.conductSeasonalActivity();
				}
			} catch (AgentException ex) {
				return finished;
			}
			// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 2
				if (currentDate.after(World.getDate())) {
					try {
						throw new SkyPigException(
								"Agent date error within run: "
								+ Integer.toString(currentDate.get(Calendar.YEAR))
								+ "-"
								+ Integer.toString(currentDate.get(Calendar.MONTH) + 1)
								+ "-"
								+ Integer.toString(currentDate.get(Calendar.DAY_OF_MONTH)));
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
			}
			// *****VERIFICATION MODE END*****//
			state.handleMessages();
			if (outbox.isEmpty()) {
				finished = true;
			}
		}
		return finished;
	}

	/**
	 * Assign a new Thread to this Agent and register it in the World.
	 *
	 */
	public final void launchMe() {
		greenhorn = true;
		currentDate = World.getDate();
		World.Instance().addAgent(this);
		return;
	}

	/**
	 * Informs if the Agent has finished its day on the basis of the comparison
	 * between its currentDate and global date.
	 *
	 * @return true if the Agent has finished the day (currentDate equals the
	 * global date), false otherwise.
	 *
	 * @verification TEST 1 checks if the currentDate is not ahead of the global
	 * date.
	 */
	public boolean getFinished() {
		if (!inbox.isEmpty()) {
			finished = false;
		}
		return finished;
	}

	/**
	 * 
	 */
	public void makeUnfinished() {
		finished = false;
	}

	/**
	 * Adds the message to inbox of this Agent and wakes it if it is sleeping.
	 * 
	 * <p> After receiving the message the Agent is activated by World and 
	 * added to the end of the execution queue. 
	 * </p>
	 *
	 * @param m Message that should be added to inbox
	 * @since 2013-01-07
	 * @verification TEST 1 Warning if the Agent receives messages in the first day 
	 * if its life, when it is not yet ready to handle them.
	 */
	public void receiveMessage(Message m) {
		inbox.addMessage(m);
		if (greenhorn) {
			// *****VERIFICATION MODE START*****//
			//***TEST 1*// 
			if (World.verificationMode && m.topic.equalsIgnoreCase("dead_agent") && m.topic.equalsIgnoreCase("ploughing")) {
				// ***TEST 1
				try {
					throw new SkyPigException("Greenhorn woken up.");
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
			// *****VERIFICATION MODE END*****//
			return;
		}
		finished = false;
		World.Instance().activateAgent(this);
		return;
	}

	/**
	 * Returns id of this Agent.
	 *
	 * @return id of this Agent
	 */
	public long getId() {
		return id;
	}

	/**
	 * 
	 * @return
	 */
	public boolean getGreenhorn() {
		return greenhorn;
	}

	/**
	 * Informs if this Agent is alive.
	 *
	 * @return true if the Agent is alive, false otherwise
	 */
	public boolean getAlive() {
		return alive;
	}

	/**
	 * Used to create the Signature with all the relevant information from this
	 * Agent that should be stored in database. Agents add to the Signature
	 * following values: id, alive, state (state name), myDate (currentDate),
	 * greenhorn.
	 *
	 * @return the Agent's signature
	 */
	public Signature getSignature() {
		Signature mySignature = new Signature(this);
		mySignature.addPrimitiveField("id", "bigint", Long.toString(id));
		mySignature.addPrimitiveField("alive", "boolean",
				Boolean.toString(alive));
		mySignature.addPrimitiveField("causeOfDeath", "text", "'"
				+ causeOfDeath + "'");
		mySignature.addPrimitiveField("state", "text", "'"
				+ state.getClass().getSimpleName() + "'");
		mySignature.addPrimitiveField("greenhorn", "boolean",
				String.valueOf(greenhorn));
		mySignature.addInternalObject(state.getSignature());
		mySignature.addPrimaryKey("id");
		return mySignature;
	}

	/**
	 * Used to get the class of this Agent state.
	 *
	 * @return the class of this Agent state
	 */
	public Class getStateClass() {
		return state.getClass();
	}

	/**
	 * Used to get this Agent state. Works only for Egg and Nestling state. <p>
	 * As Agents with state Egg and Nestling are not independent their states
	 * need to be used by their Nest for management purposes. This function
	 * should not be called by any other objects then Nest possessing these Eggs
	 * or Nestlings. </p>
	 *
	 * @param id of the Agent requesting the state.
	 * @return this Agent state 
	 * @verification TEST 1 Warning if the object which state is requested 
	 * is not a Nest. 
	 */
	public AgentState getState(long id) {
		// *****VERIFICATION MODE START*****//
		if (World.verificationMode) {
			// ***TEST 1
			Class currentClass = World.Instance().getAgent(id).getClass();
			HashSet<String> classNames = new HashSet<String>();
			while (currentClass != null) {
				classNames.add(currentClass.getSimpleName());
				currentClass = currentClass.getSuperclass();
			}
			boolean isThatNest = false;
			for (String name : classNames) {
				if (name.equalsIgnoreCase("Nest")) {
					isThatNest = true;
				}
			}
			if (!isThatNest) {
				try {
					throw new SkyPigException(
							"Dangerous request for state from: "
							+ World.Instance().getAgent(id).getClass().getSimpleName() + " "
							+ String.valueOf(id));
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
		// *****VERIFICATION MODE END*****//
		Class currentClass = state.getClass();
		while (currentClass != null) {
			if (currentClass.getSimpleName().equalsIgnoreCase("Egg")
					|| currentClass.getSimpleName().equalsIgnoreCase("Nestling")) {
				return state;
			}
			currentClass = currentClass.getSuperclass();
		}
		return null;
	}

	/**
	 * Exchange the state of this Agent for a new AgentState. Works only for Egg
	 * and Nestling state. <p> As Agents with state Egg and Nestling are not
	 * independent their states need to be used by their Nest for management
	 * purposes. This function should not be called by any other objects then
	 * Nest possessing these Eggs or Nestlings. </p>
	 *
	 */
	public void exchangeState() {
		Class currentClass = state.getClass();
		while (currentClass != null) {
			if (currentClass.getSimpleName().equalsIgnoreCase("Egg")
					|| currentClass.getSimpleName().equalsIgnoreCase("Nestling")) {
				state = state.changeState();
				break;
			}
			currentClass = currentClass.getSuperclass();
		}
	}

	/**
	 * Used to set this Agent state with the parameter state if it is null.
	 *
	 * @param state to be set
	 * @return true iif this Agent state has been set, false if it was set
	 * earlier
	 */
	public boolean initialiseState(AgentState state) {
		if (this.state == null) {
			this.state = state;
			return true;
		} else {
			return false;
		}
	}

	/**
	 * In derived classes used to create an AgentState according to the
	 * specification passed in a HashMap
	 *
	 * @param parameters a HashMap with names of parameters as keys and values
	 * as their values.
	 * @return created AgentState.
	 */
	public abstract AgentState createAgentState(
			HashMap<String, Object> parameters);

	/**
	 * 
	 */
	public final void resetState() {
		if (ready.compareTo(World.getDate()) < 0) {
			ready = World.getDate();
			state.reset();
		}
	}

	/**
	 * 
	 */
	public class AgentException extends Exception {
	}
}
