//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Level;

/**
 * Represents Agents of type Animal.
 *
 * <p> Animals are characterized by their age, sex, bodyweight, they may have
 * territory, partner and nest. Their bodyweight is splitted among basal
 * bodyweight and energy reserves. Animals calculate their energy requirements
 * using their state, and gather all the obtained resources that are later used
 * in state functions. </p>
 *
 * @author Katarzyna Kułakowska
 * @author Tomasz Kułakowski
 * @version 1.0
 *
 */
public abstract class Animal extends Agent {

	/**
	 * State containing all the functions necessary for Animals, irrespective of
	 * their species and life stage.
	 *
	 * <p> Contains functions necessary for managing energy budget of an animal
	 * and for serving the fields of Animal. </p>
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @version 1.0
	 */
	public abstract class AnimalState extends AgentState {

		/**
		 * Energy lost for daily maintenance today in kJ.
		 */
		public double energyLoss;
		/**
		 * Energy required for growth today in kJ.
		 */
		public double growthPossibilityAndCost;
		/**
		 * Energy possible to store in energy reserves today in kJ.
		 */
		public double reservesStorePossibilityAndCost;
		/**
		 *
		 */
		public double reproductionBuffer;
		/**
		 *
		 */
		public double growthEnergy;
		/**
		 * Energy for maintenance today, not reset during feeding, only reset in
		 * reset function.
		 */
		public double metabolicRate;
		/**
		 * 
		 */
		public double activityEnergy;
		/**
		 *
		 */
		public int changesInBodyweight;

		/**
		 * Constructor for AnimalState does not require any parameters in the
		 * HashMap.
		 *
		 * @param parameters that can be used for parameterization if necessary.
		 */
		public AnimalState(HashMap<String, Object> parameters) {
			super();
			changesInBodyweight = 0;
		}

		@Override
		public void manageEnergy() {
			managePesticides();
			calculateFoodRequirements();
		}

		/**
		 *
		 */
		public void managePesticides() {
			for (String pest : pesticidesForEffects.keySet()) {
				pesticidesForEffects.get(pest).pollFirst();
				if (pesticides.keySet().contains(pest)) {
					pesticidesForEffects.get(pest).add(1000.0 * pesticides.get(pest) / getBodyweight());
				} else {
					pesticidesForEffects.get(pest).add(0.0);
				}
			}
			checkEffects();
			if (getDateOfLaying().get(GregorianCalendar.YEAR) < World.getDate().get(GregorianCalendar.YEAR)
					|| getDaysSinceLeavingNest() < 0
					|| (((String) Config.getParameter("World_speciesName")).equalsIgnoreCase("Skylark")
					&& getDaysSinceLeavingNest() <= (Integer) Config.getParameter("SkylarkJuvenile_feedingLength"))) {
				for (String pesticide : pesticides.keySet()) {
					if (maxPesticides.get(pesticide) < 1000.0 * pesticides.get(pesticide) / getBodyweight()) {
						maxPesticides.put(pesticide, 1000.0 * pesticides.get(pesticide) / getBodyweight());
					}
					exposureTime.put(pesticide, exposureTime.get(pesticide) + 1);
					cumulativePesticides.put(pesticide, cumulativePesticides.get(pesticide)
							+ 1000.0 * pesticides.get(pesticide) / getBodyweight());
				}
			}
			Iterator<Map.Entry<String, Double>> iterator = pesticides.entrySet().iterator();
			while (iterator.hasNext()) {
				Map.Entry<String, Double> pesticide = iterator.next();
//				checkEffects(pesticide);
				Double newAmount;
				if (((Double) Config.getParameter(pesticide.getKey()
						+ "_animalHalfLife")) > 0) {
					newAmount = pesticide.getValue()
							* Math.pow(0.5, (1.0 / (Double) Config.getParameter(pesticide.getKey()
							+ "_animalHalfLife")));
				} else {
					newAmount = 0.0;
				}
				if (newAmount / (getBodyweight() / 1000)
						<= (Double) Config.getParameter(pesticide.getKey() + "_minimumDetectableMgPerKgBw")) {
					iterator.remove();
				} else {
					pesticides.put(pesticide.getKey(), newAmount);
				}
			}
		}

		/**
		 * 
		 * @return
		 */
		public boolean checkMortalityEffect() {
			if (effectsMagnitude.containsKey("mortalityEffect")) {
				if (Config.getGenerator().nextDouble() < effectsMagnitude.get("mortalityEffect")) {
					return true;
				} else {
					return false;
				}
			} else {
				return false;
			}
		}

		/**
		 * 
		 */
		public void checkEffects() {
			Iterator<Map.Entry<String, Boolean>> iterator = effects.entrySet().iterator();
			while (iterator.hasNext()) {
				String effect = iterator.next().getKey();
				if (effect != null) {
					for (String p : pesticidesForEffects.keySet()) {
						double x = 0.0;
						if (effect.contains("time")) {
							for (double d : pesticidesForEffects.get(p)) {
								if (d > 0) {
									x += 1.0;
								}
							}
						} else if (effect.contains("cumulative")) {
							for (double d : pesticidesForEffects.get(p)) {
								x += d;
							}
						} else if (effect.contains("maxdose")) {
							for (double d : pesticidesForEffects.get(p)) {
								if (d > x) {
									x = d;
								}
							}
						}
						if (x > 0) {
							double effectProbability = 1 / (1 + Math.exp((-x
								+ (Double) Config.getParameter(p + "_" + effect + "_ec50"))
								/ (Double) Config.getParameter(p + "_" + effect + "_slope")));
						
							effectsMagnitude.put(effect.split("_")[0], effectsMagnitude.get(effect.split("_")[0])
									+ (1 - effectsMagnitude.get(effect.split("_")[0]))
									* effectProbability);
							//		* ((1 / (1 + Math.pow(10.0, ((Double) Config.getParameter(p + "_" + effect + "_slope")
							//		* (Math.log10((Double) Config.getParameter(p + "_" + effect + "_ec50")) - Math.log10(x))))))));
						}
					}
				}
			}
		}

		/**
		 * 
		 */
		public void resetEffects() {
			Iterator<Map.Entry<String, Boolean>> iterator = effects.entrySet().iterator();
			while (iterator.hasNext()) {
				effects.put(iterator.next().getKey(), Boolean.FALSE);
			}
			for (String ef : effectsMagnitude.keySet()) {
				effectsMagnitude.put(ef, 0.0);
			}
		}

		/**
		 * 
		 * @return
		 */
		public boolean checkFeedingEffect() {
			if (effectsMagnitude.containsKey("feedingEffect")) {
				if (Config.getGenerator().nextDouble() < effectsMagnitude.get("feedingEffect")) {
					return true;
				} else {
					return false;
				}
			}
			return false;
			
//			if (effects.containsKey("feedingEffect")) {
//				if (effects.get("feedingEffect") == Boolean.TRUE) {
//					return false;
//				} else {
//					return true;
//				}
//			}
//			return true;
		}

		/**
		 * Resets Animal and all the own fields to 0.
		 */
		@Override
		public void reset() {
			resetAnimal();
			resetEffects();
			energyLoss = 0;
			metabolicRate = 0;
			activityEnergy = 0;
			growthEnergy = 0;
			reproductionBuffer = 0;
			growthPossibilityAndCost = 0;
			reservesStorePossibilityAndCost = 0;
			super.reset();
		}

		/**
		 * 
		 */
		public void resetPesticides() {
			maxPesticides.clear();
			exposureTime.clear();
			cumulativePesticides.clear();
			for (String pesticide : (ArrayList<String>) Config.getParameter("Pesticides")) {
				maxPesticides.put(pesticide, 0.0);
				cumulativePesticides.put(pesticide, 0.0);
				exposureTime.put(pesticide, 0);
			}
		}

		/**
		 * Adds the passed resource to todaysResources in Animal.
		 *
		 * @param r to be added
		 */
		public void addTodaysResource(Resource r) {
			todaysResources.add(r);
			modifyFoodRequirements(r);
		}

		/**
		 *
		 * @return
		 */
		public double calculateMinimumEnergyRequired() {
			if (desiredEnergy == null) {
				double desiredBodyweight = calculateDesiredBodyweight();
//				desiredEnergy = (Double) Config.getParameter("Animal_thresholdBodyweight")
//						* (Double) Config.getParameter("Animal_massToBasalEnergy")
//						+ (desiredBodyweight - (Double) Config.getParameter("Animal_thresholdBodyweight"))
//						* (Double) Config.getParameter("Animal_massToEnergyReserves");
				desiredEnergy = maxBasalEnergy + (desiredBodyweight
						- maxBasalEnergy / (Double) Config.getParameter("Animal_massToBasalEnergy"))
						* (Double) Config.getParameter("Animal_massToEnergyReserves");
			}
			return desiredEnergy - (basalEnergy + energyReserves - energyLoss);
		}

		/**
		 * 
		 * @return
		 */
		public double calculateDesiredBodyweight() {
			return (Double) Config.getParameter("Adult_deathProbability2")
					* maxBasalEnergy / (Double) Config.getParameter("Animal_massToBasalEnergy")
					- ((Double) Config.getParameter("Adult_deathProbability1")
					* Math.log((1 / Math.pow(
					(Double) Config.getParameter("Animal_desiredSurvivalProbability"),
					1 / (Double) Config.getParameter("Animal_desiredSurvivalTime"))) - 1));
		}

		/**
		 * Calculates maximum possible energy needed for growth today.
		 *
		 * <p> Uses mass dependent growth rate at the beginning of growth, and
		 * then constant growth rate. Takes into account protein storage
		 * efficiency. </p>
		 *
		 * @return growthPossibilityAndCost calculated in kJ @parameters
		 * Animal_thresholdBodyweight, Animal_thresholdNestlingBodyweight,
		 * Animal_maxDailyGrowth, Animal_massToBasalEnergy,
		 * Animal_proteinStorageEfficiency, Animal_initialGrowthRatio
		 * @verification TEST 1 Warning if the growth cost was not reset or was 
		 * calculated previously this day.
		 *
		 */
		public double calculateGrowthPossibilityAndCost() {
			// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 1
				if (growthPossibilityAndCost != 0) {
					try {
						throw new SkyPigException(
								"Animal calculating growthCost twice");
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
			}
			// *****VERIFICATION MODE END*****//
			if (basalEnergy < maxBasalEnergy) {
				growthPossibilityAndCost = (Double) Config.getParameter("Animal_growthParameter2")
						* Math.exp(-(Double) Config.getParameter("Animal_growthParameter1")
						* getBodyweight())
						* getBodyweight()
						* (Double) Config.getParameter("Animal_massToBasalEnergy")
						/ (Double) Config.getParameter("Animal_proteinStorageEfficiency");
				// double thresholdNestlingBodyweight = (Double)
				// Config.getParameter("Animal_thresholdNestlingBodyweight");
				// if (getBodyweight() > thresholdNestlingBodyweight) {
				// growthPossibilityAndCost = (Double)
				// Config.getParameter("Animal_maxDailyGrowth")
				// * (Double) Config.getParameter("Animal_massToBasalEnergy")
				// / (Double)
				// Config.getParameter("Animal_proteinStorageEfficiency");
				// } else {
				// growthPossibilityAndCost = (Double)
				// Config.getParameter("Animal_initialGrowthRatio")
				// * getBodyweight()
				// * (Double) Config.getParameter("Animal_massToBasalEnergy")
				// / (Double)
				// Config.getParameter("Animal_proteinStorageEfficiency");
				// }
			}
			growthEnergy = growthPossibilityAndCost;
			return growthPossibilityAndCost;
		}

		/**
		 * Calculates BMR using allometric equation.
		 *
		 * @return BMR in kJ @parameters Animal_bmrA, Animal_bmrB
		 */
		public double calculateMetabolicRate() {
			metabolicRate = (Double) Config.getParameter("Animal_bmrA")
					* Math.pow(getBodyweight(),
					(Double) Config.getParameter("Animal_bmrB"));
			return metabolicRate;
		}

		/**
		 * Calculates and returns daily energyStorePossibilityAndCost.
		 *
		 * <p> Energy can be stored only if an animal has finished growth, that
		 * means, its bodyweight is bigger then Animal_thresholdBodyweight.
		 * Energy calculated here depends on an Animal bodyweight and the energy
		 * that is stored already. The calculation takes into accuont costs of
		 * storing energy. </p>
		 *
		 * @return reservesStorePossibilityAndCost in kJ @parameters
		 * Animal_thresholdBodyweight, Animal_maxReservesStorePossibility,
		 * Animal_energyScallingFactor, Animal_massToEnergyReserves,
		 * Animal_lipidsStorageEfficiency 
		 * 
		 * @verification TEST 1 Warning if the reserves store possibility was not 
		 * reset or was calculated previously this day.
		 */
		public double calculateReservesStorePossibilityAndCost() {
			// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 1
				if (reservesStorePossibilityAndCost != 0) {
					try {
						throw new SkyPigException(
								"Animal calculating reservesStorePossibilityAndCost twice");
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
			}
			// *****VERIFICATION MODE END*****//
			if (getBodyweight() >= maxBasalEnergy / (Double) Config.getParameter("Animal_massToBasalEnergy")
					&& energyReserves < maxReserves) {
				reservesStorePossibilityAndCost = maxReservesStoreSpeed
						* (1 + (Double) Config.getParameter("Animal_lipidStorageEfficiency"));

//				reservesStorePossibilityAndCost = (Double) Config.getParameter("Animal_growthParameter2")
//						* Math.exp(-(Double) Config.getParameter("Animal_growthParameter1")
//						* getBodyweight())
//						* getBodyweight()
//						* (Double) Config.getParameter("Animal_massToEnergyReserves")
//						/ (Double) Config.getParameter("Animal_lipidStorageEfficiency");
				// reservesStorePossibilityAndCost = ((Double)
				// Config.getParameter("Animal_maxReservesStorePossibility")
				// * Math.exp((-getBodyweight() + (Double)
				// Config.getParameter("Animal_thresholdBodyweight"))
				// / (Double)
				// Config.getParameter("Animal_energyScallingFactor")) *
				// (Double) Config.getParameter("Animal_massToEnergyReserves"))
				// / (Double)
				// Config.getParameter("Animal_lipidStorageEfficiency");
			}
			return reservesStorePossibilityAndCost;
		}


		/**
		 * 
		 */
		public void chooseResourcesForNestlings() {
		}


		/**
		 * Allocates energy from the collected food to maintenance, growth,
		 * reproduction and storing reserves.
		 *
		 * <p> Adds the energy from resources stored in todaysResources to
		 * energyObtained. Then allocates it to maintenance, growth,
		 * reproduction and storing reserves (in this order). Additionally, adds
		 * all the pesticides from the Resource to pesticides.
		 *
		 */
		public void eat() {
			compensateEnergyLoss(digestResources(sliceResources(energyLoss)));
			grow(digestResources(sliceResources(growthPossibilityAndCost)));
			if (calculateMinimumEnergyRequired() < 0) {
				chooseResourcesForNestlings();
				storeReproductionBuffer(digestResources(sliceResources(reproductionBuffer)));
				buildEnergyReserves(digestResources(sliceResources(reservesStorePossibilityAndCost)));
			} else {
				buildEnergyReserves(digestResources(sliceResources(reservesStorePossibilityAndCost)));
				chooseResourcesForNestlings();
				storeReproductionBuffer(digestResources(sliceResources(reproductionBuffer)));
			}
		}

		/**
		 *
		 * @param requiredEnergy
		 * @return
		 */
		public Set<Resource> sliceResources(double requiredEnergy) {
			double resourcesEnergy = 0;
			for (Resource r : todaysResources) {
				resourcesEnergy += r.energyContent * r.amount;
			}
			double resourcesProportion = requiredEnergy / resourcesEnergy;
			HashSet<Resource> requiredResources = new HashSet<Resource>();
			for (Resource r : todaysResources) {
				requiredResources.add(r.splitResource(r.amount
						* resourcesProportion));
			}
			return requiredResources;
		}

		/**
		 *
		 * @param resources
		 * @return
		 */
		public double digestResources(Set<Resource> resources) {
			double extractedEnergy = 0;
			Iterator<Resource> iterator = resources.iterator();
			while (iterator.hasNext()) {
				Resource r = iterator.next();
				double resourceEnergy = r.amount * r.energyContent;
				extractedEnergy += resourceEnergy;
				addPesticides(r.getPesticides());
				iterator.remove();
			}
			energyObtained += extractedEnergy;
			return extractedEnergy;
		}

		/**
		 *
		 * @return
		 */
		public double getEnergyOfResources() {
			double energyOfResources = 0;
			for (Resource r : todaysResources) {
				energyOfResources += r.energyContent * r.amount;
			}
			return energyOfResources;
		}

		/**
		 * Adds the passed energy to energyReserves with some efficiency.
		 *
		 * @param energy to be stored @parameter Animal_lipidStorageEfficiency
		 */
		public void buildEnergyReserves(double energy) {
			energyReserves += energy
					* (Double) Config.getParameter("Animal_lipidStorageEfficiency");
		}

		/**
		 *
		 * @param energy
		 */
		public void storeReproductionBuffer(double energy) {
			reproductionBuffer -= energy;
		}

		/**
		 * Allocates the required energy to growth and returns the rest.
		 *
		 * <p> If the Animal bodyweight is below the Animal_thresholdBodyweight
		 * the energy passed is allocated to building proteins up to the limit
		 * of growthPossibilityAndCost, adjusted for efficiency. If the
		 * Animal_thresholdBodyweight is reached during this process before
		 * allocating the energy calculated as growthPossibilityAndCost the rest
		 * is stored as energy reserves. ed as energyLoss the rest is stored as
		 * energy reserves. If the bodyweight of the Animal is above the
		 * Animal_thresholdBodyweight, the whole energy (up to the limit of
		 * growthPossibilityAndCost) is allocated to energyReserves, adjusted
		 * for efficiency. </p>
		 *
		 * @param energy passed to allocate to growth
		 * @return energy left after allocation to growth @parameters
		 * Animal_proteinStorageEfficiency, Animal_thresholdBodyweight,
		 * Animal_massToBasalEnergy, Animal_lipidStorageEfficiency
		 */
		public double grow(double energy) {
			double growthEnergy = 0;
			if (growthPossibilityAndCost > energy) {
				growthEnergy = energy;
				growthPossibilityAndCost -= energy;
				energy = 0;
			} else {
				growthEnergy = growthPossibilityAndCost;
				energy -= growthPossibilityAndCost;
				growthPossibilityAndCost = 0;
			}
			double efficiency = (Double) Config.getParameter("Animal_proteinStorageEfficiency");
//			double thresholdEnergy = (Double) Config.getParameter("Animal_thresholdBodyweight")
//					* (Double) Config.getParameter("Animal_massToBasalEnergy");
			if (basalEnergy < maxBasalEnergy) {
				if (growthEnergy * efficiency > maxBasalEnergy - basalEnergy) {
					basalEnergy = maxBasalEnergy;
					growthEnergy -= (maxBasalEnergy - basalEnergy) / efficiency;
					energyReserves += growthEnergy * (Double) Config.getParameter("Animal_lipidStorageEfficiency");
				} else {
					basalEnergy += growthEnergy * efficiency;
					growthEnergy = 0;
				}
			}
			return energy;
		}

		/**
		 * Allocates the passed energy to compensate energy used previously on
		 * maintenance.
		 *
		 * <p> If the Animal bodyweight is below the Animal_thresholdBodyweight
		 * the energy passed is allocated to building proteins up to the limit
		 * of energyLoss, adjusted for efficiency. If the
		 * Animal_thresholdBodyweight is reached during this process before
		 * allocating the energy calculated as energyLoss the rest is stored as
		 * energy reserves. If the bodyweight of the Animal is above the
		 * Animal_thresholdBodyweight, the whole energy (up to the limit of
		 * energyLoss) is allocated to energyReserves, adjusted for efficiency.
		 * </p>
		 *
		 * @param energy passed to allocate to compensate energy loss
		 * @return energy left after allocating for compensation of energy loss
		 * @parameters Animal_proteinStorageEfficiency,
		 * Animal_thresholdBodyweight, Animal_massToBasalEnergy,
		 * Animal_lipidStorageEfficiency"
		 */
		public double compensateEnergyLoss(double energy) {
			if (energyLoss > energy) {
				energyLoss -= energy;
				energy = 0;
			} else {
				energy -= energyLoss;
				energyLoss = 0;
			}
			spendEnergy();
			if (energyLoss > 0 && getBodyweight() < calculateDesiredBodyweight()) {
				changesInBodyweight += 1;
			} else {
				changesInBodyweight = 0;
			}
			return energy;
		}

		/**
		 * Calculates and loose energy for maintenance (BMR + activities).
		 * Calculates energy required for growth.
		 *
		 * <p> If possible, energy loss is subtracted from energy reserves. If
		 * energyReserves are empty, it is subtracted from basalEnergy. </p>
		 *
		 *
		 */
		public void spendEnergy() {
			if (energyReserves > energyLoss) {
				energyReserves -= energyLoss;
			} else {
				double basalMassLoss = energyLoss - energyReserves;
				energyReserves = 0;
				basalEnergy -= basalMassLoss;
			}
		}

		/**
		 * Spends energy, then calculates the total food requirements: for
		 * maintenance, growth (together with growth costs) and storing reserves
		 * (together with cost).
		 */
		public void calculateFoodRequirements() {
			energyLoss = calculateMetabolicRate() + activityEnergy;
			calculateGrowthPossibilityAndCost();
			calculateReservesStorePossibilityAndCost();
			if (foodRequirements == 0) {
				foodRequirements = energyLoss + reproductionBuffer
						+ growthPossibilityAndCost
						+ reservesStorePossibilityAndCost;
				animalEnergyRequirements = foodRequirements;
			}
		}

		/**
		 *
		 * @return
		 */
		public double getEnergyReserves() {
			return energyReserves;
		}

		/**
		 *
		 * @param newPesticides
		 */
		public void addPesticides(HashMap<String, Double> newPesticides) {
			for (String pesticide : newPesticides.keySet()) {
				Double amount = newPesticides.get(pesticide);
				if (pesticides.containsKey(pesticide)) {
					amount += pesticides.get(pesticide);
				}
				pesticides.put(pesticide, amount);
			}
		}
	}

	/**
	 * An Animal state for Animals managed by the Nest. The only implemented
	 * method for this state is changeState.
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @version 1.0
	 */
	public class Egg extends AnimalState {

		/**
		 * Constructor for Egg does not require any parameters in the HashMap.
		 *
		 * @param parameters that can be used for parameterization if necessary.
		 */
		public Egg(HashMap<String, Object> parameters) {
			super(parameters);
		}

		@Override
		public boolean checkDeath(String causeOfDeath) {
			if (causeOfDeath == null) {
				if (Config.getGenerator().nextDouble() < (Double) Config.getParameter("Egg_infertilityProbability")) {
					return super.checkDeath("egg_infertility");
				} else {
					return false;
				}
			} else {
				return super.checkDeath(causeOfDeath);
			}
		}

		/**
		 * 
		 * @param effect
		 * @return
		 */
		public boolean checkHatchingEffect(double effect) {
			if (Config.getGenerator().nextDouble() < (effect)) {
				return checkDeath("eggsHatchingEffect");
			} else {
				return false;
			}
		}

		@Override
		public void findFood() {
			return;
		}

		@Override
		public void manageEnergy() {
			return;
		}

		/**
		 * Returns new Nestling AgentState.
		 *
		 * @return Nestling AgentState
		 */
		@Override
		public AgentState changeState() {
			HashMap<String, Object> parameters = new HashMap<String, Object>();
			parameters.put("stateName", "Nestling");
			return createAgentState(parameters);
		}
	}

	/**
	 * An Animal state for Animals managed by the Nest. The only implemented
	 * methods for this state are changeState and checkDeath.
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @version 1.0
	 */
	public class Nestling extends AnimalState {

		/**
		 * Constructor for Nestling does not require any parameters in the
		 * HashMap.
		 *
		 * @param parameters that can be used for parameterization if necessary.
		 */
		public Nestling(HashMap<String, Object> parameters) {
			super(parameters);
			hatchingDate = World.getDate();
		}

		/**
		 * Checks if the Nestling should die from starvation.
		 *
		 * For juveniles: death probability = Animal_bmrDeathProbability *exp(-a
		 * * (energyObtained - metabolicRate)/(animalEnergyRequirements -
		 * metabolicRate)), where a =
		 * -ln((1-Animal_desiredSurvivalProbability^(1/
		 * Animal_juvenileDesiredSurvivalTime)) /Animal_bmrDeathProbability)
		 *
		 * @param causeOfDeath
		 * @return true if the Nestling has died, false otherwise.
		 */
		@Override
		public boolean checkDeath(String causeOfDeath) {
			if (causeOfDeath == null) {
				if (checkMortalityEffect()) {
					return super.checkDeath("mortalityEffect");
				} else {
					double survivalProbability =
							1 - (Double) Config.getParameter("Animal_bmrDeathProbability")
							* Math.pow((1 - Math.pow(
							(Double) Config.getParameter("Animal_desiredSurvivalProbability"),
							1 / ((Double) Config.getParameter("Animal_juvenileDesiredSurvivalTime"))))
							/ (Double) Config.getParameter("Animal_bmrDeathProbability"),
							(energyObtained - metabolicRate) / (animalEnergyRequirements - metabolicRate));
					if (Config.getGenerator().nextDouble() > survivalProbability) {
						return super.checkDeath("badCondition");
					} else {
						return false;
					}
				}
			} else {
				return super.checkDeath(causeOfDeath);
			}
		}

		/**
		 * 
		 * @param effect
		 * @return
		 */
		public boolean checkSurvivalEffect(double effect) {
			if (Config.getGenerator().nextDouble() < (effect)) {
				return checkDeath("NestlingsSurvivalEffect");
			} else {
				return false;
			}
		}

		@Override
		public void findFood() {
			return;
		}

		/**
		 * Returns new Adult AgentState.
		 *
		 * @return Adult AgentState
		 */
		@Override
		public AgentState changeState() {
			HashMap<String, Object> parameters = new HashMap<String, Object>();
			parameters.put("stateName", "Adult");
			return createAgentState(parameters);
		}
	}

	/**
	 * An AnimalState for Adults.
	 *
	 * <p> Specifies the seasonal activities for Adult Animals: searching for
	 * territory and partner, building nest and reproduction, and winter
	 * behaviour. The method for reading messages is implemented here. The
	 * management of energy and dicviding the obtained resources between this
	 * Adult and its offspring is also done here. <p>
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @version 1.0
	 */
	public abstract class Adult extends AnimalState {

		/*
		 * Fields:
		 */
		/**
		 * Resources separated from obtainedResources and allocated for the
		 * offsrping.
		 */
		public Set<Resource> resourcesForNest;
		/**
		 *
		 */
		public double incubationForNest;
		/**
		 * Energy required by the offspring.
		 */
		public double nestRequirements;
		/**
		 *
		 */
		public boolean incubationRequired;
		/**
		 * SeasonalActivity currently conducted byt his Adult.
		 */
		private SeasonalActivity activity;
		/**
		 *
		 */
		public boolean growing;
		/**
		 * 
		 */
		public Calendar lastBroodEnd;
		/**
		 * 
		 */
		public Integer lastRenestingInterval;

		/*
		 * Methods:
		 */
		/**
		 * Constructor for an Adult AnimalState.
		 *
		 * <p> Constructs and Adult AnimalStae with activity set to OffSeason.
		 * Does not require any parameters in the HashMap. </p>
		 *
		 * @param parameters
		 */
		public Adult(HashMap<String, Object> parameters) {
			super(parameters);
			if (!parameters.containsKey("activityName")) {
				parameters.put("activityName", "OffSeason");
			}
			activity = createSeasonalActivity(parameters);
			incubationRequired = false;
			incubationForNest = 0;
			resourcesForNest = new HashSet<Resource>();
			setNestLeavingDate();
			growing = true;
			lastBroodEnd = null;
			lastRenestingInterval = null;
		}

		/**
		 *
		 * For juveniles: death probability = Animal_bmrDeathProbability *exp(-a
		 * * (energyObtained - metabolicRate)/(animalEnergyRequirements -
		 * metabolicRate)), where a =
		 * -ln((1-Animal_desiredSurvivalProbability^(1/
		 * Animal_juvenileDesiredSurvivalTime)) /Animal_bmrDeathProbability)
		 *
		 * For adults: logistic function for checking death probability in
		 * dependance on Bodyweight.
		 *
		 * @param causeOfDeath
		 * @return true (super.checkDeath()) if the Animal is dead, false
		 * otherwise. @parameters Adult_deathProbability1,
		 * Adult_deathProbability2,
		 */
		@Override
		public boolean checkDeath(String causeOfDeath) {
			if (causeOfDeath == null) {
				if (checkMortalityEffect()) {
					return super.checkDeath("mortalityEffect");
				} else if (checkIfShot()) {
					return super.checkDeath("shooting");
				} else {
					if (growing) {
						double survivalProbability = 1
								- (Double) Config.getParameter("Animal_bmrDeathProbability")
								* Math.pow(
								(1 - Math.pow((Double) Config.getParameter("Animal_desiredSurvivalProbability"),
								1 / ((Double) Config.getParameter("Animal_juvenileDesiredSurvivalTime"))))
								/ (Double) Config.getParameter("Animal_bmrDeathProbability"),
								(energyObtained - metabolicRate)
								/ (animalEnergyRequirements - metabolicRate));

						if (getBodyweight() < (Double) Config.getParameter("Adult_minimalBodyweight")) {
							survivalProbability = 0;
						}
						if (Config.getGenerator().nextDouble() > survivalProbability) {
							return super.checkDeath("starvation");
						} else {
							return false;
						}
					} else {
						double survivalProbability = 1 / (1 + Math.exp((-getBodyweight()
								+ (Double) Config.getParameter("Adult_deathProbability2")
								* maxBasalEnergy / (Double) Config.getParameter("Animal_massToBasalEnergy"))
								/ (Double) Config.getParameter("Adult_deathProbability1")));
						if (Config.getGenerator().nextDouble() > survivalProbability) {
							return super.checkDeath("lowBodyweight");
						} else {
							return false;
						}
					}
				}
			} else {
				return super.checkDeath(causeOfDeath);
			}
		}
		
		public boolean checkIfShot() {
			return false;
		}
		
		/**
		 * Resets nestRequirements to 0 and clears resourcesForNest, then calls
		 * super.reset().
		 */
		@Override
		public void reset() {
			resourcesForNest.clear();
			incubationForNest = 0;
			nestRequirements = 0;
			incubationRequired = false;
			if (growing && energyReserves > 0) {
				growing = false;
			}
			lastRenestingInterval = null;
//			if (World.getDate().get(Calendar.DAY_OF_YEAR)
//					== (Integer) Config.getParameter("SeasonalActivity_startOfSeason") - 1) {
//				lastBroodEnd = null;
//			}
			super.reset();
		}

		/**
		 * 
		 */
		public void resetLastBroodEnd() {
			lastBroodEnd = null;
		}

		/**
		 * Reads the passed Message, takes the proper action in response and
		 * replies to the Message.
		 *
		 * <p> Topics of Messages read: </p> <p> "feeding" - returns the message
		 * with topic "re:feeding" and resourcesForNest attached. </p> <p>
		 * "incubation" - returns the message with topic "re:incubation" and
		 * null attachment. </p> <p> "partner" - if this Animal has no partner,
		 * it set the partnerId to the Id of sender of the Message, adds it to
		 * contact list and returns the message with the topic "re:partner" and
		 * true attached. If it has a partner, it returns the Message with
		 * topic: "re:partner" and false attached. </p> <p> "re:partner" - if
		 * this Animal has no partner and the Messages attachment idicates true,
		 * it set the partnerId to the Id of sender of the Message, adds it to
		 * contact list and assigns territory. Othrewise it does nothing.
		 * Returns null </p> <p> "nest" - if this Animal has no nest, it set the
		 * nestId to the attachemt of this Message, adds it to contact list and
		 * returns the message with the topic "re:nest" and true attached. If it
		 * has a nest, it returns the Message with topic: "re:nest" and false
		 * attached. </p> <p> "re: nest" - returns null. </p> <p> "dead_agent"
		 * or "remove_agent" - removes the sender from contacts and forwards the
		 * message to super.readMessage. </p> <p> "re:dead_agent" or
		 * "re:remove_agent" - returns null. </p> <p> If a Message with any
		 * other topic is received, it is forwarded to super.readMessage. </p>
		 *
		 * @param m Message to be read
		 * @return Message to be send as reply or null if the passed Message
		 * does not require an answer. 
		 * @verification TEST 1 Warning if an animal has another nest while 
		 * getting a nest proposal.
		 */
		@Override
		public Message readMessage(Message m) {
			if (m.topic.equalsIgnoreCase("feeding")) {
				return new Message(getId(), m.senderId, "re:feeding", m,
						resourcesForNest);
			} else if (m.topic.equalsIgnoreCase("incubation")) {
				return new Message(getId(), m.senderId, "re:incubation", m,
						incubationForNest);
			} else if (m.topic.equalsIgnoreCase("partner")) {
				if (getPartner() == null && getNest() == null) {
					partnerId = (Long) m.senderId;
					addToContactList(partnerId);
					return new Message(getId(), m.senderId, "re:partner", m, getTerritory());
				}
				return new Message(getId(), m.senderId, "re:partner", m, null);
			} else if (m.topic.equalsIgnoreCase("re:partner")) {
				if (m.attachment != null) {
					partnerId = (Long) m.senderId;
					addToContactList(partnerId);
					assignTerritory(m.attachment);
				}
				return null;
			} else if (m.topic.equalsIgnoreCase("nest")) {
				if (getNest() == null) {
					HashMap<String, Object> nestProperties = ((HashMap<String, Object>) m.attachment);
					nestProperties.put("maleOwnerId", getId());
					return new Message(getId(), m.senderId, "re:nest", m,
							nestProperties);
				} else {
					// //*****VERIFICATION MODE START*****//
					if (World.verificationMode) {
						// ***TEST 1
						try {
							throw new SkyPigException(
									"Animal who got nest proposal has a nest.");
						} catch (SkyPigException ex) {
							World.log(Level.WARNING,
									Config.getLogInfo(), ex);
						}
					}
					// //*****VERIFICATION MODE END*****//
					return new Message(getId(), m.senderId, "re:nest", m, false);
				}
			} else if (m.topic.equalsIgnoreCase("re:nest")) {
				Nest myNest = World.Instance().speciesFactory.createNest((HashMap<String, Object>) m.attachment);
				nestId = myNest.getId();
				allBroods++;
				addToContactList(nestId);
				myNest.launchMe();
				sendMessage(new Message(getId(), m.senderId, "our_nest", nestId));
				return null;
			} else if (m.topic.equalsIgnoreCase("dead_agent")
					|| m.topic.equalsIgnoreCase("remove_agent")) {
				removeFromContacts(m.senderId);
				return super.readMessage(m);
			} else if (m.topic.equalsIgnoreCase("re:dead")) {
				if (m.inReplyTo.topic.equalsIgnoreCase("nest")) {
					Nest myNest = World.Instance().speciesFactory.createNest((HashMap<String, Object>) m.inReplyTo.attachment);
					nestId = myNest.getId();
					allBroods++;
					addToContactList(nestId);
					myNest.launchMe();
					return null;
				} else {
					return super.readMessage(m);
				}
			} else if (m.topic.equalsIgnoreCase("our_nest")) {
				nestId = (Long) m.attachment;
				allBroods++;
				addToContactList(nestId);
				return new Message(getId(), m.senderId, "re:our_nest", m);
			} else if (m.topic.equalsIgnoreCase("re:our_nest")) {
				return null;
			} else if (m.topic.equalsIgnoreCase("breeding_success")) {
				successfulBroods++;
				return new Message(getId(), m.senderId, "re:breeding_success", m);
			} else {
				return super.readMessage(m);
			}
		}

		/**
		 * Removes references to id passed if it is stored as partnerId or
		 * nestId and set null instead.
		 *
		 * @param senderId Id to be removed from contacts
		 */
		private void removeFromContacts(long senderId) {
			if (partnerId == null ? partnerId == null : senderId == partnerId) {
				partnerId = null;
			}
			if (nestId == null ? nestId == null : senderId == nestId) {
				nestId = null;
			}
		}

		/**
		 * 
		 * @return
		 */
		public abstract Object getTerritory();

		@Override
		public Signature getSignature() {
			Signature mySignature = super.getSignature();
			mySignature.addPrimitiveField("activity", "text", "'"
					+ getActivityName() + "'");
			//if (((Nest)World.Instance().getAgent(nestId)).getDaysSinceLaying() == 0) {
			mySignature.addPrimitiveField("renesting", "integer", String.valueOf(lastRenestingInterval));
			//}
			return mySignature;
		}

		/**
		 * Checks if the present activity should be exchanged for another
		 * SeasonalActivity and conducts the current activity.
		 */
		@Override
		public void conductSeasonalActivity() {
			activity = activity.nextSeasonalActivity();
			activity.conduct();
		}

		/**
		 * In derived classes should return a new SeasonalActivity according to
		 * specification passed as parameters.
		 *
		 * @param parameters
		 * @return null
		 */
		public abstract SeasonalActivity createSeasonalActivity(
				HashMap<String, Object> parameters);

		/**
		 * If this Adult has nestlings, checks their foodrequirements and adds
		 * to foodRequirements that should be collected today.
		 *
		 * <p> If an adult has a Nest with state Nestlings or Eggs on hatching
		 * day it checks if there is a message from Nest with the request for
		 * food for nestlings and waits until it has appeared. Then adds the
		 * requirements specified in the message attachemtn to its own
		 * foodRequirements and additionally save them to nestRequirements. </p>
		 *
		 * @param m @verification TEST 1 checks if the nestRequirements are set
		 * to 0 before adding the new requirements.
		 */
		public void acceptNestFoodRequirements(Message m) {
			if (effects.containsKey("feedingEffect")) {
				if (effects.get("feedingEffect") == Boolean.TRUE) {
					return;
				}
			}
			if (m.topic.equalsIgnoreCase("incubation")) {
				if (effects.containsKey("incubationEffect")) {
					if (effects.get("incubationEffect") == Boolean.TRUE) {
						return;
					}
				}
				incubationRequired = true;
			} else if (m.topic.equalsIgnoreCase("feeding")) {
				if (effects.containsKey("feedingNestlingsEffect")) {
					if (effects.get("feedingNestlingsEffect") == Boolean.TRUE) {
						return;
					}
				}
				nestRequirements = (Double) m.attachment;
			}
		}

		// if (nestId != null) {
		// if (World.Instance().getAgent(nestId) != null) {
		// Class currentClass =
		// ((World.Instance().getAgent(nestId)).getStateClass());
		// while (currentClass != null) {
		// if (currentClass.getSimpleName().equalsIgnoreCase(
		// "Nestlings")
		// || (currentClass.getSimpleName().equalsIgnoreCase("Eggs")
		// && ((Nest) World.Instance().getAgent(nestId)).getDaysSinceHatching()
		// == 0)) {
		// nestRequirements = ((Nest)
		// World.Instance().getAgent(nestId)).getNestFoodRequirementsPerParent();
		// } else if (currentClass.getSimpleName().equalsIgnoreCase("Eggs")
		// && ((Nest) World.Instance().getAgent(nestId)).getDaysSinceHatching()
		// != 0
		// && ((Nest) World.Instance().getAgent(nestId)).getDaysSinceLaying() !=
		// 0) {
		// incubationRequired = true;
		// }
		// currentClass = currentClass.getSuperclass();
		// }
		// }
		// }
		// }
		/**
		 *
		 */
		public void addNestFoodRequirements() {
			foodRequirements += nestRequirements;
		}

		/**
		 * Divides obtained set of Resources between parents and nestlings. <p>
		 * Each Resource from the set
		 * <code>todaysResources</code> is divided accordingly to requirements.
		 * First parents take into account energy they need to make up for
		 * todays' obligatory loss (energyLoss) and 'reseve' appropriate amount
		 * of energy for themselves. Then they check if the rest exceeds the
		 * nestlings requirements. If no, they assign the whole rest (proportion
		 * of each Resource) for nestlings. Else - they assign for nestlings
		 * exactly as much as they wanted, and take the rest for themselves to
		 * build energy reserves. If the parents' bodyweight is below a
		 * threshold bodyweight (it has no energy reserves) nestlings are not
		 * fed at all. </p> <p> Resources for nestlings come from splitting the
		 * resources collected today (todaysResources) and are saved as
		 * resourcesForNest. </p>
		 *
		 *
		 */
		public void chooseResourcesForNestlings() {
			double resourcesEnergy = 0;
			for (Resource r : todaysResources) {
				resourcesEnergy += r.energyContent * r.amount;
			}
			double foodForNestlings = 0;
			if ((resourcesEnergy > calculateMinimumEnergyRequired()) && resourcesEnergy > 0) {
				if (resourcesEnergy - calculateMinimumEnergyRequired() >= nestRequirements) {
					foodForNestlings = nestRequirements;
				} else {
					foodForNestlings = resourcesEnergy
							- calculateMinimumEnergyRequired();
				}
			} else {
				return;
			}
			double resourceNestProportion = foodForNestlings / resourcesEnergy;
			for (Resource r : todaysResources) {
				resourcesForNest.add(r.splitResource(r.amount
						* resourceNestProportion));
			}
		}

		@Override
		public void calculateFoodRequirements() {
			activity.calculateReproductionBuffer();
			calculateActivityEnergy();
			super.calculateFoodRequirements();
		}

		/**
		 * Calculates energy needed for activities above BMR.
		 *
		 * <p> Together with BMR constitutes energy for maintenance. </p>
		 *
@parameters stateName_activityEnergy
		 */
		public void calculateActivityEnergy() {
			activityEnergy = (Double) Config.getParameter(getStateClass().getSimpleName()
					+ "_activityEnergy");
		}

		/**
		 * Returns the name of the current activity.
		 *
		 * @return name of acticity
		 */
		public String getActivityName() {
			return activity.getClass().getSimpleName();
		}

		/**
		 * Should assign the territory according to the rules specified in
		 * derived classes.
		 * 
		 * @param territory 
		 */
		public abstract void assignTerritory(Object territory);

		/**
		 * Should abandon the territory according to the rules specified in
		 * derived classes.
		 */
		public abstract void abandonTerritory();

		/**
		 * Checks the if the condition of the animal is sufficient to enter the
		 * breeding season.
		 *
		 *
		 * @return true if the condition is sufficient, false otherwise
		 * @since 2012-06-28
		 */
		public abstract boolean checkBreedingCondition();

		/**
		 * SeasonalActivity should specify the behaviour that should not be
		 * executed daily, but is specific to some seasons or parts of season.
		 * <p> In the model there are SeasonalActivities that represent the
		 * behaviour within breeding season (FindTerritory, FindPartner,
		 * WaitingMale, WaitingFemale, Reproduce) and outside of the
		 * breedingSeason(OffSeason). </p>
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @version 1.0
		 */
		public abstract class SeasonalActivity {

			/**
			 *
			 */
			public double reproductionBufferSum;
			/**
			 *
			 */
			public double reproductionBufferSize;

			/**
			 * Constructor for SeasonalActivity does not require any parameters
			 * in the HashMap.
			 *
			 * @param parameters that can be used for parameterization if
			 * necessary.
			 */
			public SeasonalActivity(HashMap<String, Object> parameters) {
			}

			/*
			 * Methods:
			 */
			/**
			 * In derived classes should specify the details of animals
			 * behaviour. Here it does nothing.
			 */
			public void conduct() {
				return;
			}

			/**
			 * In derived classes should specify the next seasonalActivoty
			 * according to a choosing algorithm. Here it returns the same
			 * SeasonalActivity.
			 *
			 * @return the same SeasonalActivity
			 */
			public SeasonalActivity nextSeasonalActivity() {
				return this;
			}

			/**
			 * Checks if there is a breeding season.
			 *
			 * @return true if currentDate is in breeding season,false otherwise
			 * @parameters SeasonalActivity_startOfSeason,
			 * SeasonalActivity_endOfSeason
			 */
			public boolean inSeason() {
				if (World.getDate().get(Calendar.DAY_OF_YEAR) > (Integer) Config.getParameter("SeasonalActivity_startOfSeason")
						&& World.getDate().get(Calendar.DAY_OF_YEAR) < (Integer) Config.getParameter("SeasonalActivity_endOfSeason")) {
					return true;
				} else {
					return false;
				}
			}

			/**
			 *
			 */
			public void loadReproductionBuffer() {
				reproductionBufferSum += (reproductionBufferSize - reproductionBuffer);
			}

			/**
			 *
			 */
			public void calculateReproductionBuffer() {
			}
		}

		/**
		 * Specifies the behaviour of birds outside of the breeding season.
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @since 2012-06-28
		 */
		public class OffSeason extends SeasonalActivity {

			/**
			 * Changes the
			 * <code>partnerId</code> to null and abandons territory if the
			 * Animal has one.
			 *
			 * @param parameters (not required) 
			 * @verification Warning if the animal has nest.
			 * @since 2012-06-28
			 */
			public OffSeason(HashMap<String, Object> parameters) {
				super(parameters);
				// *****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					// ***TEST 1
					if (nestId != null) {
						try {
							throw new SkyPigException(
									"Adult going offSeason has nest.");
						} catch (SkyPigException ex) {
							World.log(Level.WARNING,
									Config.getLogInfo(), ex);
						}
					}
				}
				// //*****VERIFICATION MODE END*****//
				if (partnerId != null) {
					sendMessage(new Message(getId(), partnerId, "remove_agent"));
				}
				if (hasTerritory()) {
					abandonTerritory();
				}
				partnerId = null;
				resetLastBroodEnd();
			}

			/**
			 * Specifies the next SeasonalActivity.
			 *
			 * <p> The Seasonal Activity is changed from
			 * <code>OffSeason</code> if the bird is after its first winter,
			 * fulfilled the condition specified in
			 * <code>checkBreedingCondition</code> and there is breeding season
			 * (defined in
			 * <code>InSeason</code>). For males without territories who have
			 * reached an appropriate bodyweight it is changed to
			 * <code>FindTerritory</code>. For females without nest and partner
			 * who have reached an appropriate bodyweight it is changed to
			 * <code>FindPartner</code>. </p>
			 *
			 * @return next
			 * <code>SeasonalActivity</code> or the same, if it does not require
			 * any change.
			 *
			 * @since 2012-06-27
			 */
			@Override
			public SeasonalActivity nextSeasonalActivity() {
				if (layingDate.get(Calendar.YEAR) < World.getDate().get(
						Calendar.YEAR)
						&& checkBreedingCondition() && inSeason()) {
					HashMap<String, Object> para = new HashMap<String, Object>();
					if (mySex == Sex.MALE) {
						para.put("activityName", "FindTerritory");
					} else {
						para.put("activityName", "FindPartner");
					}
					resetPesticides();
					return createSeasonalActivity(para);
				} else {
					return this;
				}
			}
		}

		/**
		 * Seasonal activity used for males withi territory during breeding
		 * season.
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @version 1.0
		 */
		public class WaitingMale extends SeasonalActivity {

			/**
			 * Constructor for WaitingMale does not require any parameters in
			 * the HashMap.
			 *
			 * @param parameters that can be used for parameterization if
			 * necessary.
			 */
			public WaitingMale(HashMap<String, Object> parameters) {
				super(parameters);
			}

			/**
			 * Specifies the next SeasonalActivity.
			 *
			 * <p> The Seasonal Activity is changed from WaitingMale for
			 * FindTerritory if there is breeding season (defined in InSeason),
			 * the male has no territory and an appropriate bodyweight. The
			 * Seasonal Activity is changed from WaitingMale for OffSeason if it
			 * is outside the breeding season and the male has no nest. </p>
			 *
			 * @return next SeasonalActivity or the same, if it does not require
			 * any change.
			 */
			@Override
			public SeasonalActivity nextSeasonalActivity() {
				HashMap<String, Object> para = new HashMap<String, Object>();
				if (inSeason() && !hasTerritory() && checkBreedingCondition()) {
					para.put("activityName", "FindTerritory");
				} else if (!inSeason() && getNest() == null) {
					para.put("activityName", "OffSeason");
				} else {
					return this;
				}
				return createSeasonalActivity(para);
			}
		}

		/**
		 * Seasonal activity used for males searching for a territory.
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @since 2012-06-28
		 */
		public class FindTerritory extends SeasonalActivity {

			/**
			 * Constructor for
			 * <code>FindTerritory<code>.
			 *
			 * @param parameters (not required)
			 * @since 2012-06-28
			 */
			public FindTerritory(HashMap<String, Object> parameters) {
				super(parameters);
			}

			/**
			 * Specifies the next
			 * <code>SeasonalActivity</code>.
			 *
			 * <p> The Seasonal Activity is changed from
			 * <code>FindTerritory</code> to
			 * <code>WaitingMale</code> if the male has a territory. The
			 * Seasonal Activity is changed from
			 * <code>FindTerritory</code> to
			 * <code>OffSeason</code> if it is outside the breeding season. </p>
			 *
			 * @return next
			 * <code>SeasonalActivity</code> or the same, if it does not require
			 * any change.
			 * @since 2012-06-28
			 */
			public SeasonalActivity nextSeasonalActivity() {
				HashMap<String, Object> para = new HashMap<String, Object>();
				if (inSeason()) {
					if (hasTerritory()) {
						para.put("activityName", "WaitingMale");
					} else {
						return this;
					}
				} else {
					para.put("activityName", "OffSeason");
				}
				return createSeasonalActivity(para);
			}
		}

		/**
		 * Seasonal activity used for females searching for a partner.
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @since 2012-06-28
		 */
		public class FindPartner extends SeasonalActivity {

			/**
			 * Constructor for
			 * <code>FindPartner</code>.
			 *
			 * @param parameters (not required)
			 */
			public FindPartner(HashMap<String, Object> parameters) {
				super(parameters);
			}

			/**
			 * Specifies the next
			 * <code>SeasonalActivity</code>.
			 *
			 * <p> The Seasonal Activity is changed from
			 * <code>FindPartner</code> to
			 * <code>Reproduce</code> if the female has a partner. The Seasonal
			 * Activity is changed from
			 * <code>FindPartner</code> to
			 * <code>OffSeason</code> if it is outside the breeding season. </p>
			 *
			 * @return next
			 * <code>SeasonalActivity</code> or the same, if it does not require
			 * any change. @parameters Reproduce_readyForReproduction
			 * @since 2012-06-28
			 */
			@Override
			public SeasonalActivity nextSeasonalActivity() {
				HashMap<String, Object> para = new HashMap<String, Object>();
				if (inSeason()) {
					if (getPartner() != null) {
						para.put("activityName", "Reproduce");
					} else {
						return this;
					}
				} else {
					para.put("activityName", "OffSeason");
				}
				return createSeasonalActivity(para);
			}

			/**
			 * Sends a message to a chosen partner to check if he accepts.
			 *
			 * @param partner id of the chosen (male) partner. 
			 * @verification TEST 1 Warning if the chosen partner is not a male.
			 * @since 2012-06-28
			 */
			public void informPartner(long partner) {
				// *****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					// ***TEST 1
					if (((Animal) World.Instance().getAgent(partner)).getSex() != Sex.MALE) {
						try {
							throw new SkyPigException(
									"Homosexual relationship detected.");
						} catch (SkyPigException ex) {
							World.log(Level.WARNING,
									Config.getLogInfo(), ex);
						}
					}
				}
				// *****VERIFICATION MODE END*****//
				Message m = new Message(getId(), partner, "partner");
				sendMessage(m);
			}
		}

		/**
		 * Seasonal activity used for females that are preparing for
		 * reproduction and laiyng eggs.
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @since 2012-06-28
		 */
		public abstract class Reproduce extends SeasonalActivity {

			/**
			 * False until all the eggs in the current brood have been laid.
			 *
			 * @since 2012-06-28
			 */
			public boolean reproductionFinished;
			/**
			 * Weights of the intended eggs
			 *
			 * @since 2012-06-28
			 */
			public LinkedList<Double> eggsWeights;
			/**
			 * Number of eggs already prepared
			 *
			 * @since 2012-06-28
			 */
			public int numberOfEggsPrepared;

			/**
			 * Constructor for
			 * <code>Reproduce</code>.
			 *
			 * <p> Values of fields are set as follows:
			 * <code>numberOfEggsPrepared</code>: 0;
			 * <code>reproductionFinished</code>: false,
			 * <code>eggWeights</code> =
			 * <code>setEggsWeights()</code>. </p>
			 *
			 * @param parameters (not required)
			 * @since 2012-06-28
			 */
			public Reproduce(HashMap<String, Object> parameters) {
				super(parameters);
				eggsWeights = new LinkedList<Double>();
				reproductionFinished = false;
				numberOfEggsPrepared = 0;
				setEggsWeights();
			}

			/**
			 * Generates the eggs and stores the length of the last renesting
			 * interval.
			 *
			 * <p> If the amount of energy in
			 * <code>reproductionBufferSum</code> allows, prepares one egg. If
			 * it is the last egg (
			 * <code>numberOfEggsPrepared</code> equals
			 * <code>eggsWeights.size()</code>), finishes the reproduction </p>
			 * If the reproduction is finished, stores the remaining energy from
			 * <code>reproducitonBufferSum</code> to
			 * <code>energyReserves</code> and informs the partner via the
			 * Message titled "nest", sending him the parameters for nest
			 * preparation generated by
			 * <code>prepareNest</code>. Finally, if it finishes reproduction (
			 * <code>lastBrood</code> != null) and it is not the first
			 * reproduction in this season, it stores the length of the last
			 * renesting interval in
			 * <code>lastRenestingInterval</code>.
			 *
			 * @parameters Reproduce_eggCost @verification TEST 1 Warning if
			 * the recorded renesting interval is longer than breeding season.
			 *
			 * @since 2012-09-05
			 */
			@Override
			public void conduct() {
				double currentEggEnergyAndCostSum = eggsWeights.get(numberOfEggsPrepared)
						* (1.0 + ((Double) Config.getParameter("Reproduce_eggCost")));
				if (currentEggEnergyAndCostSum < reproductionBufferSum) {
					reproductionBufferSum -= currentEggEnergyAndCostSum;
					if (++numberOfEggsPrepared >= eggsWeights.size()) {
						reproductionFinished = true;
					}
				}
				if (reproductionFinished) {
					energyReserves += reproductionBufferSum;
					reproductionBufferSum = 0;
					sendMessage(new Message(getId(), partnerId, "nest",
							prepareNest()));
					if (lastBroodEnd != null) {
						lastRenestingInterval = World.calculateDateDifference(lastBroodEnd);
						// *****VERIFICATION MODE START*****//
						if (lastRenestingInterval > (Integer) Config.getParameter("SeasonalActivity_endOfSeason")
								- (Integer) Config.getParameter("SeasonalActivity_startOfSeason")) {
							// ***TEST 1
							try {
								throw new SkyPigException(
										"Renesting interval longer than breeding season.");
							} catch (SkyPigException ex) {
								World.log(Level.WARNING, Config.getLogInfo(), ex);
							}
						}
						// *****VERIFICATION MODE END*****//
						lastBroodEnd = null;
					}
					nestingAttempts++;
				}
			}

			/**
			 * If
			 * <code>reproductionBufferSum</code> is not sufficient for
			 * generating all the eggs specified in
			 * <code>eggsWeights</code>, the
			 * <code>reproducitonBuffer</code> and
			 * <code>eproductionBufferSize</code> for the current day are
			 * updated.
			 *
			 * @parameters Reproduce_eggCost, Reproduce_reproductionBuffer
			 *
			 * @since 2012-06-29
			 */
			@Override
			public void calculateReproductionBuffer() {
				double energyRequiredForEggs = 0;
				for (double d : eggsWeights) {
					energyRequiredForEggs += d * (1.0 + ((Double) Config.getParameter("Reproduce_eggCost")));
				}
				if (reproductionBufferSum < energyRequiredForEggs) {
					reproductionBuffer = (Double) Config.getParameter("Reproduce_reproductionBuffer");
					reproductionBufferSize = reproductionBuffer;
				}
			}

			/**
			 * Sets number and weights of eggs to be produced in the current
			 * reproduction attempt.
			 *
			 * <p> Gets number of eggs from
			 * <code>getNumberOfEggsa</code> and sets their weights based on the
			 * normal distribution of egg weights specified by
			 * Animal_meanBodyweight and Animal_sdBodyweight Adds the weights to
			 * <code>eggsWeights</code>. </p>
			 *
			 * @parameters Animal_meanBodyweight, Animal_sdBodyweight,
			 * Animal_massToBasalEnergy
			 *
			 * @since 2012-06-29
			 */
			public void setEggsWeights() {
				int numberOfEggs = getNumberOfEggs();
				for (int i = 0; i < numberOfEggs; i++) {
					double predictedBasalMass = (Double) Config.getParameter("Egg_meanBodyweight")
							* (Double) Config.getParameter("Animal_massToBasalEnergy");
					eggsWeights.add(predictedBasalMass);
				}
			}

			/**
			 * Should be specified in species_specific models.
			 *
			 * @return number of eggs to be produced i the current reproduction
			 * attempt
			 * @since 2012-06-29
			 */
			public abstract int getNumberOfEggs();

			/**
			 * Prepares list of parameters for the new nest.
			 *
			 * @return parameters ready for passing to a Nest constructr.
			 */
			public HashMap<String, Object> prepareNest() {
				HashMap<String, Object> para = new HashMap();
				para.put("stateName", "Eggs");
				para.put("femaleOwnerId", getId());
				para.put("eggsWeights", eggsWeights);
				for (String eff : effectsMagnitude.keySet()) {
					para.put(eff, effectsMagnitude.get(eff));
				}
				return para;
			}

			/**
			 * Specifies the next
			 * <code>SeasonalActivity</code>.
			 *
			 * <p> The Seasonal Activity is changed from
			 * <code>Reproduce</code> to
			 * <code>WaitingFemale</code> if the female has a nest. The Seasonal
			 * Activity is changed from
			 * <code>Reproduce</code> to
			 * <code>OffSeason</code> if it is outside the breeding season and
			 * the female has no nest. If the female has no partner, the
			 * activity is changed to
			 * <code>FindPartner</code>. </p>
			 *
			 * @return next
			 * <code>SeasonalActivity</code> or the same, if it does not require
			 * any change.
			 * @since 2012-06-29
			 */
			@Override
			public SeasonalActivity nextSeasonalActivity() {
				HashMap<String, Object> para = new HashMap<String, Object>();
				if (getNest() != null) {
					para.put("activityName", "WaitingFemale");
					return createSeasonalActivity(para);
				} else if (!inSeason()) {
					para.put("activityName", "OffSeason");
					return createSeasonalActivity(para);
				} else if (partnerId == null) {
					para.put("activityName", "FindPartner");
					return createSeasonalActivity(para);
				} else {
					return this;
				}
			}
		}

		/**
		 * Seasonal activity used for females in breeding season when they are
		 * not reproducing and searching for partners.
		 *
		 * @author Katarzyna Kułakowska
		 * @author Tomasz Kułakowski
		 * @since 2012-06-29
		 */
		public class WaitingFemale extends SeasonalActivity {

			/**
			 * Constructor for
			 * <code>WaitingFemale</code>.
			 *
			 * @param parameters (not required)
			 */
			public WaitingFemale(HashMap<String, Object> parameters) {
				super(parameters);
			}

			/**
			 * Specifies the next SeasonalActivity.
			 *
			 * <p> The Seasonal Activity is changed from
			 * <code>WaitingFemale</code> for
			 * <code>Reproduce</code> if there is breeding season (defined in
			 * <code>InSeason</code>), the female has no nest but has a partner.
			 * In this case
			 * <code>lastBroodEnd</code> is set for current date to start
			 * counting renesting interval. The Seasonal Activity is changed
			 * from
			 * <code>WaitingFemale</code> to
			 * <code>FindPartner</code> if there is breeding season, the female
			 * has no nest and no partner and conditions specified in
			 * <code>checkBreedingCondition</code>. The Seasonal Activity is
			 * changed from
			 * <code>WaitingFemale</code> to
			 * <code>OffSeason</code> if it is outside the breeding season and
			 * the female has no nest. </p>
			 *
			 * @return next
			 * <code>SeasonalActivity</code> or the same, if it does not require
			 * any change.
			 * @since 2012-09-05
			 */
			@Override
			public SeasonalActivity nextSeasonalActivity() {
				HashMap<String, Object> para = new HashMap<String, Object>();
				if (getNest() == null) {
					if (inSeason()) {
						if (getPartner() != null) {
							para.put("activityName", "Reproduce");
							lastBroodEnd = World.getDate();
						} else if (checkBreedingCondition()) {
							para.put("activityName", "FindPartner");
							//lastBroodEnd = World.getDate();
						} else {
							return this;
						}
					} else {
						para.put("activityName", "OffSeason");
					}
					return createSeasonalActivity(para);
				}
				return this;
			}

			public void conduct() {
			}
		}
	}

	/*
	 * Fields:
	 */
	/**
	 * Contains two sex categories.
	 */
	public enum Sex {

		/**
		 * Sex category for females.
		 */
		FEMALE,
		/**
		 * Sex category for males.
		 */
		MALE
	}
	/**
	 * Date of laing the egg this animal was hatched from.
	 */
	private final Calendar layingDate;
	/**
	 * Sex of this Animal.
	 */
	private Sex mySex;
	/**
	 * Partner id or null, if this Animal has no partner.
	 */
	private Long partnerId;
	/**
	 * Nest id or null, if this Animal has no nest.
	 */
	private Long nestId;
	/**
	 * Food requirements for the current day. Changed each time a new resource
	 * is found.
	 */
	private double foodRequirements;
	/**
	 * Food requirements for the current day, not changed during this day, reset
	 * to 0 daily.
	 */
	private double animalEnergyRequirements;
	/**
	 * Energy from food digested on the current day. Reset daily.
	 */
	private double energyObtained;
	/**
	 * Energy reserves are gathered by animals that are above minimal weight.
	 * <p> They are stored in energy units (kJ). Energy reserves are stored with
	 * some costcan be used with no further cost. </p>
	 */
	private double energyReserves;
	/**
	 * Represents the energy contained in the body without energy reserves. <p>
	 * For convenience, it is stored in energy units (kJ). This energy can be
	 * used by animal for maintenance if there is no other source of energy.
	 * </p>
	 */
	private double basalEnergy;
	private double maxReserves;
	private double maxBasalEnergy;
	/**
	 * Maximum pace of storing reserves in kJ/day;
	 *
	 * @since 2012-08-21
	 */
	private double maxReservesStoreSpeed;
	/**
	 * Resources that have been collected today. Reset daily.
	 */
	private HashSet<Resource> todaysResources;
	/**
	 * Pesticides that have been collected with the Resources and their amounts
	 * in mg.
	 *
	 * @since 2012-10-22
	 */
	private HashMap<String, Double> pesticides;
	private HashMap<String, Double> maxPesticides;
	private HashMap<String, Integer> exposureTime;
	private HashMap<String, Double> cumulativePesticides;
	private HashMap<String, LinkedList<Double>> pesticidesForEffects;
	private Calendar hatchingDate;
	private Calendar nestLeavingDate;
	private Double desiredEnergy = null;
	/**
	 * Effects of pesticides possible in the current run
	 *
	 * <p> The boolean value determines if the effect is taking place on the
	 * current day. </p>
	 *
	 * @since 2012-09-08
	 */
	private HashMap<String, Boolean> effects;
	/**
	 * Effects of pesticide tested in the current run.
	 *
	 * <p>The effects are represented without the mechanism that is responsible
	 * for them, eg. dose response, cumulative amount response, time response of
	 * threshold response. The Double represents the proportion of maximum
	 * bird's performance without pesticide taking place this day. </p>
	 *
	 * @since 2012-10-28
	 */
	private HashMap<String, Double> effectsMagnitude;
	private int successfulBroods;
	private int nestingAttempts;
	private int allBroods;

	/*
	 * Methods:
	 */
	/**
	 * Constructor for an Animal. <p> It is used to construct two types of
	 * animals: adults at initialization of the model, and eggs when they are
	 * laid during the model execution. </p> <p> Required parameters: </p> <p>
	 * boolean at_initialization - true if the constructed Animal is an adult at
	 * initialization, false it if it is a laid egg; for eggs: Calendar
	 * layingDate, optionally: basalEnergy; for adults: int age - age in days;
	 * double bodyweight - bodyweight in gramms; optionally: string sex (MALE
	 * ofr FEMALE); </p>
	 *
	 * @param parameters
	 * <code>HashMap<String, Object></code> mapping the name of the field with
	 * its value.
	 */
	public Animal(HashMap<String, Object> parameters) {
		super();
		layingDate = World.getDate();
		partnerId = null;
		nestId = null;
		foodRequirements = 0;
		energyObtained = 0;
		todaysResources = new HashSet<Resource>();
		pesticides = new HashMap<String, Double>();
		maxPesticides = new HashMap<String, Double>();
		cumulativePesticides = new HashMap<String, Double>();
		exposureTime = new HashMap<String, Integer>();
		effects = new HashMap<String, Boolean>();
		effectsMagnitude = new HashMap<String, Double>();
		pesticidesForEffects = new HashMap<String, LinkedList<Double>>();
		successfulBroods = 0;
		nestingAttempts = 0;
		allBroods = 0;
		for (String pesticide : (ArrayList<String>) Config.getParameter("Pesticides")) {
			maxPesticides.put(pesticide, 0.0);
			cumulativePesticides.put(pesticide, 0.0);
			exposureTime.put(pesticide, 0);
			pesticidesForEffects.put(pesticide, new LinkedList<Double>());
			for (int i = 1; i <= (Integer) Config.getParameter("Animal_pesticideEffectLength"); i++) {
				if ((Integer) Config.getParameter("Animal_pesticideEffectLength") != null) {
					pesticidesForEffects.get(pesticide).add(0.0);
				}
			}
		}
		for (String effect : (ArrayList<String>) Config.getParameter("Animal_pesticidesEffects")) {
			if (effect != null) {
				effects.put(effect, Boolean.FALSE);
				effectsMagnitude.put(effect.split("_")[0], 0.0);
			}
		}
		if (parameters.get("sex") != null) {
			setSex((String) parameters.get("sex"));
		} else {
			setSex();
		}
		double bodyweight;
		if (parameters.get("bodyweight") != null) {
			bodyweight = (Double) parameters.get("bodyweight");
		} else {
			if (mySex.equals(Sex.FEMALE)) {
				bodyweight = (Double) Config.getParameter("Animal_meanFemaleMaxBodyweight")
						+ (Double) Config.getParameter("Animal_sdFemaleMaxBodyweight")
						* Config.getGenerator().nextGaussian();
			} else {
				bodyweight = (Double) Config.getParameter("Animal_meanMaleMaxBodyweight")
						+ (Double) Config.getParameter("Animal_sdMaleMaxBodyweight")
						* Config.getGenerator().nextGaussian();
			}
		}
		maxReserves = (Double) Config.getParameter("Animal_proportionOfReserves")
				* bodyweight
				* (Double) Config.getParameter("Animal_massToEnergyReserves");
		maxBasalEnergy = (1 - (Double) Config.getParameter("Animal_proportionOfReserves"))
				* bodyweight
				* (Double) Config.getParameter("Animal_massToBasalEnergy");
		maxReservesStoreSpeed = (maxReserves - maxBasalEnergy)
				/ ((Integer) Config.getParameter("Animal_timeOfEnergyStorage")).doubleValue();
		HashMap<String, Object> para = new HashMap<String, Object>();
		if (!(Boolean) parameters.get("at_initialization")) {
//			setSex();
			energyReserves = 0;
			if (parameters.get("basalEnergy") != null) {
				basalEnergy = (Double) parameters.get("basalEnergy");
			} else {
				basalEnergy = (Double) Config.getParameter("Egg_meanBodyweight")
						* (Double) Config.getParameter("Animal_massToBasalEnergy");
			}
			para.put("stateName", "Egg");
		} else {
			if (parameters.get("age") != null) {
			layingDate.add(Calendar.DAY_OF_MONTH,
					-(Integer) parameters.get("age"));
			} else {
				int age = Config.getGenerator().nextInt((Integer) Config.getParameter("Animal_maxInitialisationAge")
						- (Integer) Config.getParameter("Animal_minInitialisationAge")) 
						+ (Integer) Config.getParameter("Animal_minInitialisationAge");
				layingDate.add(Calendar.DAY_OF_MONTH, -age);
			}
			hatchingDate = (Calendar) layingDate.clone();
			nestLeavingDate = (Calendar) layingDate.clone();
			energyReserves = maxReserves;
			basalEnergy = maxBasalEnergy;
			para.put("stateName", "Adult");
		}
		super.initialiseState(createAgentState(para));
	}

	/**
	 * Provides access to age of this Animal.
	 *
	 * @return age in days from laing the egg the Animal was hatched from
	 */
	public int getDaysSinceLaying() {
		return World.calculateDateDifference(layingDate);
	}

	/**
	 * 
	 * @return
	 */
	public Calendar getDateOfLaying() {
		return (Calendar) layingDate.clone();
	}

	/**
	 * Provides access to age of this Animal.
	 *
	 * @return age in days from hatching of the Animal
	 */
	public int getDaysSinceHatching() {
		if (hatchingDate != null) {
			return World.calculateDateDifference(hatchingDate);
		} else {
			return -1;
		}
	}

	/**
	 * Provides access to age of this Animal.
	 *
	 * @return age in days from leaving the nest by the Animal
	 */
	public int getDaysSinceLeavingNest() {
		if (nestLeavingDate != null) {
			return World.calculateDateDifference(nestLeavingDate);
		} else {
			return -1;
		}
	}

	/**
	 *
	 */
	public void setNestLeavingDate() {
		if (nestLeavingDate == null) {
			nestLeavingDate = World.getDate();
		}
	}

	/**
	 * Returns
	 * <code>mySex</code> of this
	 * <code>Animal</code>.
	 *
	 * @return an Enum<code>mySex</code> sex of this
	 * <code>Animal</code>
	 */
	public Sex getSex() {
		return mySex;
	}

	/**
	 * 
	 * @return
	 */
	public double getBasalEnergy() {
		return basalEnergy;
	}

	/**
	 * 
	 * @return
	 */
	public double getMaxBasalEnergy() {
		return maxBasalEnergy;
	}

	/**
	 * 
	 * @return
	 */
	public HashMap<String, Double> getMaxPesticides() {
		HashMap<String, Double> mp = new HashMap<String, Double>();
		for (Entry<String, Double> e : maxPesticides.entrySet()) {
			mp.put(new String((String) e.getKey()), new Double((Double) e.getValue()));
		}
		return mp;
	}

	/**
	 * 
	 * @return
	 */
	public HashMap<String, Double> getCumulativePesticides() {
		HashMap<String, Double> cp = new HashMap<String, Double>();
		for (Entry<String, Double> e : cumulativePesticides.entrySet()) {
			cp.put(new String((String) e.getKey()), new Double((Double) e.getValue()));
		}
		return cp;
	}
	
	/**
	 * 
	 * @return
	 */
	public HashMap<String, Double> getPesticides() {
		HashMap<String, Double> pest = new HashMap<String, Double>();
		for (Entry<String, Double> e : pesticides.entrySet()) {
			pest.put(new String((String) e.getKey()), new Double((Double) e.getValue()));
		}
		return pest;
	}

	/**
	 * 
	 * @return
	 */
	public HashMap<String, Integer> getExposureTime() {
		HashMap<String, Integer> et = new HashMap<String, Integer>();
		for (Entry<String, Integer> e : exposureTime.entrySet()) {
			et.put(new String((String) e.getKey()), new Integer((Integer) e.getValue()));
		}
		return et;
	}

	/**
	 * Returns
	 * <code>Id</code> of the partner of this
	 * <code>Animal</code>.
	 *
	 * @return Integer
	 * <code>Id</code> of the partner of this
	 * <code>Animal</code>
	 */
	public Long getPartner() {
		if (partnerId != null) {
			if ((World.Instance().getAgent(partnerId)) != null) {
				return partnerId;
			} else {
				return null;
			}
		} else {
			return null;
		}
	}

	/**
	 * Calculates and provides the bodyweight of this Animal.
	 *
	 * <p> The bodyweight is calculated on the basis on basal bodyweight and
	 * energy reserves recalculated from kJ of stored energy to gramms of
	 * bodyweigt. </p>
	 *
	 * @return bodyweight in gramms
	 */
	public double getBodyweight() {
		return (basalEnergy
				/ (Double) Config.getParameter("Animal_massToBasalEnergy") + energyReserves
				/ (Double) Config.getParameter("Animal_massToEnergyReserves"));
	}

	/**
	 * Provides the food requirements of this animal as currently set in the
	 * field foodrequirements, no matter if they have been updated today.
	 *
	 * @return foodRequirements
	 */
	public double getFoodRequirements() {
		return foodRequirements;
	}

	/**
	 *
	 * @return
	 */
	public double getAnimalEnergyRequirements() {
		return animalEnergyRequirements;
	}

	/**
	 * Returns
	 * <code>Id</code> of the nest of this
	 * <code>Animal</code> or
	 * <code>null</code> if there is no nest. <p> Does not check if the nest is
	 * alive. </p>
	 *
	 * @return
	 * <code>Id</code> of the nests of this
	 * <code>Animal</code>
	 */
	public Long getNest() {
		if (nestId != null) {
			if ((World.Instance().getAgent(nestId)) != null) {
				return nestId;
			} else {
				return null;
			}
		} else {
			return null;
		}
	}

	/**
	 * Provides information about this Animal territory.
	 *
	 * @return true of the Animal has its own territory, false otherwise.
	 */
	public abstract boolean hasTerritory();

	/**
	 * Provides the energy obtained as food today, not changing the value of
	 * energyObtained field.
	 *
	 * @return energy obtained as food today
	 */
	public double getEnergyObtained() {
		return energyObtained;
	}

	/**
	 * Modifies foodrequirements subtrackting the energy content of the passed
	 * resource.
	 *
	 * <p> If the energy contant of the passed resource is higher then
	 * feedRequirements, foodRequirements are set to 0. </p>
	 *
	 * @param obtainedResource
	 */
	public void modifyFoodRequirements(Resource obtainedResource) {
		if (obtainedResource != null) {
			if (obtainedResource.energyContent * obtainedResource.amount
					>= getFoodRequirements() - (Double) Config.getParameter("Config_precision")) {
				foodRequirements = 0;
			} else {
				foodRequirements -= obtainedResource.energyContent
						* obtainedResource.amount;
			}
		}
	}

	/**
	 * Sets
	 * <code>mySex</code> of this
	 * <code>Animal</code>.
	 *
	 * Sex is set randomly, with equal probability of being
	 * <code>MALE</code> or
	 * <code>FEMALE</code>.
	 */
	private void setSex() {
		if (Config.getGenerator().nextBoolean()) {
			mySex = Sex.FEMALE;
		} else {
			mySex = Sex.MALE;
		}
	}

	/**
	 * Sets
	 * <code>mySex</code> of this
	 * <code>Animal</code> according to the parameter.
	 *
	 * @param sex to be set, either MALE or FEMALE.
	 */
	private void setSex(String sex) {
		if (sex.equalsIgnoreCase("MALE")) {
			mySex = Sex.MALE;
		} else if (sex.equalsIgnoreCase("FEMALE")) {
			mySex = Sex.FEMALE;
		}
	}

	/**
	 * 
	 * @return
	 */
	public int getSuccessfulBroods() {
		return successfulBroods;
	}

	/**
	 * 
	 * @return
	 */
	public int getNestingAttempts() {
		return nestingAttempts;
	}

	/**
	 * 
	 * @return
	 */
	public int getAllBroods() {
		return allBroods;
	}

	/**
	 * 
	 */
	public void resetSuccessfulBroods() {
		successfulBroods = 0;
	}

	/**
	 * 
	 */
	public void resetNestingAttempts() {
		nestingAttempts = 0;
	}
	/**
	 * Resets the fields that are required by the model to be calculated each
	 * day.
	 *
	 * <p> Here it resets: foodRequirements = 0, energyObtained = 0,
	 * animalEnergyRequirements = 0 and clears todaysResources. </p>
	 */
	private void resetAnimal() {
		foodRequirements = 0;
		energyObtained = 0;
		todaysResources.clear();
		animalEnergyRequirements = 0;
		if (World.getDate().get(Calendar.DAY_OF_YEAR)
				== (Integer) Config.getParameter("SeasonalActivity_startOfSeason") + 1) {
			resetSuccessfulBroods();
			allBroods = 0;
		}
	}

	/**
	 * Used to create the Signature with all the relevant information from this
	 * Animal that should be stored in database.
	 *
	 * <p> Animal adds to the Signature following values: sex, date_of_birth,
	 * bodyweight, basalEnergy (recalculated to weight units), energyReserves
	 * (recalculated to weight units), partner id, nest id. </p>
	 *
	 * @return the Animal's signature
	 */
	@Override
	public Signature getSignature() {
		// add a table of intigers for storing ids of resources eaten this day
		Signature newSignature = super.getSignature();
		newSignature.addPrimitiveField("sex", "text", "'" + mySex.toString()
				+ "'");
		newSignature.addPrimitiveField("bodyweight", "double precision",
				((Double) getBodyweight()).toString());
		newSignature.addPrimitiveField("basalEnergy", "double precision",
				((Double) (basalEnergy / (Double) Config.getParameter("Animal_massToBasalEnergy"))).toString());
		newSignature.addPrimitiveField("energyReserves", "double precision",
				((Double) (energyReserves / (Double) Config.getParameter("Animal_massToEnergyReserves"))).toString());
		newSignature.addPrimitiveField("foodRequirements", "double precision",
				(Double.toString(foodRequirements)));
		newSignature.addPrimitiveField("energyObtained", "double precision",
				(Double.toString(energyObtained)));
		newSignature.addPrimitiveField("animalEnergyRequirements",
				"double precision", Double.toString(animalEnergyRequirements));
		if (partnerId != null) {
			newSignature.addPrimitiveField("partner", "bigint",
					Long.toString(partnerId));
		} else {
			newSignature.addPrimitiveField("partner", "bigint", "null");
		}
		if (nestId != null) {
			newSignature.addPrimitiveField("nest", "bigint",
					Long.toString(nestId));
		} else {
			newSignature.addPrimitiveField("nest", "bigint", "null");
		}
		newSignature.addPrimitiveField("agesincelaying", "integer",
				Integer.toString(getDaysSinceLaying()));
		newSignature.addPrimitiveField("agesincehatching", "integer",
				Integer.toString(getDaysSinceHatching()));
		newSignature.addPrimitiveField("agesinceleavingnest", "integer",
				Integer.toString(getDaysSinceLeavingNest()));
		for (String pesticide : pesticides.keySet()) {
			Signature pestSignature = new Signature(this);
			pestSignature.setTableName((String) Config.getParameter("Animal_pesticideTableName"));
			pestSignature.addPrimitiveField("id", "bigint",
					Long.toString(getId()));
			pestSignature.addPrimitiveField("pesticide", "text", "'"
					+ pesticide + "'");
			pestSignature.addPrimitiveField("amount", "double precision",
					Double.toString(pesticides.get(pesticide)));
			pestSignature.addPrimaryKey("id");
			pestSignature.addPrimaryKey("pesticide");
			newSignature.addInternalObject(pestSignature);
		}
		return newSignature;
	}
}
