//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
package model.core;

import java.util.*;
import java.util.logging.Level;

/**
 *
 * @author kasia
 */
public abstract class LandscapeUnit extends Agent {

	/**
	 *
	 */
	public abstract class Vegetation extends AgentState {
		/*
		 * Fields:
		 */

		/**
		 * Name of the current Vegetation type
		 */
		public String name;
		/**
		 * The period in agricultural regime: after sowing or after harvest
		 */
		public String agriculturalTime;
		/**
		 * The day in year in which the vegetation changes to fallow.
		 *
		 * <p> If the vegetation is not changing (eg permanent fallow, pasture
		 * or meadow), The next year vegetation type is read from crop rotation.
		 * </p>
		 */
		public Integer ploughingDay;
		/**
		 *
		 */
		public Integer sowingDay;
		/**
		 *
		 */
		public Integer harvestingDay;
		/**
		 *
		 */
		public Integer nextVegetationSowingDay;
		/**
		 * True if the vegetation type is listed in the crop rotation, false for
		 * unlisted fallow.
		 */
		public boolean inCropRotation;
		/**
		 * For the given vegetation from crop rotation true before the first
		 * June of this crop, false later
		 */
		public boolean seasonIncomplete;

		/*
		 * Methods:
		 */
		/**
		 *
		 * @param parameters
		 */
		public Vegetation(HashMap<String, Object> parameters) {
			name = (String) parameters.get("stateName");
			resources.clear();
			seasonIncomplete = true;
//			if (parameters.containsKey("inCropRotation")) {
//				inCropRotation = (Boolean) parameters.get("inCropRotation");
			if (name.equalsIgnoreCase("Fallow")) {
				inCropRotation = false;
			} else {
				inCropRotation = true;
			}
			initialiseResources(getListOfResources());
			sowingDay = World.getDate().get(Calendar.DAY_OF_YEAR);
			if (parameters.containsKey("nextVegetationSowingDay")) {
				nextVegetationSowingDay = (Integer) parameters.get("nextVegetationSowingDay");
			} else {
				setHarvestAndNextSowingDay();
			}
			setPloughingDay();
			sow();
		}

		private void setHarvestAndNextSowingDay() {
			Integer harvestingDayStart = (Integer) Config.getParameter(name
					+ "_harvestingDayStart");
			Integer harvestingDayEnd = (Integer) Config.getParameter(name
					+ "_harvestingDayEnd");
			Integer sowingDayStart = (Integer) Config.getParameter(myRotation.getNextVegetationName() + "_sowingDayStart");
			Integer sowingDayEnd = (Integer) Config.getParameter(myRotation.getNextVegetationName() + "_sowingDayEnd");
			if (harvestingDayEnd == null || harvestingDayStart == null) {
				harvestingDay = null;
			}
			if (sowingDayEnd == null || sowingDayStart == null) {
				nextVegetationSowingDay = null;
			}
			if (harvestingDayEnd != null && harvestingDayStart != null && sowingDayStart == null && sowingDayEnd == null) {
				harvestingDay = chooseRandomInt(harvestingDayStart, harvestingDayEnd);
			}
			if (sowingDayEnd != null && sowingDayStart != null && harvestingDayStart == null && harvestingDayEnd == null) {
				nextVegetationSowingDay = chooseRandomInt(sowingDayStart, sowingDayEnd);
			}
			if (sowingDayEnd != null && sowingDayStart != null && harvestingDayStart != null && harvestingDayEnd != null) {
				if (sowingDayEnd - harvestingDayStart < 2 && sowingDayEnd > (Integer) Config.getParameter("Vegetation_halfYear")) {
					World.log(Level.SEVERE, "Incorrect input for agricultural activities: "
							+ "mismatch between {0}_harvestingDayStart and {1}_sowingDayEnd.",
							new Object[]{name, myRotation.getNextVegetationName()});
					throw new Error("Incorrect input for agricultural activities: "
							+ "mismatch between " + name + "_harvestingDayStart and "
							+ myRotation.getNextVegetationName() + "_sowingDayEnd.");
				}
				if (harvestingDayEnd < sowingDayStart || sowingDayEnd < harvestingDayStart) {
					harvestingDay = chooseRandomInt(harvestingDayStart, harvestingDayEnd);
					nextVegetationSowingDay = chooseRandomInt(sowingDayStart, sowingDayEnd);
				} else {
					TreeSet<Integer> dates = new TreeSet<Integer>();
					dates.add(harvestingDayStart);
					dates.add(harvestingDayEnd);
					dates.add(sowingDayStart);
					dates.add(sowingDayEnd);
					dates.pollFirst();
					dates.pollLast();
					int middle = chooseRandomInt(dates.pollFirst() + 1, dates.pollLast());
					harvestingDay = chooseRandomInt(harvestingDayStart, middle);
					nextVegetationSowingDay = chooseRandomInt(middle + 1, sowingDayEnd + 1);
				}
			}
		}

		private int chooseRandomInt(int earlier, int later) {
			int delay = Config.getGenerator().nextInt(calculateTimeDifference(later, earlier));
			if (earlier + delay > 365) {
				return 365 - (earlier + delay);
			} else {
				return earlier + delay;
			}
		}

		/**
		 * Should return the list of the names of resources in the current
		 * vegetation
		 *
		 * @return list of the names of resources in the current vegetation
		 */
		public abstract HashMap<String, Double> getListOfResources();

		/**
		 * Creates new empty resources for the vegetation.
		 *
		 * @param listOfResources to be created
		 */
		public void initialiseResources(HashMap<String, Double> listOfResources) {
			if (resources.isEmpty()) {
				if (listOfResources != null) {
					for (String r : listOfResources.keySet()) {
						HashMap<String, Object> para = new HashMap<String, Object>();
						para.put("name", r);
						para.put("amount", listOfResources.get(r) * size);
						para.put("size", size);
						resources.add(new Resource(para));
					}
				}
			}
		}

		/**
		 * Sets the ending date for this vegetaiton in crop rotation.
		 *
		 * <p> For cases when the vegetation is ploughed and after a period of
		 * fallow the next crop is sown, the ploughing day is randomly chosen
		 * between harves and sowing of the next crop. For crops followed by a
		 * pasture or fallow lasting at least one season, the ploughing day is
		 * chosen randomly between harvest and the parameter
		 * Vegetation_latestVegetationEndingDay. For pasture, meadow,fallow and
		 * other possible vegetation types wihout a specified date of harvest,
		 * the ploughing day is chosen randomly in the period given by a
		 * parameter Vegetation_MaxPeriodOfFallow before sowig of the next
		 * vegetation. For cases when the vegetation is not changing and not
		 * turned into a fallow (permanent pasture, meadow or fallow) the
		 * ploughing day means the day in which the next season of this
		 * vegetation begins given by the parameter
		 * Vegetation_latestVegetationEndingDay. Should be overwritten for
		 * fallows changing directly into a new sown crop. </p>
		 *
		 * @parameters Vegetation_latestVegetationEndingDay,
		 * Vegetation_MaxPeriodOfFallow, "vegetationName"_harvestingDay
		 */
		public void setPloughingDay() {
			if (nextVegetationSowingDay == null && harvestingDay == null) {
				ploughingDay = (Integer) Config.getParameter("Vegetation_latestVegetationEndingDay");
			} else {
				if (nextVegetationSowingDay == null) {
					nextVegetationSowingDay = (Integer) Config.getParameter("Vegetation_latestVegetationEndingDay");
				}
				int dayBeforeSowing = 0;
				if (harvestingDay == null) {
					if (nextVegetationSowingDay > (Integer) Config.getParameter("Vegetation_MaxPeriodOfFallow")) {
						dayBeforeSowing = nextVegetationSowingDay
								- (Integer) Config.getParameter("Vegetation_MaxPeriodOfFallow");
					} else {
						dayBeforeSowing = 365 - ((Integer) Config.getParameter("Vegetation_MaxPeriodOfFallow") - nextVegetationSowingDay);
					}
					harvestingDay = dayBeforeSowing;
				}
				int timeBetweenHarvestAndSowing = calculateTimeDifference(
						nextVegetationSowingDay, harvestingDay);
				int timeBetweenHarvestAndPloughing = Config.getGenerator().nextInt(timeBetweenHarvestAndSowing - 1) + 1;
				if (harvestingDay + timeBetweenHarvestAndPloughing <= 365) {
					ploughingDay = harvestingDay
							+ timeBetweenHarvestAndPloughing;
				} else {
					ploughingDay = harvestingDay
							+ timeBetweenHarvestAndPloughing - 365;
				}
				if (nextVegetationSowingDay - ploughingDay == 1) {
					int ala = 0;
				}
			}
		}

		/**
		 *
		 */
		public void setIntermediateFallowPloughingDay() {
			if (!inCropRotation) {
				if (nextVegetationSowingDay == null) {
					if ((Integer) Config.getParameter("Vegetation_latestVegetationEndingDay") >= 365) {
						ploughingDay = 1;
					} else {
						ploughingDay = 1 + (Integer) Config.getParameter("Vegetation_latestVegetationEndingDay");
					}
				} else {
					ploughingDay = nextVegetationSowingDay;
				}
			}
		}

		/**
		 *
		 * @param resource
		 * @param newAmount
		 */
		public void changeResourceAmount(String resource, double newAmount, boolean pesticide) {
			if (newAmount < 0.0) {
				newAmount = 0.0;
			}
			for (Resource r : resources) {
				if (r.getName().equalsIgnoreCase(resource)) {
					if (!pesticide) {
						r.modifyPesticides(0.0);
					} else if (r.amount > newAmount) {
						r.modifyPesticides(newAmount / r.amount);
					}
					r.amount = newAmount;
				}
			}
		}

		/**
		 *
		 * @param resource
		 * @param newDensity
		 */
		public void changeResourceDensity(String resource, double newDensity) {
			for (Resource r : resources) {
				if (r.getName().equalsIgnoreCase(resource)) {
					r.density = newDensity;
				}
			}
		}

		/**
		 *
		 * @param resource
		 * @param energyContent
		 */
		public void changeResourceEnergyContent(String resource,
				double energyContent) {
			for (Resource r : resources) {
				if (r.getName().equalsIgnoreCase(resource)) {
					r.energyContent = energyContent;
				}
			}
		}

		/**
		 *
		 */
		@Override
		public void findFood() {
			return;
		}

		/**
		 * @param causeOfDeath @verification TEST 1 Warning if the LandscapeUnit
		 * is dead.
		 */
		@Override
		public final boolean checkDeath(String causeOfDeath) {
			if (causeOfDeath == null) {
				return false;
			} else {
				// *****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					try {
						throw new SkyPigException(
								"LandscapeUnit tired to die giving "
								+ causeOfDeath + " as causeOfDeath:"
								+ String.valueOf(getId()));
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
				// *****VERIFICATION MODE END*****//
				return false;
			}
		}

		/**
		 *
		 */
		@Override
		public void conductSeasonalActivity() {
			if (harvestingDay != null) {
				if (World.getDate().get(Calendar.DAY_OF_YEAR) == harvestingDay
						&& !seasonIncomplete) {
					harvest();
				}
			}
			if (((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")) != null) {
				if (Config.getParameter("Pesticides") != null) {
					for (String pesticide : (ArrayList<String>) Config.getParameter("Pesticides")) {
						if (((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).containsKey(pesticide)) {
							if (((String) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(
									pesticide).get("mode")).equalsIgnoreCase("spray")) {
								if ((Integer) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide).get("time") == World.getDate().get(Calendar.DAY_OF_YEAR)) {
									if (Config.checkIfPesticideApplicationYear(pesticide, name)
											&& Config.getGenerator().nextDouble()
											< ((Double) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide).get("probability"))) {
										Agent lu = World.Instance().getAgent(getId());
										if (((Boolean) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide).get("yeartime")
												&& (Integer) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide).get("time") == World.getDate().get(Calendar.DAY_OF_YEAR))
												|| (!(Boolean) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide).get("yeartime")
												&& calculateTimeDifference(World.getDate().get(Calendar.DAY_OF_YEAR),
												sowingDay) == (Integer) ((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide).get("time"))) {
											for (Resource r : resources) {
												r.addPesticide(pesticide,
														calculatePesticideAmount(pesticide, r));
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}

		

		private double calculatePesticideAmount(String pesticide, Resource r) {
			return (Double) (((HashMap<String, HashMap<String, Object>>) Config.getParameter(name + "_pesticide")).get(pesticide)).get("density")
					* (Double) Config.getParameter(pesticide + "_rud")
					* r.amount / 1000;
//		For weeds overgrown by the crop for which we know interception factor:
//			return  (Double) (((HashMap<String, HashMap<String, Object>>) 
//					Config.getParameter(name + "_pesticide")).get(pesticide)).get("density")
//					* (Double) Config.getParameter(pesticide + "_pesticideInterception")
//					* getSize();
		}

		/**
		 *
		 */
		public void sow() {
			agriculturalTime = "afterSowing";
		}

		/**
		 *
		 */
		public void harvest() {
			agriculturalTime = "afterHarvest";
		}

		@Override
		public void manageEnergy() {
			for (Resource r : resources) {
				r.decayPesticides();
			}
		}

		/**
		 *
		 * @return
		 */
		public CropRotation getMyRotation() {
			return myRotation;
		}

		/**
		 *
		 * @return
		 */
		@Override
		public AgentState changeState() {
			if (World.getDate().get(Calendar.MONTH) == 5) {
				seasonIncomplete = false;
			}
			if (World.getDate().get(Calendar.DAY_OF_YEAR) == ploughingDay
					&& seasonIncomplete == false) {
				if (nextVegetationSowingDay == null
						&& harvestingDay == null
						&& name.equalsIgnoreCase(myRotation.getNextVegetationName())) {
					seasonIncomplete = true;
					myRotation.getNextVegetation();
					setHarvestAndNextSowingDay();
					setPloughingDay();
					return this;
				} else {
					HashMap<String, Object> vegetationParameters = new HashMap<String, Object>();
					vegetationParameters.put("stateName", "Fallow");
					vegetationParameters.put("nextVegetationSowingDay", nextVegetationSowingDay);
					//vegetationParameters.put("inCropRotation", false);
					return createAgentState(vegetationParameters);
				}
			} else {
				return this;
			}
		}

		/**
		 *
		 * @return
		 */
		public AgentState getNextStateForFallow() {
			if (inCropRotation) {
				return null;
			} else if (World.getDate().get(Calendar.DAY_OF_YEAR) == ploughingDay) {
				String nextVegetation = myRotation.getNextVegetation();
				if (!nextVegetation.equalsIgnoreCase(name)) {
					HashMap<String, Object> vegetationParameters = new HashMap<String, Object>();
					vegetationParameters.put("stateName", nextVegetation);
					return createAgentState(vegetationParameters);
				} else {
					setPloughingDay();
					seasonIncomplete = true;
					inCropRotation = true;
					return this;
				}
			} else {
				return this;
			}
		}

		@Override
		public Signature getSignature() {
			Signature newSignature = super.getSignature();
			newSignature.addPrimitiveField("name", "text", "'" + name + "'");
			newSignature.addPrimitiveField("agriculturalTime", "text", "'"
					+ agriculturalTime + "'");
			newSignature.addPrimitiveField("inCropRotation", "boolean",
					Boolean.toString(inCropRotation));
			newSignature.addPrimitiveField("ploughingDay", "integer",
					Integer.toString(ploughingDay));
			if (sowingDay != null) {
				newSignature.addPrimitiveField("sowingDay", "integer",
						Integer.toString(sowingDay));
			} else {
				newSignature.addPrimitiveField("sowingDay", "integer", "null");
			}
			if (harvestingDay != null) {
				newSignature.addPrimitiveField("harvestingDay", "integer",
						Integer.toString(harvestingDay));
			} else {
				newSignature.addPrimitiveField("harvestingDay", "integer",
						"null");
			}
			if (nextVegetationSowingDay != null) {
				newSignature.addPrimitiveField("nextSowingDay", "integer",
						Integer.toString(nextVegetationSowingDay));
			} else {
				newSignature.addPrimitiveField("nextSowingDay", "integer",
						"null");
			}
			newSignature.addPrimitiveField("nextindex", "integer",
					Integer.toString(myRotation.getNextVegetationIndex()));
			return newSignature;
		}
	}

	/*
	 * Fields:
	 */
	/**
	 * The coordinates of this
	 * <code>LandscapeUnit</code> centroid.
	 */
	private ArrayList<Double> centroid;
	// private Vegetation vegetation;
	/**
	 * Size of this
	 * <code>LanscapeUnit</code> in m^2.
	 */
	private double size;
	/**
	 * Collection of the shortest distances between borders if this
	 * <code>LanscapeUnit</code> and borders of all the other
	 * <code>LanscapeUnits</code>.
	 */
	private TreeMap<Double, ArrayList<LandscapeUnit>> borderDistances;
	/**
	 * Collection of distances between the centroid of this
	 * <code>LanscapeUnit</code> and the centroids of all the other
	 * <code>LanscapeUnits</code>.
	 */
	// private TreeMap<Double, ArrayList<LandscapeUnit>> centreDistances;
	/**
	 * Collection of border lengths between this
	 * <code>LandscapeUnit</code> and all the other
	 * <code>LanscapeUnits</code>.
	 */
	private TreeMap<Double, ArrayList<LandscapeUnit>> borderLengths;
	/**
	 * Collection of all territories stored on this
	 * <code>LanscapeUnit</code>.
	 */
	// private TreeSet<SkylarkTerritory> territories;
	private CropRotation myRotation;
	/**
	 * Description of resources on this
	 * <code>Vegetation</code>.
	 */
	private HashSet<Resource> resources;

	/*
	 * Methods:
	 */
	/**
	 * Creates a new
	 * <code>LandscapeUnit</code> based on the specification. Executed at the
	 * beginning of the program.
	 *
	 * @param parameters
	 */
	public LandscapeUnit(HashMap<String, Object> parameters) {
		super();
		size = (Double) parameters.get("size");
		borderDistances = new TreeMap();
		borderLengths = new TreeMap();
		resources = new HashSet<Resource>();
		centroid = (ArrayList<Double>) parameters.get("centroid");
		myRotation = CropRotation.getCropRotation(
				(String) parameters.get("cropRotation"),
				(Integer) parameters.get("cropIndex"));
		HashMap<String, Object> para = new HashMap<String, Object>();
		if ((Integer) Config.getParameter(myRotation.getNextVegetationName()
				+ "_sowingDayStart") != null) {
			para.put("stateName", "Fallow");
			//para.put("inCropRotation", false);
		} else {
			para.put("stateName", myRotation.getNextVegetation());
		}
		Vegetation currentVeg = (Vegetation) createAgentState(para);
		super.initialiseState(currentVeg);
	}

	/**
	 * Returns a set of the names of resources available on this
	 * <code>LandscapeUnit</code>.
	 *
	 * @return
	 */
	public Set<Resource> getResources() {
		return Collections.unmodifiableSet(resources);
	}

	/**
	 *
	 * @param distance
	 * @return
	 */
	public NavigableMap<Double, ArrayList<LandscapeUnit>> getLandscapeUnits(
			double distance) {
		return borderDistances.headMap(distance, true);
	}

	/**
	 *
	 * @param x
	 * @return
	 */
	public double calculateCentralDistance(LandscapeUnit x) {
		ArrayList<Double> centroidX = x.getCentroid();
		return Math.pow(
				(Math.pow((centroid.get(0) - centroidX.get(0)), 2) + Math.pow(
				(centroid.get(1) - centroidX.get(1)), 2)), 0.5);
	}

	/**
	 *
	 * @return
	 */
	public LandscapeUnit getCloseLandscapeUnit() {
		TreeMap<Double, ArrayList<LandscapeUnit>> invertedBorderDistances = new TreeMap<Double, ArrayList<LandscapeUnit>>();
		for (Map.Entry<Double, ArrayList<LandscapeUnit>> pair : borderDistances.entrySet()) {
			invertedBorderDistances.put(
					(borderDistances.lastKey() - pair.getKey())
					* pair.getValue().size(), pair.getValue());
		}
		double sum = 0.0;
		TreeMap<Double, ArrayList<LandscapeUnit>> cumulativeBorderDistances = new TreeMap<Double, ArrayList<LandscapeUnit>>();
		for (Map.Entry<Double, ArrayList<LandscapeUnit>> pair : invertedBorderDistances.entrySet()) {
			sum += pair.getKey();
			cumulativeBorderDistances.put(sum, pair.getValue());
		}
		Double rand = Config.getGenerator().nextDouble() * sum;
		Iterator<Double> it = cumulativeBorderDistances.navigableKeySet().iterator();
		while (it.hasNext()) {
			Double currentDistance = it.next();
			if (rand <= currentDistance) {
				int n = 0;
				if (cumulativeBorderDistances.get(currentDistance).size() > 1) {
					n = Config.getGenerator().nextInt(cumulativeBorderDistances.get(
							currentDistance).size());
				}
				return cumulativeBorderDistances.get(currentDistance).get(n);
			}
		}
		return null;
	}

	/**
	 *
	 * @return
	 */
	public ArrayList<Double> getCentroid() {
		return centroid;
	}

	/**
	 *
	 * @return
	 */
	public double getSize() {
		return size;
	}

	/**
	 *
	 * @param borderDistances
	 */
	public void setBorderDistances(
			TreeMap<Double, ArrayList<LandscapeUnit>> borderDistances) {
		if (this.borderDistances.isEmpty()) {
			this.borderDistances = borderDistances;
		}
	}

	/**
	 *
	 * @param borderLengths
	 */
	public void setBorderLengths(
			TreeMap<Double, ArrayList<LandscapeUnit>> borderLengths) {
		if (this.borderLengths.isEmpty()) {
			this.borderLengths = borderLengths;
		}
	}

	/**
	 * Returns the time difference in days between two
	 * <code>Calendar.DAY_OF_YEAR</code>, not necesserily in the same year, but
	 * only if the difference is less then one year.
	 *
	 * @param later
	 * @param earlier
	 * @return difference in days
	 */
	public int calculateTimeDifference(int later, int earlier) {
		if (later >= earlier) {
			return later - earlier;
		} else {
			return (365 - earlier) + later;
		}
	}

	/**
	 * Returns value of a logistic function reflecting probability of getting a
	 * territory on this LandscapeUnit.
	 *
	 * @param x - the ratio of territories occupied to the mean number of
	 * territories for this Landscape Unit
	 * @return probability of getting a territory
	 * @since 2012-06-28
	 */
	public double calculateTerritoryGettingProbability(double x) {
		return (1 / (1 + Math.exp((Double) Config.getParameter("LandscapeUnit_territoryGettingProbabilitySlope")
				* (x - 1))));
	}

	/**
	 * Should be overwritten.
	 *
	 * @param parameters
	 * @return
	 */
	public Resource giveResource(HashMap<String, Object> parameters) {
		return null;
	}

	/**
	 *
	 * @return
	 */
	public double getEnergyDensity() {
		double en = 0;
		for (Resource r : resources) {
			double x = (r.amount * r.energyContent) / (size * 10000);
			if (x > en) {
				en = x;
			}
		}
		return en;
	}

	/**
	 * Should be overwritten.
	 *
	 * @param parameters
	 * @return
	 */
	public HashSet<Resource> giveResources(HashMap<String, Object> parameters) {
		return resources;
	}

	/**
	 *
	 * @param id
	 */
	public void addToWaitingList(long id) {
	}

	/**
	 * Should be overwritten to return partner id.
	 *
	 * @return null
	 * @since 2012-06-28
	 */
	public Long providePartnerId() {
		return null;
	}

	@Override
	public Signature getSignature() {
		Signature newSignature = super.getSignature();
		newSignature.addPrimitiveField("size", "double precision",
				Double.toString(size));
		newSignature.addPrimitiveField("rotation", "text",
				"'" + myRotation.getName() + "'");
		String resourcesIds = "'{";
		for (Resource r : resources) {
			resourcesIds += String.valueOf(r.getId()) + ",";
		}
		if (resourcesIds.equalsIgnoreCase("'{")) {
			resourcesIds += "}'";
		} else {
			resourcesIds = resourcesIds.substring(0, resourcesIds.length() - 1)
					+ "}'";
		}
		newSignature.addPrimitiveField("resourcesIds", "bigint[]", resourcesIds);
		for (Resource r : resources) {
			newSignature.addInternalObject(r.getSignature());
		}
		return newSignature;// (primitiveFields, internalObjects,
		// World.Instance().getDay(), tableName, this);
	}
}
