//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.util.logging.Level;

/**
 * Serves for the information exchange.
 *
 * <p>
 * There are four constructors for the Message, each with different parameter set.
 * For the given constructor, parameters must not be null.
 * </p>
 *
 * @author Katarzyna Kułakowska
 * @author Tomasz Kułakowski
 * @version 1.0
 */
public class Message {

    /*
     * Fields
     */
    /**
     *  The id of an Agent sending the Message.
     */
    public final Long senderId;
    /**
     * The id of an Agent receiving the Message.
     */
    public final Long recipientId;
    /**
     * Topic of the Message.
     * <p>
     * Messages are recognised by their topis. Each topic used needs to be
     * handled in the readMesage function of AgentState or derived classes.
     * </p>
     */
    public final String topic;
    /**
     * Message to which this message replies, null if this is the first
     * message in thie current information exchange.
     */
    public final Message inReplyTo;
    /**
     * Attachment may be null or may contain specific information for the recipient.
     */
    public final Object attachment;

    /**
     * Creates a new <code>Message</code> without specified attachment.
     * 
     *
     * <p>
     * This is the first message, not a reply to another message.
     * </p>
     *
     * @param senderId
     * @param recipientId
     * @param topic
     * @verification TEST 1 Warning if any of the obligatory fields in message 
	 * specification (topic, senderId, recipientId) is empty
     */
    public Message(Long senderId, Long recipientId, String topic) {
        this.senderId = senderId;
        this.recipientId = recipientId;
        this.topic = topic;
        this.inReplyTo = null;
        this.attachment = null;
//*****VERIFICATION MODE START*****//
        if (this.topic == null || this.senderId == null || this.recipientId == null) {
            //***TEST 1
            try {
                throw new SkyPigException("Wrong message specification: "
                        + "senderId: " + String.valueOf(this.senderId)
                        + ", recipientId: " + String.valueOf(this.recipientId)
                        + ", topic: " + this.topic);
            } catch (SkyPigException ex) {
                World.log(Level.WARNING, Config.getLogInfo(), ex);
            }
        }
//*****VERIFICATION MODE END*****//
    }

    /**
     *Creates a new <code>Message</code> without specified attachment.
     *
     * <p>
     * The message created is a reply to another message passed as a parameter.
     * <code>inReplyTo</code>.
     * </p>
     *
     * @param senderId
     * @param recipientId
     * @param topic
     * @param inReplyTo
     * @verification TEST 1 Warning if any of the obligatory fields in message 
	 * specification (topic, senderId, recipientId) is empty
     */
    public Message(Long senderId, Long recipientId, String topic, Message inReplyTo) {
        this.senderId = senderId;
        this.recipientId = recipientId;
        this.topic = topic;
        this.inReplyTo = inReplyTo;
        this.attachment = null;
//*****VERIFICATION MODE START*****//
        if (this.topic == null || this.senderId == null || this.recipientId == null) {
            //***TEST 1
            try {
                throw new SkyPigException("Wrong message specification: "
                        + "senderId: " + String.valueOf(this.senderId)
                        + ", recipientId: " + String.valueOf(this.recipientId)
                        + ", topic: " + this.topic);
            } catch (SkyPigException ex) {
                World.log(Level.WARNING, Config.getLogInfo(), ex);
            }
        }
//*****VERIFICATION MODE END*****//
    }

    /**
     * Creates a new <code>Message</code> with a specified attachment.
     *
     * <p>
     * This is the first message, not a reply to another message.
     * </p>
     *
     * @param senderId
     * @param recipientId
     * @param topic
     * @param attachment
     * @verification TEST 1 Warning if any of the obligatory fields in message 
	 * specification (topic, senderId, recipientId) is empty
	 * */
    public Message(Long senderId, Long recipientId, String topic, Object attachment) {
        this.senderId = senderId;
        this.recipientId = recipientId;
        this.topic = topic;
        this.inReplyTo = null;
        this.attachment = attachment;
//*****VERIFICATION MODE START*****//
        if (this.topic == null || this.senderId == null || this.recipientId == null) {
            //***TEST 1
            try {
                throw new SkyPigException("Wrong message specification: "
                        + "senderId: " + String.valueOf(this.senderId)
                        + ", recipientId: " + String.valueOf(this.recipientId)
                        + ", topic: " + this.topic);
            } catch (SkyPigException ex) {
                World.log(Level.WARNING, Config.getLogInfo(), ex);
            }
        }
//*****VERIFICATION MODE END*****//
    }

    /**
     * Creates a new <code>Message</code> with a specified attachment.
     *
     * <p>
     * The message is a reply to another message passed as the parameter
     * <code>inReplyTo</code>.
     * </p>
     *
     * @param senderId
     * @param recipientId
     * @param topic
     * @param inReplyTo
     * @param attachment
     * @verification TEST 1 Warning if any of the obligatory fields in message 
	 * specification (topic, senderId, recipientId) is empty
     */
    public Message(Long senderId, Long recipientId, String topic,
            Message inReplyTo, Object attachment) {
        this.senderId = senderId;
        this.recipientId = recipientId;
        this.topic = topic;
        this.inReplyTo = inReplyTo;
        this.attachment = attachment;
//*****VERIFICATION MODE START*****//
        if (this.topic == null || this.senderId == null || this.recipientId == null) {
            //***TEST 1
            try {
                throw new SkyPigException("Wrong message specification: "
                        + "senderId: " + String.valueOf(this.senderId)
                        + ", recipientId: " + String.valueOf(this.recipientId)
                        + ", topic: " + this.topic);
            } catch (SkyPigException ex) {
                World.log(Level.WARNING, Config.getLogInfo(), ex);
            }
        }
//*****VERIFICATION MODE END*****//
    }
}
