//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.logging.Level;

import model.core.Animal.AnimalState;

/**
 *
 * @author kasia
 */
public abstract class Nest extends Agent {

	/**
	 * Specifies the state of nests with eggs during incubation time.
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @since 2012-06-30
	 */
	public abstract class Eggs extends AgentState {

		/*
		 * Methods:
		 */
		/**
		 * Constructor for Eggs does not require any additional parameters.
		 *
		 * @param parameters
		 */
		public Eggs(HashMap<String, Object> parameters) {
			super();
			if (parameters.containsKey("maleOwnerId")) {
				addToContactList((Long) parameters.get("maleOwnerId"));
			}
			addToContactList((Long) parameters.get("femaleOwnerId"));
		}

		/**
		 *
		 *
		 * @param causeOfDeath
		 */
		@Override
		public boolean checkDeath(String causeOfDeath) {
			if (causeOfDeath == null) {
				deadOffspring.clear();
				if (checkIfPredated()) {
					return super.checkDeath("predated");
				} else if (checkIfDestroyed()) {
					return super.checkDeath("destroyed");
				} else {
					Iterator<Animal> iterator = offspring.iterator();
					while (iterator.hasNext()) {
						Animal egg = iterator.next();
						if (egg.getState(getId()).checkDeath(null)
								|| (checkIfEggsHatch()
								&& ((Animal.Egg) egg.getState(getId())).checkHatchingEffect(eggsHatchingEffect))) {
							iterator.remove();
							deadOffspring.add(egg);
						}
					}
					if (offspring.isEmpty()) {
						return super.checkDeath("noOffspring");
					} else {
						return false;
					}
				}
			} else {
				return super.checkDeath(causeOfDeath);
			}
		}

		/**
		 * Checks if nest was predated. Needs implementation in species specific
		 * classes.
		 *
		 * @return
		 */
		public abstract boolean checkIfPredated();

		/**
		 * Does nothing.
		 */
		public void manageEnergy() {
			return;
		}

		/**
		 * Sends the messages to both parents (or one, if the second is dead)
		 * requesting for incubation.
		 *
		 * @since 2012-06-30
		 */
		@Override
		public void findFood() {
			if (getMaleOwnerId() != null) {
				sendMessage(new Message(getId(), getMaleOwnerId(), "incubation"));
			}
			if (getFemaleOwnerId() != null) {
				sendMessage(new Message(getId(), getFemaleOwnerId(), "incubation"));
			}
		}

		/**
		 * Replies and conducts appropriate actions when
		 *
		 * @param m
		 * @return
		 */
		@Override
		public Message readMessage(Message m) {
			if (m.topic.equalsIgnoreCase("re:incubation")) {
				receiveIncubation((Double) m.attachment);
				return null;
			} else if (m.topic.equalsIgnoreCase("dead_agent")
					|| m.topic.equalsIgnoreCase("remove_agent")
					|| m.topic.equalsIgnoreCase("re:remove_agent")
					|| m.topic.equalsIgnoreCase("re:not_existing")) {
				removeFromContacts(m.senderId);
				return super.readMessage(m);
			} else {
				return super.readMessage(m);
			}

		}

		/**
		 *
		 * @param senderId
		 */
		private void removeFromContacts(long senderId) {
			if (maleOwnerId == null ? maleOwnerId == null : senderId == maleOwnerId) {
				maleOwnerId = null;
			}
			if (femaleOwnerId == null ? femaleOwnerId == null : senderId == femaleOwnerId) {
				femaleOwnerId = null;

			}
		}

		/**
		 *
		 * @return
		 */
		public AgentState changeState() {
			if (!offspring.isEmpty() && checkIfEggsHatch()) {
				for (Animal egg : offspring) {
					egg.exchangeState();
				}
				dateOfHatching = World.getDate();
				HashMap<String, Object> para = new HashMap<String, Object>();
				para.put("stateName", "Nestlings");
				return createAgentState(para);
			} else {
				return this;
			}
		}

		/**
		 *
		 * @return
		 */
		public abstract boolean checkIfEggsHatch();

		/**
		 *
		 * @param incubationTime
		 */
		public abstract void receiveIncubation(double incubationTime);

		/**
		 *
		 *
		 * @param date
		 */
		public void setDateOfHatching(Calendar date) {
			dateOfHatching = date;
		}
	}

	/**
	 *
	 */
	public class Nestlings extends AgentState {

		/**
		 *
		 */
		private HashSet<Resource> todaysResources;

		/**
		 *
		 */
		/**
		 *
		 * @param parameters
		 */
		public Nestlings(HashMap<String, Object> parameters) {
			super();
			todaysResources = new HashSet<Resource>();
		}

		@Override
		public boolean checkDeath(String causeOfDeath) {
			if (causeOfDeath == null) {
				deadOffspring.clear();
				if (offspring.isEmpty()) {
					causeOfDeath = "fledged";
				} else {
					if (checkIfPredated()) {
						causeOfDeath = "predated";
					} else if (checkIfDestroyed()) {
						return super.checkDeath("destroyed");
					} else {
						Iterator<Animal> iterator = offspring.iterator();
						while (iterator.hasNext()) {
							Animal nestling = iterator.next();
							if (nestling.getState(getId()).checkDeath(null)
									|| ((Animal.Nestling) nestling.getState(getId())).checkSurvivalEffect(nestlingsSurvivalEffect)) {
								iterator.remove();
								deadOffspring.add(nestling);
							}
						}
						if (offspring.isEmpty()) {
							causeOfDeath = "deadOffspring";
						}
					}
				}
			}
			if (offspring.isEmpty() && success && femaleOwnerId != null) {
				sendMessage(new Message(getId(), femaleOwnerId, "breeding_success"));
			}
			if (causeOfDeath != null) {
				return super.checkDeath(causeOfDeath);
			} else {
				return false;
			}
		}

		/**
		 *
		 * @return
		 */
		public boolean checkIfPredated() {
			return Config.getGenerator().nextDouble() < (Double) Config.getParameter("Nestlings_predationRate");
		}

		/**
		 *
		 */
		@Override
		public void reset() {
			nestFoodRequirements = 0;
			todaysResources.clear();
			for (Animal nestling : offspring) {
				nestling.resetState();
			}
			super.reset();
		}

		/**
		 *
		 */
		@Override
		public void findFood() {
			if (maleOwnerId != null && femaleOwnerId != null) {
				sendMessage(new Message(getId(), maleOwnerId, "feeding",
						0.5 * nestFoodRequirements));
				sendMessage(new Message(getId(), femaleOwnerId, "feeding",
						0.5 * nestFoodRequirements));
			} else if (maleOwnerId != null && femaleOwnerId == null) {
				sendMessage(new Message(getId(), maleOwnerId, "feeding",
						nestFoodRequirements));
			} else if (maleOwnerId == null && femaleOwnerId != null) {
				sendMessage(new Message(getId(), femaleOwnerId, "feeding",
						nestFoodRequirements));
			} else {
				for (Animal nestling : offspring) {
					((Animal.Nestling) nestling.getState(getId())).eat();
				}
			}
		}

		/**
		 *
		 */
		@Override
		public void manageEnergy() {
			if (nestFoodRequirements == 0) {
				for (Animal nestling : offspring) {
					if (nestling.getAlive() == true) {
						((AnimalState) nestling.getState(getId())).manageEnergy();
						nestFoodRequirements += nestling.getAnimalEnergyRequirements();
					}
				}
			}
		}

		/**
		 *
		 * @param m
		 * @return
		 */
		@Override
		public Message readMessage(Message m) {
			if (m.topic.equalsIgnoreCase("re:feeding")
					|| m.topic.equalsIgnoreCase("re:dead")) {
				if (m.attachment != null) {
					todaysResources.addAll((HashSet<Resource>) m.attachment);
				}
				if (!checkIfMessagePending("feeding")) {
					for (Resource r : todaysResources) {
						double amount = r.amount;
						for (Animal nestling : offspring) {
							double possibleAmount = (nestling.getAnimalEnergyRequirements() / nestFoodRequirements)
									* amount;
							((AnimalState) nestling.getState(getId())).addTodaysResource(r.splitResource(possibleAmount));
						}
					}
					Iterator<Animal> iterator = offspring.iterator();
					while (iterator.hasNext()) {
						Animal nestling = iterator.next();
						((Animal.Nestling) nestling.getState(getId())).eat();
						if (nestling.getDaysSinceLeavingNest() >= 0) {
							nestling.exchangeState();
							iterator.remove();
							nestling.launchMe();
						}
					}
				}
				return null;
			} else if (m.topic.equalsIgnoreCase("dead_agent")
					| m.topic.equalsIgnoreCase("remove_agent")) {
				removeFromContacts(m.senderId);
				return super.readMessage(m);
			} else if (m.topic.equalsIgnoreCase("re:breeding_success")) {
				return null;
			} else {
				return super.readMessage(m);
			}
		}

		/**
		 *
		 * @param senderId
		 */
		private void removeFromContacts(long senderId) {
			if (maleOwnerId == null ? maleOwnerId == null : senderId == maleOwnerId) {
				maleOwnerId = null;
			}
			if (femaleOwnerId == null ? femaleOwnerId == null : senderId == femaleOwnerId) {
				femaleOwnerId = null;

			}
		}

		/**
		 *
		 * @return
		 */
		@Override
		public AgentState changeState() {
			return this;
		}

		/**
		 * Hatches eggs to nestlings when they reach appropriate age and mass.
		 */
		@Override
		public void conductSeasonalActivity() {
			Iterator<Animal> iterator = offspring.iterator();
			while (iterator.hasNext()) {
				Animal nestling = iterator.next();
				if (nestling.getAlive()) {
					if (getDaysSinceHatching() > (Integer) Config.getParameter("Nestlings_fledgingAge")
							&& nestling.getBodyweight() > (Double) Config.getParameter("Nestlings_fledgingBodyweight")) {
						if (nestling.getBodyweight() > (Double) Config.getParameter("Nestlings_maxFledgingBodyweight")
								|| getDaysSinceHatching() > (Integer) Config.getParameter("Nestlings_maxFledgingAge")
								|| Config.getGenerator().nextDouble() < (Double) Config.getParameter("Nestlings_fledgingProbability")) {
							prepareNestling(nestling);
							nestling.setNestLeavingDate();
							numberFledged++;
							success = true;
						}
					}
				}
			}
		}

		/**
		 *
		 * @param nestling
		 */
		public void prepareNestling(Animal nestling) {
		}

		/**
		 *
		 * @return
		 */
		public HashSet<Animal> getOffspring() {
			return offspring;
		}
	}

	/*
	 * Fields:
	 */
	/**
	 * Set of alive Animals in state
	 * <code>egg</code> or
	 * <code>nestling</code> in the nest.
	 */
	private HashSet<Animal> offspring;
	/**
	 * Set of dead Animals in state
	 * <code>egg</code> or
	 * <code>nestling</code> in the nest.
	 */
	private HashSet<Animal> deadOffspring;
	/**
	 * Number of eggs laid in this nest.
	 */
	private int eggsLaid;
	/**
	 * Id of the male owner of the territory, if there is one. If there is no
	 * female owner, points to
	 * <code>null</code>.
	 */
	private Long maleOwnerId;
	/**
	 * Id of the female owner of the territory, if there is one. If there is no
	 * female owner, points to
	 * <code>null</code>.
	 */
	private Long femaleOwnerId;
	/**
	 * Date when eggs were laid.
	 */
	private Calendar dateOfLaying;
	/**
	 * Date when offspring hatched.
	 */
	private Calendar dateOfHatching;
	private double nestFoodRequirements;
	private boolean success;
	private int numberFledged;
	private double eggsHatchingEffect;
	private double nestlingsSurvivalEffect;

	/**
	 * Constructor for Nest. <p> Parameters required in argument:
	 * HashSet<Double> eggWeights, Integer femaleOwnerId, Integer maleOwnerId,
	 * parameters required by createAgentState(). </p>
	 *
	 * @param parameters
	 */
	public Nest(HashMap<String, Object> parameters) {
		super();
		//System.out.println("nest");
		offspring = new HashSet<Animal>();
		deadOffspring = new HashSet<Animal>();
		femaleOwnerId = (Long) parameters.get("femaleOwnerId");
		if (parameters.containsKey("maleOwnerId")) {
			maleOwnerId = (Long) parameters.get("maleOwnerId");
		} else {
			maleOwnerId = null;
		}
		eggsLaid = ((LinkedList<Double>) parameters.get("eggsWeights")).size();
		dateOfLaying = World.getDate();
		dateOfHatching = null;
		nestFoodRequirements = 0;
		initialiseState(createAgentState(parameters));
		Iterator<Double> weightsIterator = ((LinkedList<Double>) parameters.get("eggsWeights")).iterator();
		for (int i = 0; i < eggsLaid; i++) {
			parameters.put("at_initialization", false);
			parameters.put("basalEnergy", weightsIterator.next());
			offspring.add(World.Instance().speciesFactory.createAnimal(parameters));
		}
		success = false;
		numberFledged = 0;
		if (parameters.containsKey("eggsHatchingEffect")) {
			eggsHatchingEffect = (Double) parameters.get("eggsHatchingEffect");
		} else {
			eggsHatchingEffect = 0.0;
		}
		if (parameters.containsKey("nestlingsSurvivalEffect")) {
			nestlingsSurvivalEffect = (Double) parameters.get("nestlingsSurvivalEffect");
		} else {
			nestlingsSurvivalEffect = 0.0;
		}
	}
	
	public boolean checkIfDestroyed() {
		return false;
	}

	/**
	 * Getter for male owner ID.
	 *
	 * @return ID of male owner of the nest or
	 * <code>null</code> if it does not exist.
	 */
	public Long getMaleOwnerId() {
		return maleOwnerId;
	}

	/**
	 * Getter for female owner ID.
	 *
	 * @return ID of female owner of the nest or
	 * <code>null</code> if it does not exist.
	 */
	public Long getFemaleOwnerId() {
		return femaleOwnerId;
	}

	/**
	 * Calculates age in days since laying.
	 *
	 * @return Number of days since laying.
	 */
	public int getDaysSinceLaying() {
		return World.calculateDateDifference(dateOfLaying);
	}

	/**
	 * Calculates age in days since hatching.
	 *
	 * @return Number of days since hatching.
	 */
	public int getDaysSinceHatching() {
		if (dateOfHatching != null) {
			return World.calculateDateDifference(dateOfHatching);
		} else {
			return -1;
		}
	}

	/**
	 * Getter for date of laying.
	 *
	 * @return Copy of date of laying.
	 */
	public Calendar getDateOfLaying() {
		return (Calendar) dateOfLaying.clone();
	}

	/**
	 * Returns the number of offspring present in the nest.
	 *
	 * @return Number of offspring.
	 */
	public int getOffspringSize() {
		return offspring.size();
	}

	/**
	 *
	 * @return
	 */
	public HashMap<Long, HashMap<String, Double>> getOffspringMaxPesticides() {
		HashMap<Long, HashMap<String, Double>> result = new HashMap<Long, HashMap<String, Double>>();
		for (Animal a : offspring) {
			result.put(a.getId(), a.getMaxPesticides());
		}
		return result;
	}

	/**
	 *
	 * @return
	 */
	public HashMap<Long, HashMap<String, Double>> getOffspringCumulativePesticides() {
		HashMap<Long, HashMap<String, Double>> result = new HashMap<Long, HashMap<String, Double>>();
		for (Animal a : offspring) {
			result.put(a.getId(), a.getCumulativePesticides());
		}
		return result;
	}

	/**
	 *
	 * @return
	 */
	public HashMap<Long, HashMap<String, Integer>> getOffspringExposureTime() {
		HashMap<Long, HashMap<String, Integer>> result = new HashMap<Long, HashMap<String, Integer>>();
		for (Animal a : offspring) {
			result.put(a.getId(), a.getExposureTime());
		}
		return result;
	}

	@Override
	public Signature getSignature() {
		// add a table of intigers for storing ids of resources eaten this day
		Signature mySignature = super.getSignature();
		if (femaleOwnerId != null) {
			mySignature.addPrimitiveField("femaleOwnerId", "bigint",
					Long.toString(femaleOwnerId));
		} else {
			mySignature.addPrimitiveField("femaleOwnerId", "bigint", "null");
		}
		if (maleOwnerId != null) {
			mySignature.addPrimitiveField("maleOwnerId", "bigint",
					Long.toString(maleOwnerId));
		} else {
			mySignature.addPrimitiveField("maleOwnerId", "bigint", "null");
		}
		String offspringIds = "'{";
		for (Animal young : offspring) {
			if (young.getDaysSinceLeavingNest() < 0) {
				offspringIds += String.valueOf(young.getId()) + ",";
			}
		}
		if (offspringIds.equalsIgnoreCase("'{")) {
			offspringIds += "}'";
		} else {
			offspringIds = offspringIds.substring(0, offspringIds.length() - 1)
					+ "}'";
		}
		mySignature.addPrimitiveField("offspringIds", "bigint[]", offspringIds);
		for (Animal young : offspring) {
			if (young.getDaysSinceLeavingNest() < 0) {
				mySignature.addInternalObject(young.getSignature());
			}
		}
		mySignature.addPrimitiveField("agesincelaying", "int", Integer.toString(getDaysSinceLaying()));
		mySignature.addPrimitiveField("agesincehatching", "int", Integer.toString(getDaysSinceHatching()));
		Class stateClass = getStateClass();
		int nestlings = 0;
		int eggs = 0;
		int animals = 0;
		while (stateClass != null) {
			if (stateClass.getSimpleName().equalsIgnoreCase("Eggs")) {
				eggs = offspring.size();
				break;
			} else if (stateClass.getSimpleName().equalsIgnoreCase("Nestlings")) {
				nestlings = offspring.size();
				break;
			} else {
				stateClass = stateClass.getSuperclass();
			}
		}
		mySignature.addPrimitiveField("numberOfNestlings", "integer", Integer.toString(nestlings));
		mySignature.addPrimitiveField("numberOfEggs", "integer", Integer.toString(eggs));
		return mySignature;
	}
	
	public int getNumberFledged() {
		return numberFledged;
	}
}
