//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.sql.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Collects <code>Signatures</code> of the <code>Observable</code> objects 
 * and stores them  together with the configuration and initialisation settings 
 * in the PostgreSQL database.
 *
 * @author Tomasz Kułakowski
 * @author Katarzyna Kułakowska
 * @since 2013-01-04
 */
public class Observer {

	/*
	 * Fields:
	 */
	/**
	 *
	 */
	private Connection connection;
	/**
	 *
	 */
	private HashSet<String> createdTables;
	private Statement statement;

	/*
	 * Methods:
	 */
	/**
	 * Creates
	 * <code>Observer</code> using the specified database.
	 *
	 * @param databaseName
	 */
	public Observer(String databaseName) {
		createdTables = new HashSet<String>();
		try {
			Class.forName("org.postgresql.Driver");
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		try {
//			connection = DriverManager.getConnection("jdbc:postgresql://10.1.1.1:5432/"
//					+ databaseName);
			connection = DriverManager.getConnection("jdbc:postgresql:"
					+ databaseName);
			statement = connection.createStatement();
		} catch (SQLException ex) {
			Logger.getLogger(Observer.class.getName()).log(Level.SEVERE, null,
					ex);
		}
	}

	/**
	 * Passes the
	 * <code>signature</code> of the specified
	 * <code>Observable</code> object to method
	 * <code>storeSignature</code>.
	 *
	 * @param observed
	 * <code>Observable</code> object
	 */
	public void observe(Collection<Observable> observed) {
		try {
//			Statement statement = connection.createStatement();
			for (Observable o : observed) {
				storeSignature(o.getSignature(), statement);
			}
			statement.executeBatch();
//			statement.close();
		} catch (SQLException e) {
			Logger.getLogger(Observer.class.getName()).log(Level.SEVERE, null,
					e);
			Logger.getLogger(Observer.class.getName()).log(Level.SEVERE, null,
					e.getNextException());
		}
	}

	/**
	 *
	 */
	public void shutdown() {
		try {
			statement.close();
			connection.close();
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}

	/**
	 * 
	 * @param firstSignature
	 * @param statement
	 * @throws SQLException
	 */
	public void ensureTableExists(Signature firstSignature, Statement statement)
			throws SQLException {
		if (!getTables(firstSignature.getTableName()).next()) {
			statement.executeUpdate(generateCreateTableStatement(firstSignature));
		}
	}

	/**
	 * Executes update of given
	 * <code>signature</code> of all its
	 * <code>primitiveFields</code> and all subsequent
	 * <code>internalObjects</code>
	 * <code>signatures</code>.
	 *
	 * @param signature
	 * @param statement
	 * @throws SQLException  
	 */
	public void storeSignature(Signature signature, Statement statement)
			throws SQLException {
		if (signature != null) {
			if (createdTables.add(signature.getTableName())) {
				ensureTableExists(signature, statement);
			}
			statement.addBatch(generateUpdate(signature));
			for (Signature subSignature : signature.getInternalObjects()) {
				storeSignature(subSignature, statement);
			}
		}
	}

	/**
	 * Generates a string with SQL command to store contents of given
	 * <code>signature</code>
	 * <code>primitiveFields</code> into database.
	 *
	 * @param signature
	 * @return
	 */
	public String generateUpdate(Signature signature) {
		String update = "INSERT INTO " + signature.getTableName() + " (";
		for (String fieldName : signature.getPrimitiveFields().navigableKeySet()) {
			update += fieldName + ", ";
		}
		update = update.substring(0, update.length() - 2) + ") VALUES (";
		for (String fieldName : signature.getPrimitiveFields().navigableKeySet()) {
			update += signature.getPrimitiveFields().get(fieldName)[1] + ", ";
		}
		return update.substring(0, update.length() - 2) + ");";
	}

	/**
	 *
	 * @param signature
	 * @return
	 */
	public String generateCreateTableStatement(Signature signature) {
		String createTableStatement = "CREATE TABLE IF NOT EXISTS "
				+ signature.getTableName() + " (";
		for (String fieldName : signature.getPrimitiveFields().navigableKeySet()) {
			createTableStatement += fieldName + " "
					+ signature.getPrimitiveFields().get(fieldName)[0] + ", ";
		}
		createTableStatement += signature.getPrimaryKey();
		return createTableStatement + ");";
	}

	/**
	 * 
	 * @param filename
	 */
	public void storeConfiguration(String filename) {
		try {
			DatabaseMetaData metaData = connection.getMetaData();
			ResultSet tables = metaData.getTables(null, null,
					(String) Config.getParameter("Observer_configurationTableName"), null);
			if (!tables.next()) {
				statement.executeUpdate("CREATE TABLE IF NOT EXISTS "
						+ (String) Config.getParameter("Observer_configurationTableName")
						+ " (starttime text, parameter text, type text, value text);");
			}
		} catch (SQLException e) {
			Logger.getLogger(Observer.class.getName()).log(Level.SEVERE, null,
					e);
		}
		BufferedReader in = null;
		try {
			in = new BufferedReader(new FileReader(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.out.println("Configuration file not found. Aborting reading.");
			return;
		}

		String line;
		Scanner myScanner;
		try {
			while ((line = in.readLine()) != null) {
				if (!line.startsWith("#")) {
					myScanner = new Scanner(line);
					if (myScanner.hasNext()) {
						String update = "INSERT INTO "
								+ (String) Config.getParameter("Observer_configurationTableName")
								+ " VALUES ('" + Config.getParameter("startTime") + "', '";
						while (myScanner.hasNext()) {
							update += myScanner.next() + "', '";
						}
						update = update.substring(0, update.length() - 3) + ");";
						statement.executeUpdate(update);
					}
				}
			}
			statement.executeUpdate("DELETE FROM "
					+ (String) Config.getParameter("Observer_configurationTableName")
					+ " WHERE starttime = '"
					+ Config.getParameter("startTime")
					+ "' AND parameter = 'World_randomSeed';");
			statement.executeUpdate("INSERT INTO "
								+ (String) Config.getParameter("Observer_configurationTableName")
								+ " VALUES ('" + Config.getParameter("startTime") 
					+ "', 'World_randomSeed', 'long', '" 
					+ Long.toString((Long) Config.getParameter("World_randomSeed")) 
					+ "');");
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("IOException during reading of configuration file. Aborting.");
		} catch (SQLException e) {
			Logger.getLogger(Observer.class.getName()).log(Level.SEVERE, null, e);
		}

	}

	/**
	 * 
	 */
	public void storeInitialisation() {
		try {
			DatabaseMetaData metaData = connection.getMetaData();
			ResultSet tables = metaData.getTables(null, null,
					(String) Config.getParameter("Observer_initialisationTableName"), null);
			if (!tables.next()) {
				statement.executeUpdate("CREATE TABLE "
						+ (String) Config.getParameter("Observer_initialisationTableName")
						+ " (name text PRIMARY KEY, file oid);");
			}
			ResultSet result = statement.executeQuery("SELECT name FROM "
					+ (String) Config.getParameter("Observer_initialisationTableName")
					+ " WHERE name = '"
					+ (String) Config.getParameter("World_initialisationFilename") + "';");
			if (!result.next()) {
				statement.executeUpdate("INSERT INTO "
						+ (String) Config.getParameter("Observer_initialisationTableName")
						+ " VALUES ('" + (String) Config.getParameter("World_initialisationFilename")
						+ "', lo_import('" + (String) Config.getParameter("World_initialisationFilename")
						+ "'));");
			}
		} catch (SQLException e) {
			Logger.getLogger(Observer.class.getName()).log(Level.SEVERE, null, e);
		}
	}
	
	/**
	 * 
	 * @throws SQLException
	 */
	public void executeBatch() throws SQLException {
		statement.executeBatch();
	}
	/**
	 * 
	 * @param batch
	 * @throws SQLException
	 */
	public void addBatch(String batch) throws SQLException {
		statement.addBatch(batch);
	}
	/**
	 * 
	 * @param name
	 * @return
	 * @throws SQLException
	 */
	public ResultSet getTables(String name) throws SQLException {
		DatabaseMetaData metaData = connection.getMetaData();
		return metaData.getTables(null, null,
				name, null);
	}
	/**
	 * 
	 * @param s
	 * @throws SQLException
	 */
	public void executeStatement(String s) throws SQLException {
		statement.execute(s);
	}
	
	/**
	 * 
	 * @param x
	 * @return
	 */
	public static String prepareUpdate(Double x) {
		String up;
		if (x != null) {
			up = Double.toString(x);
		} else {
			up = "null";
		}
		return up;
	}

	/**
	 * 
	 * @param x
	 * @return
	 */
	public static String prepareUpdate(Integer x) {
		String up;
		if (x != null) {
			up = Integer.toString(x);
		} else {
			up = "null";
		}
		return up;
	}
	
		/**
	 * 
	 * @param x
	 * @return
	 */
	public static String prepareUpdate(Boolean x) {
		String up;
		if (x != null) {
			up = Boolean.toString(x);
		} else {
			up = "null";
		}
		return up;
	}
	
	/**
	 * 
	 * @param x
	 * @return
	 */
	public static String prepareUpdate(ArrayList<Double> x) {
		String up = "'{";
		for (double d : x) {
			up += String.valueOf(d) +  ",";
		}
		if(up.equalsIgnoreCase("'{")) {
				up += "}'";
		} else {
			up = up.substring(0, up.length() - 1)+ "}'";
		}

		return up;
	}
	
	public static String prepareUpdate(SortedSet<Double> x) {
		String up = "'{";
		for (double d : x) {
			up += String.valueOf(d) +  ",";
		}
		if(up.equalsIgnoreCase("'{")) {
				up += "}'";
		} else {
			up = up.substring(0, up.length() - 1)+ "}'";
		}

		return up;
	}

	/**
	 * 
	 * @param list
	 * @return
	 */
	public static Double avg(ArrayList<Double> list) {
		if (list.size() > 0) {
			double tot = 0.0;
			for (Double d : list) {
				tot += d;
			}
			return tot / new Double(list.size());
		} else {
			return null;
		}
	}

	public static Double avg(List list) {
		if (list.size() > 0) {
			double tot = 0.0;
			Iterator it = list.iterator();
			while (it.hasNext()) {
				Object x = it.next();
				try {
					int i = (Integer) x;
					tot += i;
				} catch(ClassCastException e1) {
					try {
						double d = (Double) x;
						tot += d;
					} catch(ClassCastException e2) {
						// *****VERIFICATION MODE START*****//
						if (World.verificationMode) {
							try {
								throw new SkyPigException(
										"Error: Avarage calculated neither from Integers nor Doubles.");
							} catch (SkyPigException e) {
								World.log(Level.SEVERE,
										Config.getLogInfo(), e);
								World.log(Level.SEVERE,
										Config.getLogInfo(), e1);
								World.log(Level.SEVERE,
										Config.getLogInfo(), e2);
							}
						}
						//*****VERIFICATION MODE END*****//
					}
				}
			}
			return tot / new Double(list.size());
		} else {
			return null;
		}
	}
	
	/**
	 * 
	 * @param list
	 * @return
	 */
	public static Double min(List<Double> list) {
		if (list.size() > 0) {
			double m = list.get(0);
			for (Double d : list) {
				m = Math.min(m, d);
			}
			return m;
		} else {
			return null;
		}
	}

	public static Integer min_i(List<Integer> list) {
		if (list.size() > 0) {
			int m = list.get(0);
			for (Integer i : list) {
				m = Math.min(m, i);
			}
			return m;
		} else {
			return null;
		}
	}

	/**
	 * 
	 * @param list
	 * @return
	 */
	public static Double max(List<Double> list) {
		if (list.size() > 0) {
			double m = list.get(0);
			for (Double d : list) {
				m = Math.max(m, d);
			}
			return m;
		} else {
			return null;
		}
	}

	public static Integer max_i(List<Integer> list) {
		if (list.size() > 0) {
			int m = list.get(0);
			for (Integer i : list) {
				m = Math.max(m, i);
			}
			return m;
		} else {
			return null;
		}
	}

	/**
	 * 
	 * @param x
	 * @return
	 */
	public static double sqr(double x) {
		return x * x;
	}

	/**
	 * 
	 * @param list
	 * @return
	 */
	public static Double variance(ArrayList<Double> list) {
		if (list.size() > 0) {
			Double avg = avg(list);
			double sumsq = 0.0;
			for (Double d : list) {
				sumsq += sqr(avg - d);
			}
			return sumsq / new Double(list.size());
		} else {
			return null;
		}

	}

	/**
	 * 
	 * @param list
	 * @return
	 */
	public static Double sdev(ArrayList<Double> list) {
		if (list.size() > 0) {
			return Math.sqrt(variance(list));
		} else {
			return null;
		}
	}

	/**
	 * 
	 * @param list
	 * @return
	 */
	public static Double serr(ArrayList<Double> list) {
		if (list.size() > 0) {
			return sdev(list) / Math.sqrt(new Double(list.size()));
		} else {
			return null;
		}
	}

	/**
	 * 
	 * @param list
	 * @param perc
	 * @return
	 */
	public static Double percentile(List<Double> list, double perc) {
		if (list.size() > 0) {
			Collections.sort(list);
			int index = (int) Math.round(perc * new Double(list.size()) / 100);
			if (index >= list.size()) {
				return list.get(list.size() - 1);
			} else {
				return list.get(index);
			}
		} else {
			return null;
		}
	}

	public static Integer percentile_i(List<Integer> list, double perc) {
		if (list.size() > 0) {
			Collections.sort(list);
			int index = (int) Math.round(perc * new Double(list.size()) / 100);
			if (index >= list.size()) {
				return list.get(list.size() - 1);
			} else {
				return list.get(index);
			}
		} else {
			return null;
		}
	}

	/**
	 * 
	 * @param list
	 * @param value
	 * @return
	 */
	public static Double proportionOfValues(ArrayList<Double> list, double value) {
		if (list.size() > 0) {
			double n = 0;
			for (Double i : list) {
				if (i == value) {
					n++;
				}
			}
			return n / new Double(list.size());
		} else {
			return null;
		}
	}

	/**
	 * 
	 * @param x
	 * @param y
	 * @return
	 */
	public static Double ratio(int x, int y) {
		if (y != 0) {
			return new Double(x) / new Double(y);
		} else {
			return null;
		}
	}
	
		/**
	 * 
	 * @param list
	 * @return
	 */
	public static ArrayList<Double> prepareResult(ArrayList<Double> list) {
		if (list.size() > 1000) {
			ArrayList<Double> newList = new ArrayList<Double>();
			HashSet<Integer> indices = new HashSet<Integer>();
			while (indices.size() < 1000) {
				indices.add(Config.getGenerator().nextInt(list.size()));
			}
			for (int i : indices) {
				newList.add(list.get(i));
			}
			return newList;
		} else {
			return list;
		}
	}
}
