//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.util.HashSet;
import java.util.logging.Level;

/**
 * Serves for sending Messages and storing their copies until they are replied
 * to.
 *
 * @author Katarzyna Kułakowska
 * @author Tomasz Kułakowski
 * @version 1.0
 */
public class Outbox {

	/*
	 * Fields:
	 */
	/**
	 * Set of sent Messages.
	 */
	private HashSet<Message> storage;

	/*
	 * Methods:
	 */
	/**
	 * Creates a new Outbox with an empty storage.
	 */
	public Outbox() {
		storage = new HashSet<Message>();
	}

	/**
	 * Checks if the Outbox contains any unreplied Messages.
	 *
	 * @return true if the storage is empty, false otherwise
	 */
	public boolean isEmpty() {
		return storage.isEmpty();
	}

	/**
	 * Deliveres the Message to the recipients' inbox.
	 *
	 * <p> If the message is a start of a new communication (inReplyTo is null)
	 * it is added to the outbox storage. If it is a reply, it is not added.
	 * </p>
	 *
	 * @param m Message to be sent 
	 * @verification TEST 1 Warning if the recipient of the message is not specified, 
	 * TEST 2 Warning if the same message is sent twice.
	 */
	public void sendMessage(Message m) {
		if (World.Instance().getAgent(m.recipientId) == null) {
//*****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					//***TEST 1
					try {
							throw new SkyPigException("No recipient for the message "
									+ m.topic);
						} catch (SkyPigException ex) {
							World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
//*****VERIFICATION MODE END*****//	
			sendMessage(new Message(m.recipientId, m.senderId, "re:not_existing", m));
		} else {
			if (m.inReplyTo == null) {
//*****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					//***TEST 1
					if (storage.contains(m)) {
						try {
							throw new SkyPigException("Message added to outbox twice: "
									+ m.topic);
						} catch (SkyPigException ex) {
							World.log(Level.WARNING, Config.getLogInfo(), ex);
						}
					}
				}
//*****VERIFICATION MODE END*****//
				storage.add(m);
			}
			World.Instance().getAgent(m.recipientId).receiveMessage(m);
		}
	}

	/**
	 *
	 * @param m
	 * @return true if the requested message 
	 * @verification TEST 1 Warning if the message specified for removal is not 
	 * contained in the storage. 
	 */
	public boolean removeMessage(Message m) {
		boolean isInStorage = storage.remove(m);
//*****VERIFICATION MODE START*****//
		if (World.verificationMode) {
			//***TEST 
			if (!isInStorage && m != null) {
				try {
					throw new SkyPigException("Message to be removed not"
							+ " found in the Outbox ");// + m.topic);
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);

				}
			}
		}
//*****VERIFICATION MODE END*****//
		return isInStorage;
	}
	/**
	 * Returns true of there is a message with the specified topic in the outbox.
	 * 
	 * @param topic
	 * @return 
	 */
	public boolean findMessage(String topic) {
        for (Message m : storage) {
            if (m.topic.equals(topic)) {
                return true;
            }
        }
        return false;
	}
}
