//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package model.core;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

/**
 * @author beholder
 *
 */
public class Resource implements Cloneable, Observable {

	/**
	 *
	 */
	private static long ID = 0;
	/**
	 *
	 */
	private long id;
	/**
	 *
	 */
	private String name;
	/**
	 * Pesticides and their amounts in mg
	 * 
	 * @since 2012-10-22
	 */
	private HashMap<String, Double> pesticides;
	/**
	 * Amount of resource in grams
	 * 
	 * @since 2012-10-22
	 */
	public double amount;
	/**
	 *
	 */
	public double energyContent;
	
	/**
	 * Density in g/ha
	 * 
	 * @since 2012-10-22
	 */
	public double density;

	/**
	 *
	 * @param parameters
	 */
	public Resource(HashMap<String, Object> parameters) {
		name = (String) parameters.get("name");
		id = ID++;
		pesticides = new HashMap<String, Double>();
		amount = (Double) parameters.get("amount");
		energyContent = (Double) Config.getParameter(name
				+ "_metabolizableEnergy");
		density = 0;

	}

	/**
	 *
	 * @param name
	 * @param amount
	 * @param energyContent  
	 */
	public Resource(String name, double amount, double energyContent) {
		id = ID++;
		this.name = name;
		this.amount = amount;
		pesticides = new HashMap<String, Double>();
		this.energyContent = energyContent;
		density = 0;		
	}

	/**
	 *
	 * @param amount
	 * @return
	 */
	public Resource splitResource(double amount) {
		Resource res;
		if (amount < this.amount) {
			double proportion = amount / this.amount;
			this.amount -= amount;
			res = new Resource(name, amount, energyContent);
			for (String pesticide : pesticides.keySet()) {
				double originalPesticideAmount = pesticides.get(pesticide);
				double pesticideAmount = proportion * originalPesticideAmount;
				originalPesticideAmount -= pesticideAmount;
				pesticides.put(pesticide, originalPesticideAmount);
				res.addPesticide(pesticide, pesticideAmount);
			}
		} else {
			amount = this.amount;
			this.amount = 0.0;
			res = new Resource(name, amount, energyContent);
			for (String pesticide : pesticides.keySet()) {
				res.addPesticide(pesticide, pesticides.get(pesticide));
			}
			pesticides.clear();
		}
		return res;
	}

	/**
	 * 
	 * @param amount
	 * @return
	 */
	public Resource shareResource(double amount) {
		Resource res;
		if (amount < this.amount) {
			double proportion = amount / this.amount;
			res = new Resource(name, amount, energyContent);
			for (String pesticide : pesticides.keySet()) {
				res.addPesticide(pesticide, proportion * getPesticideAmount(pesticide));
			}
		} else {
			amount = this.amount;
			res = new Resource(name, amount, energyContent);
			for (String pesticide : pesticides.keySet()) {
				res.addPesticide(pesticide, getPesticideAmount(pesticide));
			}
		}
		return res;
	}

	/**
	 *
	 * @param name 
	 * @param amount  
	 */
	public void addPesticide(String name, Double amount) {
		pesticides.put(name, amount);
	}

	/**
	 * 
	 * @return
	 */
	public HashMap<String, Double> getPesticides() {
		return pesticides;
	}

	/**
	 * 
	 */
	public void decayPesticides(/*double area*/) {
		Iterator<Map.Entry<String, Double>> iterator = pesticides.entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<String, Double> pesticide = iterator.next();
			Double newAmount;
			if ((Double) Config.getParameter(pesticide.getKey() + "_fieldHalfLife") > 0) {
				newAmount = pesticide.getValue() * Math.pow(0.5,
						(1.0 / (Double) Config.getParameter(pesticide.getKey() + "_fieldHalfLife")));
			} else {
				newAmount = 0.0;
			}
			if (newAmount/(amount/1000.0)/*/area*/ <= (Double) Config.getParameter(pesticide.getKey() + "_minimumDetectableMgPerKgFood")) {
				iterator.remove();
			} else {
				pesticides.put(pesticide.getKey(), newAmount);
			}
		}
	}
	
	/**
	 * 
	 * @param proportion
	 */
	public void modifyPesticides(double proportion/*, double area*/) {
		Iterator<Map.Entry<String, Double>> iterator = pesticides.entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<String, Double> pesticide = iterator.next();
			Double newAmount = pesticide.getValue() *  proportion;
			if (newAmount/(amount/1000.0)/*/area*/ <= (Double) Config.getParameter(pesticide.getKey() + "_minimumDetectableMgPerKgFood")) {
				iterator.remove();
			} else {
				pesticides.put(pesticide.getKey(), newAmount);
			}
		}
	}

	/**
	 *
	 *
	 * @return
	 */
	public String getName() {
		return name;
	}

	/**
	 *
	 * @return
	 */
	public long getId() {
		return id;
	}
	
	/**
	 * 
	 * @param size
	 * @return
	 */
	public double getDensity(double size) {
		if (density > amount/size) {
			return density;
		} else {
			return amount/size;
		}
	}

	/*
	 * (non-Javadoc) @see model.core.Observable#getSignature()
	 */
	/**
	 *
	 * @return
	 */
	@Override
	public Signature getSignature() {
		Signature mySignature = new Signature(this);
		mySignature.addPrimitiveField("id", "bigint", Long.toString(id));
		mySignature.addPrimitiveField("name", "text", "'" + name + "'");
		mySignature.addPrimitiveField("amount", "double precision", Double.toString(amount));
		mySignature.addPrimitiveField("energyContent", "double precision", Double.toString(energyContent));
		mySignature.addPrimaryKey("id");
		for (String pesticide : pesticides.keySet()) {
			Signature pestSignature = new Signature(this);
			pestSignature.setTableName((String) Config.getParameter("Resource_pesticideTableName"));
			pestSignature.addPrimitiveField("id", "bigint", Long.toString(id));
			pestSignature.addPrimitiveField("pesticide", "text", "'" + pesticide + "'");
			pestSignature.addPrimitiveField("amount", "double precision", Double.toString(pesticides.get(pesticide)));
			pestSignature.addPrimaryKey("id");
			pestSignature.addPrimaryKey("pesticide");
			mySignature.addInternalObject(pestSignature);
		}
		return mySignature;
	}
	
	private double getPesticideAmount(String pesticide) {
		return pesticides.get(pesticide);
	}
}
