//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
package model.core;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.*;

import model.skylark.SkylarkFactory;
import model.skylark.SkylarkLandscapeUnit;
import model.woodpigeon.WoodpigeonFactory;

/**
 * The core part of the model.
 *
 * <p> Initialized at the beginning of the model execution. It is responsible
 * for selecting species, initializing Agents, logging, time management and
 * removing dead Agents. After the specified model length it returns. </p> <p>
 * Enables initializing of the verification mode. In this mode all the
 * conditional verification statements are executed and throw specified
 * exceptions. The exceptions are caught and logged to a text file named
 * startTime_configurationNumber.log. </p> <p> Includes several elements that
 * are species-specific: weather selection and initialization of territories for
 * skylarks. </p>
 *
 *
 * @author Katarzyna Kułakowska
 * @author Tomasz Kułakowski
 * @since 2012-08-07
 */
public class World implements Observable {
	/*
	 * Fields:
	 */

	/**
	 * Determines the mode of model execution: verification mode if true,
	 * otherwise normal.
	 *
	 * @since 2012-08-06
	 */
	public final static boolean verificationMode = false;
	/**
	 * Used for storing information while in the verification mode.
	 *
	 * <p> There are several levels of logginng messages: (1)SEVERE - serious
	 * failure usually stopping the model execution emerging from Java
	 * exceptions; (2) WARNING - problem indicated by exceptions defined by the
	 * authors that usually do not stop model execution but should exclude this
	 * run from result analysis; (3) INFO - information messages about reaching
	 * the milestone during the model execution; (4) FINE - tracing message
	 * documenting normal programm execution.
	 *
	 * @since 2012-08-06
	 *
	 */
	private static final Logger logger = getLogger();// Logger.getLogger("mainLogger");
	/**
	 * Collection of LandscapeUnits.
	 *
	 * @since 2012-08-06
	 */
	private HashSet<LandscapeUnit> landscapeUnits;
	/**
	 * While initialized, stores the created World.
	 *
	 * @since 2012-08-06
	 */
	private static World instance = null;
	/**
	 * Stores the current model date.
	 *
	 * @since 2012-08-06
	 */
	private Calendar date;
	/**
	 * Stores the date of the apocalypse generated from delivered model
	 * execution length.
	 *
	 * @since 2012-08-06
	 */
	private Calendar apocalypse;
	/**
	 * Collection of all alive Agents. Agents are added to it while launched and
	 * removed while dead after logging their signatures.
	 *
	 * @since 2012-08-06
	 */
	private ConcurrentHashMap<Long, Agent> agents;
	/**
	 * Stores the Observer responsible for logging the model results to a
	 * database.
	 *
	 * @since 2012-08-06
	 */
	private HashSet<Observer> observers;
	/**
	 * Stores SpeciesFactory that creates objects coming from the specified
	 * species-package (currently: woodpigeon package or skylark package).
	 *
	 * @since 2012-08-06
	 */
	public final SpeciesFactory speciesFactory;
	/**
	 * Determines the weather conditions: rain if true, otherwise no rain. <p>
	 * Used in skylark model. </p>
	 *
	 * @since 2012-08-06
	 */
	private boolean rain;
	/**
	 * Collection of active agents during the day.
	 *
	 * <p> Agents are added to this list each day after the date has been
	 * changed and removed after their execution. </p>
	 *
	 * @since 2012-08-06
	 */
	private final LinkedList<Agent> runningAgents;

	/*
	 * Methods:
	 */
	/**
	 * Makes the instance of World available to all the other classes.
	 *
	 * @return
	 * <code>World</code> instance
	 * @since 2012-08-06 @verification Warning if the
	 * <code>World</code> does not exist
	 */
	public static World Instance() {
		// *****VERIFICATION MODE START*****//
		if (verificationMode) {
			// ***TEST 1
			if (instance == null) {
				try {
					throw new SkyPigException(
							"World not properly initialised.");
				} catch (SkyPigException ex) {
					log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
		// *****VERIFICATION MODE END*****//
		return instance;
	}

	/**
	 * @return
	 */
	private static Logger getLogger() {
		if (verificationMode) {
			return Logger.getLogger("mainLogger");
		} else {
			return null;
		}
	}

	/**
	 * @param fine
	 * @param msg
	 * @param logInfo
	 */
	public static boolean log(Level fine, String msg, String logInfo) {
		if (verificationMode) {
			logger.log(Level.FINE, "Day finshed: {0}", Config.getLogInfo());
		}
		return verificationMode;
	}

	public static boolean log(Level level, String msg, Throwable thrown) {
		if (verificationMode) {
			logger.log(level,msg,thrown);
		}
		return verificationMode;
	}

	public static boolean log(Level level, String msg, Object[] params) {
		if (verificationMode) {
			logger.log(level,msg,params);
		}
		return verificationMode;
	}

	/**
	 * At the beginning of model execution creates the instance of World calling
	 * its constructor.
	 *
	 * @param observers
	 * @return newly created World instance when executed for the first time,
	 * existing World instance when executed later.
	 * @since 2012-08-06
	 */
	public static World Instance(HashSet<Observer> observers) {
		if (instance == null) {
			instance = new World(observers);
		}
		return instance;
	}

	/**
	 * Manages the Agents and time in the model. <p> If the current date is
	 * earlier then or equal to the apocalypse date, checks if all the Agents
	 * have finished their activities. If they have finished, the night begins.
	 * World logs the results, advances the date and removes dead Agents from
	 * the list
	 * <code>runningAgents</code>. Selects the weather for the next day in
	 * skylark model. Wakes Agents that start a new day. </p> <p> Otherwise, if
	 * the current date is after the apocalypse date, it returns. </p>
	 *
	 * @verification TEST 1 Warning if there are any unfinished Agents after the
	 * day has finished and before the date is changed to the next day.
	 * @parameters World_initialisationFilename, World_speciesName
	 *
	 * @since 2012-08-07
	 */
	public void run() {
		initialise((String) Config.getParameter("World_initialisationFilename"));
		date.add(Calendar.DATE, 1);
		loadAgents();
		while (date.compareTo(apocalypse) <= 0
				&& (((Boolean) Config.getParameter("World_finishWhenNoAnimals")
				&& runningAgents.size() > landscapeUnits.size())
				|| !(Boolean) Config.getParameter("World_finishWhenNoAnimals"))
				&& (runningAgents.size() - landscapeUnits.size() < (Integer) Config.getParameter("World_finishWhenMaxAnimals")
				|| (Integer) Config.getParameter("World_finishWhenMaxAnimals") == 0)) {
			checkIfDayFinished();
//			System.out.println("Agents: " + agents.size());
//			System.out.println("day "
//					+ String.valueOf(date.get(Calendar.DAY_OF_YEAR))
//					+ ", year " + String.valueOf(date.get(Calendar.YEAR)));
//			log(Level.FINE, "Day finshed: {0}", Config.getLogInfo());
			storeResults();
			// *****VERIFICATION MODE START*****//
			if (verificationMode) {
				// ***TEST 1
				for (Agent ag : agents.values()) {
					if (!ag.getFinished()) {
						try {
							throw new SkyPigException(
									"Agent not finished at night: "
									+ String.valueOf(ag.getId()));
						} catch (SkyPigException ex) {
							log(Level.WARNING, Config.getLogInfo(), ex);
						}
					}
				}
			}
			// *****VERIFICATION MODE END*****//
			date.add(Calendar.DATE, 1);
			removeDeadAgents();
			if (((String) Config.getParameter("World_speciesName")).equalsIgnoreCase("skylark")) {
				selectWeather();
			}
			loadAgents();
		}
		for (Observer observer : observers) {
			observer.shutdown();
		}
		return;
	}

	/**
	 * Adds all the agents from
	 * <code>agents</code> to
	 * <code>runningAgents</code>. <p> LandscapeUnits are added at the beginning
	 * of the list in fixed order. Nests and birds are added later in random
	 * order. </p>
	 *
	 * @since 2012-08-07
	 */
	private void loadAgents() {
		for (LandscapeUnit l : landscapeUnits) {
			l.makeUnfinished();
		}
		runningAgents.addAll(landscapeUnits);
		ArrayList<Agent> a = new ArrayList<Agent>();
		a.addAll(agents.values());
		int n = a.size();
		while (n > 0) {
			int m = Config.getGenerator().nextInt(n);
			Agent x = a.get(m);
			if (!runningAgents.contains(x)) {
				x.makeUnfinished();
				runningAgents.add(x);
			}
			a.remove(m);
			n = a.size();
		}

	}

	/**
	 * Executes agents from the
	 * <code>runningAgents</code> list as long is it is not empty.
	 *
	 * @since 2012-08-07
	 */
	private void checkIfDayFinished() {
		while (!runningAgents.isEmpty()) {
			executeAgent(runningAgents.poll());
		}
	}

	/**
	 * Enables others checking the current
	 * <code>World date</code>.
	 *
	 * @return cloned date
	 * @since 2012-08-07
	 */
	public static Calendar getDate() {
		return (Calendar) Instance().date.clone();
	}

	/**
	 * Calculates the difference in days between the current date and the data
	 * passed as parameter.
	 *
	 * @param time date for comparison with current date
	 * @return difference in days between current and passed date, positive if
	 * the passed value is earlier then current day, otherwise negative.
	 * @since 2012-08-07
	 */
	public static int calculateDateDifference(Calendar time) {
		long today = getDate().getTimeInMillis();
		long birth = time.getTimeInMillis();
		long diff = today - birth;
		return ((Long) (diff / (1000 * 60 * 60 * 24))).intValue();
	}

	/**
	 * Returns random landscapeUnit from the list
	 * <code>landscapeUnits</code>.
	 *
	 * @return random LandscapeUnit @verification TEST 1 Warning if called
	 * before the initialization of
	 * <code>World</code>
	 * @since 2012-08-07
	 */
	public static LandscapeUnit getRandomLandscapeUnit() {
		if (verificationMode) {
			// ***TEST 1
			if (instance == null) {
				try {
					throw new SkyPigException("World not instantiated.");
				} catch (SkyPigException ex) {
					log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
			// *****VERIFICATION MODE END*****//			
		}
		Object[] fields = instance.landscapeUnits.toArray();
		int randomField = Config.getGenerator().nextInt(fields.length);
		return (LandscapeUnit) fields[randomField];
	}

	/**
	 * Enables others checking the
	 * <code>apocalypse</code> date.
	 *
	 * @return cloned
	 * <code>apocalypse</code> date
	 * @since 2012-08-07
	 */
	public Calendar getApocalypse() {
		return (Calendar) apocalypse.clone();
	}

	/**
	 * Provides access to the Agents from
	 * <code>agents</code> collection.
	 *
	 * <p> Warning: do not store references to obtained agents in any other
	 * place in the model - always use
	 * <code>World</code> to access them; otherwise undeleted dead agents may
	 * cause memory leaks. Not applicable to Agents that persist through the
	 * whole model execution (LandscapeUnits). </p>
	 *
	 * @param id of the desired Agent
	 * @return Agent from
	 * <code>agents</code> collection specified by its id @verification TEST 1
	 * Warning if the if the Agent ID is below 0 <p> TEST 2 Warning if the
	 * requested Agent is not listed in
	 * <code>agents</code>. </p>
	 * @since 2012-08-07
	 */
	public Agent getAgent(long id) {
		// *****VERIFICATION MODE START*****//
		if (verificationMode) {
			// ***TEST 1
			if (id <= 0) {
				try {
					throw new SkyPigException("Agent ID valueOutOfRange: "
							+ String.valueOf(id));
				} catch (SkyPigException ex) {
					log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
			// ***TEST 2
			if (!agents.containsKey(id)) {
				try {
					throw new SkyPigException(
							"Agent ID not found on the runningAgents list "
							+ String.valueOf(id));
				} catch (SkyPigException ex) {
					log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
		// *****VERIFICATION MODE END*****//
		return agents.get(id);
	}

	/**
	 * Adds a new Agent to
	 * <code>agents</code> collection.
	 *
	 * @param a Agent to be added to runningAgents @verification TEST 1 Warning
	 * if the id of added Agent is already noted in the
	 * <code>agents</code> list.
	 * @since 2012-08-07
	 */
	public void addAgent(Agent a) {
		// *****VERIFICATION MODE START*****//
		if (verificationMode) {
			// ***TEST 1
			if (agents.containsKey(a.getId())) {
				try {
					throw new SkyPigException(
							"Agent id error: adds"
							+ " an Agent thet already exsits in the agents");
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
		// *****VERIFICATION MODE END*****//
		agents.put(a.getId(), a);
		return;
	}

	/**
	 * Adds the passed agent to the
	 * <code>runningAgents</code> collection.
	 *
	 * @param a Agent to be activated
	 * @since 2012-08-07
	 */
	public void activateAgent(Agent a) {
		if (!runningAgents.contains(a)) {
			runningAgents.add(a);
		}
	}

	/**
	 * Initializes the
	 * <code>run</code> method of the passed Agent.
	 *
	 * <p> If the function returns
	 * <code>false</code>, calls
	 * <code>activateAgent</code>. </p>
	 *
	 * @param a Agent to be executed
	 * @since 2012-08-07
	 */
	private void executeAgent(Agent a) {
		if (!a.run()) {
			activateAgent(a);
		}
	}

	/**
	 * Returns running LandscapeUnits that may be stored for reference.
	 *
	 * @return the set of LandscapeUnits.
	 * @since 2012-08-07
	 */
	public Set<LandscapeUnit> getLandscapeUnits() {
		return Collections.unmodifiableSet(landscapeUnits);
	}

	/**
	 * Used together with
	 * <code>skylark</code> package.
	 *
	 * @return
	 * <code>true</code> if it is raining on this day;
	 * <code>false</code> otherwise.
	 * @since 2012-0807
	 */
	public boolean getRain() {
		return rain;
	}

	/**
	 * Used to create the Signature with all the relevant information from the
	 * World that should be stored in database. World adds to the Signature
	 * following values: numberofAgents (number of Agents stored in the
	 * runningAgents Collection), numberOfAnimals (number of Animals stored in
	 * the runningAgents Collection) numberOfNests (number of nests stored in
	 * the runningAgents Collection)
	 *
	 * @return the World's signature
	 */
	@Override
	public Signature getSignature() {
		Signature mySignature = new Signature(this);
		mySignature.addPrimitiveField("numberOfAgents", "integer",
				String.valueOf(agents.size()));
		int[] adultsAndNests = getNumberOfAdultsAndNests();
		mySignature.addPrimitiveField("numberOfAdults", "integer", Integer.toString(adultsAndNests[0]));
		mySignature.addPrimitiveField("numberOfNests", "integer", Integer.toString(adultsAndNests[1]));
		if (((String) Config.getParameter("World_speciesName")).equalsIgnoreCase("skylark")) {
			mySignature.addPrimitiveField("rain", "Boolean",
					Boolean.toString(rain));
		}
		return mySignature;
	}

	/**
	 *
	 * @return
	 */
	public int[] getNumberOfAdultsAndNests() {
		int adults = 0;
		int nests = 0;
		for (Agent a : agents.values()) {
			Class currentClass = a.getClass();
			while (currentClass != null) {
				if (currentClass.getSimpleName().equalsIgnoreCase("Animal")) {
					adults++;
					break;
				} else if (currentClass.getSimpleName().equalsIgnoreCase("Nest")) {
					nests++;
				}
				currentClass = currentClass.getSuperclass();
			}
		}
		int[] array = new int[2];
		array[0] = adults;
		array[1] = nests;
		return array;
	}

	/**
	 * Creates a new
	 * <code>World</code>.
	 *
	 * <p> Creates a new random number generator. Initializes logging the
	 * information to log file, named: startTime_configurationNumber.log.
	 * Initializes the current date and sets the date of the end of the model
	 * (apocalypse). Initializes empty lists of Agents: runningAgents and
	 * landscapeUnits. Creates Observer. Initializes speciesFactory according to
	 * the input species name. Sets rain to false. </p> @parameters
	 * configurationName, World_randomSeed, World_startDate, World_length,
	 * World_dataBaseName, World_speciesName
	 *
	 * @since 2012-08-07
	 */
	private World(Collection<Observer> observers) {
		if (logger != null) {
			try {
				String configurationFilename = (String) Config.getParameter("configurationFile");
				if (configurationFilename.contains("/")) {
					configurationFilename = configurationFilename.substring(configurationFilename.lastIndexOf("/") + 1);
				} else if (configurationFilename.contains("\\")) {
					configurationFilename = configurationFilename.substring(configurationFilename.lastIndexOf("\\") + 1);
				}
				logger.addHandler(new FileHandler(Config.getParameter("World_speciesName")
						+ "_" + Config.getParameter("startTime").toString()
						+ "_" + configurationFilename
						+ "_" + Config.getParameter("configurationName") + ".log"));
			} catch (IOException e) {
				logger.addHandler(new ConsoleHandler());
			}
			for (Handler h : logger.getHandlers()) {
				h.setFormatter(new SimpleFormatter());
			}
			logger.setLevel(Level.WARNING);
		}
		runningAgents = new LinkedList<Agent>();
		date = (Calendar) ((GregorianCalendar) Config.getParameter("World_startDate")).clone();
		apocalypse = (Calendar) date.clone();
		apocalypse.add(Calendar.YEAR,
				(Integer) Config.getParameter("World_length"));
		agents = new ConcurrentHashMap<Long, Agent>();
		landscapeUnits = new HashSet<LandscapeUnit>();
		this.observers = new HashSet<Observer>(observers);
		rain = false;
		if (((String) Config.getParameter("World_speciesName")).equalsIgnoreCase("skylark")) {
			speciesFactory = new SkylarkFactory();
		} else {
			if (((String) Config.getParameter("World_speciesName")).equalsIgnoreCase("woodpigeon")) {
				speciesFactory = new WoodpigeonFactory();
			} else {
				System.out.print("Error: Species not supported yet.");
				speciesFactory = null;
			}
		}
	}

	/**
	 * Calls the observer to observe (store results) for each Agent and the
	 * World.
	 *
	 * @since 2012-08-07
	 */
	private void storeResults() {
		Collection<Observable> observed = new HashSet<Observable>();
		observed.add(this);
		observed.addAll(agents.values());
		for (Observer observer : observers) {
			observer.observe(observed);
		}
	}

	/**
	 * Generates weather for the next day.
	 *
	 * <p> Used in skylark model. Sets the rain variable in the basis of its
	 * current value and probability defined in the configuration file. The
	 * probability of weather change may be different rain = true and rain =
	 * false. </p>
	 *
	 * @parameters World_rain_probability_after_rain, World_speciesName
	 * @verification TEST 1 Warning if called in other then skylark models
	 *
	 * @since 2012-08-07
	 */
	private void selectWeather() {
		if (rain) {
			if (Config.getGenerator().nextDouble() < (Double) Config.getParameter("World_rain_probability_after_rain")) {
				rain = true;
			} else {
				rain = false;
			}
		} else {
			if (Config.getGenerator().nextDouble() < (Double) Config.getParameter("World_rain_probability_after_sun")) {
				rain = true;
			} else {
				rain = false;
			}
		}
		// *****VERIFICATION MODE START*****//
		if (verificationMode) {
			// ***TEST 1
			if (!((String) Config.getParameter("World_speciesName")).equalsIgnoreCase("skylark")) {
				try {
					throw new SkyPigException(
							"Wheater used in new species: "
							+ ((String) Config.getParameter("World_speciesName")));
				} catch (SkyPigException ex) {
					log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
		// *****VERIFICATION MODE END*****//
	}

	/**
	 * Removes dead Agents (Animals and Nests) from the
	 * <code>agents</code>.
	 *
	 * @since 2012-08-07
	 */
	private void removeDeadAgents() {
		Iterator<Map.Entry<Long, Agent>> iterator = agents.entrySet().iterator();
		while (iterator.hasNext()) {
			Map.Entry<Long, Agent> id = iterator.next();
			if (!id.getValue().getAlive()) {
				iterator.remove();
			}
		}
	}

	/**
	 * Creates Agents.
	 *
	 * <p> Expects the following section names: CropRotations, LandscapeUnits,
	 * BorderDistances, BorderLengths, Territories, Animals. Other section names
	 * are unrecognised and produce errors, but not all of the above are
	 * obligatory. </p>
	 *
	 * @param filename of initialisation file with values for creating objects,
	 * including Agents.
	 * @since 2012-08-07.
	 */
	private void initialise(String filename) {
		BufferedReader in = null;
		try {
			in = new BufferedReader(new FileReader(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.out.println("Initialisation file not found. Aborting reading.");
			return;
		}
		String line = null;
		Scanner myScanner = null;
		String word = null;
		String currentSection = null;
		ArrayList<Long> landscapeCentreIds = null;
		ArrayList<Long> landscapeBorderIds = null;
		ArrayList<Long> landscapeBorderLengthIds = null;
		HashMap<String, CropRotation> cropRotationMap = new HashMap<String, CropRotation>();
		try {
			while ((line = in.readLine()) != null) {
				myScanner = new Scanner(line);
				if (myScanner.hasNext()) {
					word = myScanner.next();
					if (!word.startsWith("#")) {
						if (word.startsWith(":")) {
							word = word.replaceFirst(":", "");
							currentSection = word;
							continue;
						} else if (currentSection.compareToIgnoreCase("CropRotations") == 0) {
							String name = word;
							ArrayList<String> list = new ArrayList<String>();
							while (myScanner.hasNext()) {
								list.add(myScanner.next());
							}
							cropRotationMap.put(name, new CropRotation(name,
									list));
						} else if (currentSection.compareToIgnoreCase("LandscapeUnits") == 0) {
							CropRotation.initialise(cropRotationMap);
							LandscapeUnit lu = speciesFactory.createLandscapeUnit((Config.lineToMap(
									word, myScanner)));
							lu.launchMe();
							landscapeUnits.add(lu);
						} else if (currentSection.compareToIgnoreCase("BorderDistances") == 0) {
							if (landscapeBorderIds == null) {
								landscapeBorderIds = new ArrayList<Long>();
								landscapeBorderIds.add(Long.valueOf(word));
								while (myScanner.hasNextLong()) {
									landscapeBorderIds.add(myScanner.nextLong());
								}
							} else {
								LandscapeUnit land = (LandscapeUnit) agents.get(Long.valueOf(word));
								TreeMap<Double, ArrayList<LandscapeUnit>> borderDistances = new TreeMap<Double, ArrayList<LandscapeUnit>>();
								int i = 0;
								while (myScanner.hasNextDouble()) {
									double key = Math.round(myScanner.nextDouble());
									if (!borderDistances.containsKey(key)) {
										ArrayList<LandscapeUnit> lands = new ArrayList<LandscapeUnit>();
										lands.add((LandscapeUnit) agents.get(landscapeBorderIds.get(i)));
										borderDistances.put(key, lands);
									} else {
										borderDistances.get(key).add(
												(LandscapeUnit) agents.get(landscapeBorderIds.get(i)));
									}
									i++;
								}
								((LandscapeUnit) land).setBorderDistances(borderDistances);
							}
						} else if (currentSection.compareToIgnoreCase("BorderLengths") == 0) {
							if (landscapeBorderLengthIds == null) {
								landscapeBorderLengthIds = new ArrayList<Long>();
								landscapeBorderLengthIds.add(Long.valueOf(word));
								while (myScanner.hasNextLong()) {
									landscapeBorderLengthIds.add(myScanner.nextLong());
								}
							} else {
								LandscapeUnit land = (LandscapeUnit) agents.get(Long.valueOf(word));
								TreeMap<Double, ArrayList<LandscapeUnit>> borderLengths = new TreeMap<Double, ArrayList<LandscapeUnit>>();
								int i = 0;
								while (myScanner.hasNextDouble()) {
									double key = Math.round(myScanner.nextDouble());
									if (!borderLengths.containsKey(key)) {
										ArrayList<LandscapeUnit> lands = new ArrayList<LandscapeUnit>();
										lands.add((LandscapeUnit) agents.get(landscapeBorderLengthIds.get(i)));
										borderLengths.put(key, lands);
									} else {
										borderLengths.get(key).add((LandscapeUnit) agents.get(landscapeBorderLengthIds.get(i)));
									}
									i++;
								}
								((LandscapeUnit) land).setBorderLengths(borderLengths);
							}
						} else if (currentSection.compareToIgnoreCase("Territories") == 0) {
							HashMap<String, Object> parameters = new HashMap<String, Object>();
							long landscapeID = Long.valueOf(word);
							SkylarkLandscapeUnit slu = null;
							for (LandscapeUnit l : getLandscapeUnits()) {
								if (l.getId() == landscapeID) {
									slu = (SkylarkLandscapeUnit) l;
									break;
								}
							}
							parameters.put("id", myScanner.nextInt());
							parameters.put("x", myScanner.nextDouble());
							parameters.put("y", myScanner.nextDouble());
							parameters.put("radiusesAndAreas",
									new HashMap<Double, HashMap<SkylarkLandscapeUnit, Double>>());
							while (myScanner.hasNextDouble()) {
								double r = myScanner.nextDouble();
								HashMap<SkylarkLandscapeUnit, Double> LUAreas = new HashMap<SkylarkLandscapeUnit, Double>();
								while (myScanner.hasNextLong()) {
									long luid = myScanner.nextLong();
									double area = myScanner.nextDouble();
									for (LandscapeUnit lu : getLandscapeUnits()) {
										if (lu.getId() == luid) {
											LUAreas.put(
													(SkylarkLandscapeUnit) lu,
													area);
											break;
										}
									}
								}
								((HashMap<Double, HashMap<SkylarkLandscapeUnit, Double>>) parameters.get("radiusesAndAreas")).put(r,
										LUAreas);
							}
							slu.addTerritory(parameters);
						} else if (currentSection.compareToIgnoreCase("Animals") == 0) {
							speciesFactory.createAnimal(
									(Config.lineToMap(word, myScanner))).launchMe();
						} else if (currentSection.compareToIgnoreCase("AnimalsRandomised") == 0) {
							for (int i=0; i< (Integer) Config.getParameter("World_numberOfAnimals"); i++) {
								HashMap<String, Object> parameters =  new HashMap<String, Object>();
								parameters.put("at_initialization", true);
								parameters.put("bodyweight", null);
								parameters.put("sex", null);
								speciesFactory.createAnimal(parameters).launchMe();
							}
						} else {
							System.out.println("Unrecognised line. Out of section or wrong section name: "
									+ currentSection);
							continue;
						}
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("IOException during reading of initialisation file. Aborting.");
		}
	}
}
