//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
package model.woodpigeon;

import java.util.*;
import model.core.*;
import java.util.logging.Level;

/**
 * Represents the Animal from woodpigeon species (Columba palumbus).
 *
 * <p> Implements the details of foragingn behaviour that is species specific
 * and behaviour typical for seasonal activities of the species. <p>
 *
 * @author Katarzyna Kułakowska
 */
public class WoodpigeonAnimal extends Animal {

	/**
	 * An AgentState for adult woodpigeons.
	 *
	 * <p> Implements detailed behaviour of the species. Used immediately after
	 * leaving the nest, also by juveniles and first-year adults. </p>
	 *
	 * @author Katarzyna Kułakowska
	 */
	public class WoodpigeonAdult extends Adult {

		/*
		 * Fields:
		 */
		/**
		 * Time slot for a single foraging attempt specified for this date. In
		 * winter it is usually the whole day, while in breeding season when
		 * birds return often to their nest to feed nestlings it is shorter.
		 */
		public int timeIntervalForForagingToday;
		/**
		 *
		 */
		public int totalTimeForForagingToday;
		/**
		 *
		 */
		public TimetableSlot[] feedingTimetable;
		/**
		 * The number of TimetableSlots available for today;
		 */
		public int photoperiod;
		/**
		 *
		 */
		public int timeLeftForForagingToday;
		/**
		 *
		 */
		public double cropContent;
		/**
		 *
		 */
		public int fieldsRevisited;

		/*
		 * Methods:
		 */
		/**
		 * Constructor for WoodpigeonAdult does not require any parameters in
		 * the HashMap.
		 *
		 * <p> Values of WoodpigeonAdult variables are set as follows:
		 * timeIntervalForForagingToday = daylength, currentStartTime = 0,
		 * totalForagingTime = 0, feedingPlaces = empty Set. Memory is created.
		 * </p>
		 *
		 * @param parameters that can be used for parameterization if necessary.
		 */
		public WoodpigeonAdult(HashMap<String, Object> parameters) {
			super(parameters);
//			feedingTimetable = new TreeMap<Integer, TimetableSlot>();
//			feedingTimetable.put(0, new TimetableSlot(false));
			feedingTimetable = new TimetableSlot[Config.getLongestDayLength()];
			resetTimeForForagingToday();
			resetFeedingTimetable();
			succesfulFeedingRoutine = new ArrayList<Boolean>();
			for (int i = 1; i <= (Integer) Config.getParameter("WoodpigeonAdult_feedingRoutineLength"); i++) {
				succesfulFeedingRoutine.add(false);
			}
			createMemory();
		}

		/**
		 *
		 * @param r
		 * @param LU
		 * @return
		 */
		public int calculateCropFillingTime(Resource r, LandscapeUnit LU) {
			if (r == null) {
				return 0;
//			} else if (timeLeftForForagingToday == totalTimeForForagingToday) {
			} else if (cropContent < (Double) Config.getParameter("WoodpigeonAdult_cropVolume")) {
				double ingRate = calculateMaxIngestionRate(r, LU);
				double digRate = calculateDigestionRate(r);
				if (ingRate > digRate) {
					return (int) (Math.round((((Double) Config.getParameter("WoodpigeonAdult_cropVolume") - cropContent)
							/ (ingRate - digRate)) / (((Integer) Config.getParameter("Config_timeFactor")).doubleValue() / 60.0)));
				} else {
					return 0;
				}
			} else {
				return 0;
			}
		}

		/**
		 *
		 * @param r
		 * @param LU
		 * @param cropFillingTime
		 * @return
		 */
		public double getEnergyIntakeRate(Resource r, LandscapeUnit LU, int cropFillingTime) {
			if (r == null) {
				return 0;
			} else {
				double ingRate = calculateMaxIngestionRate(r, LU);
				double digRate = calculateDigestionRate(r);
				if (cropFillingTime > 0 || digRate > ingRate) {
					return ingRate * r.energyContent * (Integer) Config.getParameter("Config_timeFactor");
				} else {
					return digRate * r.energyContent * (Integer) Config.getParameter("Config_timeFactor");
				}
			}
		}

		/**
		 * Returns the energy in g digested per 1min.
		 *
		 * @param r
		 * @return
		 */
		public double calculateDigestionRate(Resource r) {
			double digestionRate = (Double) Config.getParameter(r.getName()
					+ "_digestionRate") / 60.0;
			return digestionRate;
		}

		/**
		 * Returns the energy in g ingested per 1 min.
		 *
		 * @param r
		 * @param LU
		 * @return
		 */
		public double calculateMaxIngestionRate(Resource r, LandscapeUnit LU) {
			double ingRate;
			double pacesBasedIngestionRate = (r.getDensity(LU.getSize()))
					* (Double) Config.getParameter(getClass().getSimpleName()
					+ "_feedingFraction")
					* (Double) Config.getParameter(getClass().getSimpleName()
					+ "_feedingSpeed") / 10000.0;
			double pecksBasedIngestionRate = (Double) Config.getParameter(r.getName() + "_maxPeckingRate")
					* (Double) Config.getParameter(r.getName() + "_peckSize");
			if (pacesBasedIngestionRate < pecksBasedIngestionRate) {
				ingRate = pacesBasedIngestionRate;
			} else {
				ingRate = pecksBasedIngestionRate;
			}
			if (getBasalEnergy() < getMaxBasalEnergy()
					|| getDaysSinceHatching() < (Integer) Config.getParameter("WoodpigeonAdult_startOfAdulthood")) {
				ingRate = ingRate * (Double) Config.getParameter("WoodpigeonAdult_ingestionRateLimitation");
			}
			return ingRate;
		}

		/**
		 * Creates memory of WoodpigeonAdult comprising all the LandscapeUnits
		 * currently used in the model with their best resources. <p> the
		 * Memorable objects are created with current date. </p>
		 *
		 * @verification TEST 1 Warning of not all the available landscapeUnits
		 * were added to memory.
		 */
		private void createMemory() {
			if (memory == null) {
				memory = new TreeMap<Double, Set<Memorable>>();
				for (LandscapeUnit lu : World.Instance().getLandscapeUnits()) {
					double energyAsRate = getEnergyIntakeRate(chooseBestResource(lu), lu, 0);
					if (!memory.containsKey(energyAsRate)) {
						Set<Memorable> mem = Collections.synchronizedSet(new HashSet<Memorable>());
						mem.add(new Memorable(lu, World.getDate()));
						memory.put(energyAsRate, mem);
					} else {
						(memory.get(energyAsRate)).add(new Memorable(lu, World.getDate()));
					}
				}
//*****VERIFICATION MODE START*****//
				if (World.verificationMode) {
					//***TEST 1
					int memorySize = 0;
					for (Double key : memory.keySet()) {
						memorySize += memory.get(key).size();
					}
					if (memorySize != World.Instance().getLandscapeUnits().size()) {
						try {
							throw new SkyPigException("Memory not initialized properly: "
									+ String.valueOf(memorySize) + " instead of "
									+ String.valueOf(World.Instance().getLandscapeUnits().size())
									+ "landscapeUnits added.");
						} catch (SkyPigException ex) {
							World.log(Level.WARNING, Config.getLogInfo(), ex);
						}
					}
				}
//*****VERIFICATION MODE END*****//
			}
		}

		/**
		 * Adds the passed LandscapeUnit to memory.
		 *
		 * <p> First the LandscapeUnit is removed from the memory of it is
		 * there. Next it is added with an up to date energy assimilation rate
		 * calculated for the best resource on this LandscapeUnit, and the date
		 * in Memorized is set according to World.getDate(). </p>
		 *
		 * @param x LandscapeUnit to be added to memory. @verification TEST 1
		 * Warning if the landscapeUnit that is added to is not contained there
		 * already, or if it was, it has been properly removed.
		 */
		private void addToMemory(LandscapeUnit x, double energyAssimilationRate) {
//			synchronized (memory) {
			Iterator<Double> keyIterator = memory.keySet().iterator();
			while (keyIterator.hasNext()) {
				Set<Memorable> memorablesAtKey = memory.get(keyIterator.next());
				Iterator<Memorable> valueIterator = memorablesAtKey.iterator();
				boolean found = false;
				while (valueIterator.hasNext()) {
					Memorable inMemory = valueIterator.next();
					if (inMemory.lu.equals(x)) {
						valueIterator.remove();
						found = true;
						if (World.calculateDateDifference(inMemory.dateMemorized) == 1) {
							if (energyAssimilationRate > 0) {
								fieldsRevisited++;
							}
						}
						break;
					} else {
//*****VERIFICATION MODE START*****//
						if (World.verificationMode) {
							//***TEST 1
							if (inMemory.lu.getId() == x.getId()) {
								try {
									throw new SkyPigException("LandscapeUnit not properly removed from memory.");
								} catch (SkyPigException ex) {
									World.log(Level.WARNING, Config.getLogInfo(), ex);
								}
							}
						}
//*****VERIFICATION MODE END*****//
					}
				}
				if (memorablesAtKey.isEmpty()) {
					keyIterator.remove();
				}
				if (found) {
					break;
				}
			}
			if (!memory.containsKey(energyAssimilationRate)) {
				Set<Memorable> mem = Collections.synchronizedSet(new HashSet<Memorable>());
				mem.add(new Memorable(x, World.getDate()));
				memory.put(energyAssimilationRate, mem);
			} else {
				(memory.get(energyAssimilationRate)).add(new Memorable(x, World.getDate()));
			}
//			}
		}

		@Override
		public boolean checkIfShot() {
			if ((Boolean) Config.getParameter("WoodpigeonAnimal_shooting")) {
				ArrayList<Integer> years = (ArrayList<Integer>) Config.getParameter("WoodpigeonAnimal_shootingYears");
				for (Integer year : years) {
					if (World.getDate().get(Calendar.YEAR) == year
							&& World.getDate().get(Calendar.DAY_OF_YEAR)
							>= (Integer) Config.getParameter("WoodpigeonAnimal_shootingStartDay")
							&& World.getDate().get(Calendar.DAY_OF_YEAR)
							<= (Integer) Config.getParameter("WoodpigeonAnimal_shootingEndDay")) {
						int yearIndex = years.indexOf(year);
						if (Config.getGenerator().nextDouble()
								< ((ArrayList<Double>) Config.getParameter("WoodpigeonAnimal_shootingProbability")).get(yearIndex)) {
							return true;
						}
					}
				}
			}
			return false;
		}

		/**
		 * Resets currentStartTime to 0, totalFragingTime to 0,
		 * timeIntervalForForagingToday is reset using method
		 * <code>resetTimeForForagingToday()</code> and memory is cleared from
		 * memories older then specified.
		 *
		 * @parameters WoodpigeonAdult_memoryLength
		 */
		@Override
		public void reset() {
			resetTimeForForagingToday();
			resetFeedingTimetable();
			resetMemory();
			cropContent = 0;
			fieldsRevisited = 0;
			super.reset();
		}

		private void resetMemory() {
//			synchronized (memory) {
			Iterator<Double> keyIterator = memory.keySet().iterator();
			while (keyIterator.hasNext()) {
				Set<Memorable> memorablesAtKey = memory.get(keyIterator.next());
				Iterator<Memorable> valueIterator = memorablesAtKey.iterator();
				while (valueIterator.hasNext()) {
					if (World.calculateDateDifference(valueIterator.next().dateMemorized) > (Integer) Config.getParameter("WoodpigeonAdult_memoryLength")) {
						valueIterator.remove();
					}
				}
				if (memorablesAtKey.isEmpty()) {
					keyIterator.remove();
				}
			}
//			}
		}

		/**
		 * Resets the feedingTimetable adjusting it to the current photoperiod,
		 * clearing the foraging places in timeslots and adjusting availability
		 * according to sex and current seasonal activities. <p> If any time is
		 * unavailable for feeding for females and males while looking for
		 * territories, it is planned in two periods in the morning and in the
		 * evening. If any time is unavailable for males not looking for
		 * territories, it is set in the middle of the day (according to
		 * incubation routine). Should be executed after
		 * <code>resetTimeForForagingToday</code>. </p>
		 *
		 * @verification TEST 1 checks if the length of feeding timetable equals
		 * to photoperiod
		 */
		private void resetFeedingTimetable() {
			photoperiod = Config.getDayLength(World.getDate());
			for (int i = 0; i < feedingTimetable.length; i++) {
				if (i < photoperiod) {
					if (feedingTimetable[i] == null) {
						feedingTimetable[i] = new TimetableSlot(false);
					}
				} else {
					feedingTimetable[i] = null;
				}
			}
			int timeUnavailable = photoperiod - totalTimeForForagingToday;
			if ((getSex() == Sex.FEMALE)
					|| getActivityName().equalsIgnoreCase("WoodpigeonFindTerritory")) {
				for (int i = 0; i < Math.floor(timeUnavailable / 2); i++) {
					feedingTimetable[i].available = false;
					feedingTimetable[i].foragingPlace = -1;
				}
				for (int i = (int) (Math.floor(timeUnavailable / 2)); i < Math.floor(timeUnavailable / 2) + totalTimeForForagingToday; i++) {
					feedingTimetable[i].available = true;
					feedingTimetable[i].foragingPlace = -1;
				}
				for (int i = (int) (Math.floor(timeUnavailable / 2) + totalTimeForForagingToday); i < photoperiod; i++) {
					feedingTimetable[i].available = false;
					feedingTimetable[i].foragingPlace = -1;
				}
			} else {
				for (int i = 0; i < (int) (Math.floor(totalTimeForForagingToday / 2)); i++) {
					feedingTimetable[i].available = true;
					feedingTimetable[i].foragingPlace = -1;
				}
				for (int i = (int) (Math.floor(totalTimeForForagingToday / 2)); i < (int) (Math.floor(totalTimeForForagingToday / 2)) + timeUnavailable; i++) {
					feedingTimetable[i].available = false;
					feedingTimetable[i].foragingPlace = -1;
				}
				for (int i = (int) (Math.floor(totalTimeForForagingToday / 2)) + timeUnavailable; i < photoperiod; i++) {
					feedingTimetable[i].available = true;
					feedingTimetable[i].foragingPlace = -1;
				}
			}
		}

		/**
		 * Resets the timeIntervalForForagingToday and totalTimeForForagingtoday
		 * according to season and sex.
		 *
		 * @parameters WoodpigeonAdult_timeForTerritory, Config_timeFactor,
		 * WoodpigeonAdult_ContinuousBroodingEnd, WoodpigeonAdult_BroodingEnd
		 *
		 * @verification TEST 1 Warning if the totalTimeForForagingToday exceeds
		 * day length.
		 */
		private void resetTimeForForagingToday() {
			if (getActivityName().equalsIgnoreCase("WoodpigeonFindTerritory")
					|| getActivityName().equalsIgnoreCase("WoodpigeonFindPartner")) {
				timeIntervalForForagingToday = (int) Math.round(Double.valueOf(Config.getDayLength(World.getDate()))
						- ((Integer) Config.getParameter("WoodpigeonAdult_timeForTerritory")).doubleValue() * 60.0
						/ ((Integer) Config.getParameter("Config_timeFactor")).doubleValue());
				totalTimeForForagingToday = timeIntervalForForagingToday;
			} else if ((getNest() != null
					&& ((((Nest) World.Instance().getAgent(getNest())).getDaysSinceHatching())
					<= (Integer) Config.getParameter("WoodpigeonAdult_continuousBroodingEnd")))
					|| ((getPartner() != null && getNest() == null))) {
				totalTimeForForagingToday = Math.round(Config.getDayLength(World.getDate()) / 2);
				if (getNest() != null && ((Nest) World.Instance().getAgent(getNest())).getDaysSinceHatching() <= 0
						|| getNest() == null) {
					if (getSex() == Sex.FEMALE) {
						timeIntervalForForagingToday = totalTimeForForagingToday;
					} else {
						timeIntervalForForagingToday = totalTimeForForagingToday / 2;
					}
				} else {
					timeIntervalForForagingToday = 
							(int) Math.round(((Integer) Config.getParameter("WoodpigeonAdult_broodingForagingTime")).doubleValue() * 60.0
							/ ((Integer) Config.getParameter("Config_timeFactor")).doubleValue());
				}
			} else if ((getNest() != null)
					&& ((Nest) World.Instance().getAgent(getNest())).getDaysSinceHatching()
					> (Integer) Config.getParameter("WoodpigeonAdult_continuousBroodingEnd")) {
				if (((Nest) World.Instance().getAgent(getNest())).getDaysSinceHatching()
						< (Integer) Config.getParameter("WoodpigeonAdult_broodingEnd")) {
					double broodingForagingTime = (int) Math.round(((Integer) Config.getParameter("WoodpigeonAdult_broodingForagingTime")).doubleValue() * 60.0
							/ ((Integer) Config.getParameter("Config_timeFactor")).doubleValue());
					double continuousBroodingEnd = (Integer) Config.getParameter("WoodpigeonAdult_continuousBroodingEnd");
					timeIntervalForForagingToday = (int) (broodingForagingTime
							+ ((Math.round(Config.getDayLength(World.getDate()) / 3) - broodingForagingTime)
							/ ((((Nest) World.Instance().getAgent(getNest())).getDaysSinceHatching()
							- continuousBroodingEnd)
							/ (((Integer) Config.getParameter("WoodpigeonAdult_broodingEnd")).doubleValue() - continuousBroodingEnd))));
					totalTimeForForagingToday = Config.getDayLength(World.getDate());
				} else {
					timeIntervalForForagingToday = Math.round(Config.getDayLength(World.getDate()) / 3);
					totalTimeForForagingToday = Config.getDayLength(World.getDate());
				}
			} else {
				timeIntervalForForagingToday = Config.getDayLength(World.getDate());
				totalTimeForForagingToday = timeIntervalForForagingToday;
			}
//*****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				//***TEST 1
				if (totalTimeForForagingToday > Config.getDayLength(World.getDate())) {
					try {
						throw new SkyPigException("totalTimeForForagingToday length error: "
								+ String.valueOf(totalTimeForForagingToday)
								+ ", photoperiod: "
								+ String.valueOf(Config.getDayLength(World.getDate())));
					} catch (SkyPigException ex) {
						World.log(Level.WARNING, Config.getLogInfo(), ex);
					}
				}
			}
//*****VERIFICATION MODE END*****//
			timeLeftForForagingToday = totalTimeForForagingToday;
		}

		/**
		 * Specifies the foraging algorithm for pigeons. <p> This methods
		 * operates in a loop assumming the bird has still time for foraging and
		 * its food requirements are not fulfilled. The function chooseBestLU is
		 * executed, followed by chooseFieldsBetween. From the set of fields
		 * between the birds' current location and field choosen from memory the
		 * fields directly flown over are chosen in the method
		 * chooseFieldsOnTheWay. This foraging time interval is calculated in
		 * the form of hours from timeTable available for this feeding attempt.
		 * Afterwards the inner loop is executed: As long as the foraging
		 * interval has not expired and the food requirements are > 0, the bird
		 * chooses the foraging places from the previously choosen set or the
		 * field chosen from the memory, if the set is empty. Another condition
		 * for this loop is that the energy assimilation rate on the previous
		 * field was unsatisfactory and the real foraging time was therefore 1
		 * time unit. The food is gathered from the chosen field and the
		 * remaining time in this foraging interval is updated. After
		 * satisfactory foraging if the conditions are met the outer loop is
		 * executed. </p> <p> If certain conditions are met, additional feeding
		 * time may be available: For incubating birds or birds in
		 * pre-incubation phase, if their food requirements were not fulfilled
		 * and the time for foraging expired, they spend additional time on
		 * foraging instead of incubation. This may affect eggs. The choice of
		 * feeding grounds during this time is made according to the above
		 * algorithm. For birds preparing for reproduction, searching for
		 * territories or partner and defending territories, they may spend
		 * additional time of foraging in the wood in which their territory is
		 * located or where their roosting place is located if they have no
		 * territory. </p> <p> After foraging the food is shared between the
		 * bird and its offspring and consumed (shareFood). </p>
		 */
		@Override
		public void findFood() {
			boolean alternativeFeeding = true;
			if (!checkFeedingEffect()) {
				while (timeLeftForForagingToday > 0 && getFoodRequirements() > 0) {
					alternativeFeeding = false;
					goForForaging(alternativeFeeding);
				}
				while (hasAlternativeForagingTime()
						&& (getAnimalEnergyRequirements() + nestRequirements - getFoodRequirements() < calculateMinimumEnergyRequired())) {
					alternativeFeeding = true;
					goForForaging(alternativeFeeding);
				}
			}
			succesfulFeedingRoutine.remove(0);
			if (alternativeFeeding) {
				succesfulFeedingRoutine.add(false);
			} else {
				succesfulFeedingRoutine.add(true);
			}
			if (myTerritory != null) {
				location = myTerritory;
			} else {
				location = roostingPlace;
			}
//			chooseResourcesForNestlings();
			eat();
		}

		/**
		 *
		 * @return
		 */
		public TreeMap<Double, Set<Memorable>> getMemory() {
			return memory;
		}

		private void goForForaging(boolean alternativeFeeding) {
			LandscapeUnit x = WoodpigeonForaging.chooseBestLU(getCurrentEnergyRequirements(alternativeFeeding), this);
//			LandscapeUnit x = chooseBestLU(getCurrentEnergyRequirements(alternativeFeeding));
			LinkedHashSet<Integer> thisForagingInterval = calculateThisForagingInterval(alternativeFeeding);
			Set<Integer> realForagingInterval = new HashSet<Integer>();
			if ((Boolean) Config.getParameter("WoodpigeonAdult_flocking") == true) {
				LandscapeUnit[] fields = chooseFieldsBetween(x);
				HashSet<LandscapeUnit> set = chooseFieldsOnTheWay(location.calculateCentralDistance(x), fields);
				Iterator<LandscapeUnit> setIterator = set.iterator();
				boolean allFieldsChecked = false;
				while (thisForagingInterval.size() > 0 && getCurrentEnergyRequirements(alternativeFeeding) > 0
						&& realForagingInterval.size() <= 1 && !allFieldsChecked) {
					LandscapeUnit foragingPlace = null;
					if (setIterator.hasNext()) {
						foragingPlace = checkFlockSize((WoodpigeonLandscapeUnit) setIterator.next(), thisForagingInterval.iterator().next());
					}
					if (foragingPlace == null && !setIterator.hasNext()) {
						foragingPlace = x;
						allFieldsChecked = true;
					}
					if (foragingPlace != null) {
						realForagingInterval = gatherFood(foragingPlace, thisForagingInterval, getCurrentEnergyRequirements(alternativeFeeding));
						thisForagingInterval.removeAll(realForagingInterval); //check if necessary
						location = foragingPlace;
					}
				}
			} else {
				realForagingInterval = gatherFood(x, thisForagingInterval, getCurrentEnergyRequirements(alternativeFeeding));
				thisForagingInterval.removeAll(realForagingInterval); //check if necessary
				location = x;
			}


		}

		private double getCurrentEnergyRequirements(boolean alternativeFeeding) {
			if (alternativeFeeding) {
				return calculateMinimumEnergyRequired()
						- (getAnimalEnergyRequirements() + nestRequirements - getFoodRequirements());
			} else {
				return getFoodRequirements();
			}
		}

		/**
		 * On the basis of feedingTimetable returns the set of hours available
		 * for this feeding attempt.
		 *
		 * <p> The set contains the time slots (only hours) available for
		 * foraging during this foraging interval. If alternativeFeeding is
		 * true, the set contains the time slots (only hours) normally
		 * unavailable for foraging, but exceptionally available during this
		 * foraging interval. </p>
		 *
		 * @return set of time units available for foraging in the current time
		 * interval.
		 */
		private LinkedHashSet<Integer> calculateThisForagingInterval(boolean alternativeFeeding) {
			LinkedHashSet<Integer> thisForagingInterval = new LinkedHashSet<Integer>();
			ArrayList<Integer> list = new ArrayList<Integer>();
			for (int i = 0; i < photoperiod; i++) {
				if ((feedingTimetable[i].available || alternativeFeeding)
						&& feedingTimetable[i].foragingPlace == -1) {
					list.add(i);
				}
			}
			int random = Config.getGenerator().nextInt(list.size());
			for (int i = random; i < random + timeIntervalForForagingToday; i++) {
				if (i < list.size()) {
					thisForagingInterval.add(list.get(i));
				} else {
					thisForagingInterval.add(list.get(i % list.size()));
				}
			}
			return thisForagingInterval;
		}

		/**
		 * Returns true if there are nya unused timeSlots in the
		 * feedingTimetable.
		 *
		 * @return true if there are unused timeslots, false otherwise.
		 */
		private boolean hasAlternativeForagingTime() {
			for (int i = 0; i < photoperiod; i++) {
				if (feedingTimetable[i].foragingPlace == -1) {
					return true;
				}
			}
			return false;
		}

		/**
		 * Calls the method of a superclass and additionally accepts the food
		 * requirments of nest.
		 */
		@Override
		public void manageEnergy() {
			super.manageEnergy();
//			acceptNestFoodRequirements();
			addNestFoodRequirements();
		}

		/**
		 * Reads the passed Message, takes the proper action in response and
		 * replies to the Message.
		 *
		 * <p> Topics of Messages read: </p> <p> "incubation" - returns the
		 * message with topic "re:incubation" and the number of timeslots spent
		 * on incubation today (based on the feedingTimetable). </p> <p> If a
		 * Message with any other topic is received, it is forwarded to
		 * super.readMessage. </p>
		 *
		 * @param m Message to be read
		 * @return Message to be send as reply or null if the passed Message
		 * does not require an answer.
		 */
		@Override
		public Message readMessage(Message m) {
			if (m.topic.equalsIgnoreCase("incubation")) {
				if (incubationRequired) {
					incubationForNest = photoperiod - totalTimeForForagingToday + timeLeftForForagingToday;
				}
				return new Message(getId(), m.senderId, "re:incubation", m, incubationForNest);
			}
			if (m.topic.equalsIgnoreCase("partner")) {
				if (checkFeedingRoutine()) {
					return super.readMessage(m);
				} else {
					return new Message(getId(), m.senderId, "re:partner", m,
							null);
				}
			} else {
				return super.readMessage(m);
			}
		}

		/**
		 * Chooses the Resource on the given LandscapeUnit that give the highest
		 * energy intake rate at a given time.
		 *
		 * @param foragingPlace LandcsapeUnit to choose the best Resource from.
		 * @return Resource of the highest energy intake rate.
		 */
		public Resource chooseBestResource(LandscapeUnit foragingPlace) {
			Set<Resource> resources = foragingPlace.getResources();
			Resource maxResource = null;
			for (Resource r : resources) {
				if (maxResource == null) {
					maxResource = r;
				} else {
					int cropFilling = 0;
					if (cropContent < (Double) Config.getParameter("WoodpigeonAdult_cropVolume")) {
						cropFilling = 1;
					}
					if (getEnergyIntakeRate(r, foragingPlace, cropFilling) > getEnergyIntakeRate(maxResource, foragingPlace, cropFilling)) {
						maxResource = r;
					}
				}
			}
			return maxResource;
		}

		/**
		 *
		 * @param x
		 * @param thisForagingInterval
		 * @param requirements
		 * @return
		 */
		public TreeSet<Integer> gatherFood(LandscapeUnit x, LinkedHashSet<Integer> thisForagingInterval, double requirements) {
			TreeSet<Integer> realForagingInterval = new TreeSet<Integer>();
			Resource obtainedResource = null;
			Resource maxResource = chooseBestResource(x);
			int cropFillingTime = calculateCropFillingTime(maxResource, x);
			double energyAssimilationRate = getEnergyIntakeRate(maxResource, x, cropFillingTime);
			addToMemory(x, energyAssimilationRate);
			if (energyAssimilationRate <= 0) {
				return realForagingInterval;
			}
			int desiredTimeForaging = 1;
			if (WoodpigeonForaging.achieveOptimalEnergyAssimilationRate(energyAssimilationRate, requirements, this)) {
				desiredTimeForaging = (int) Math.ceil(requirements
						/ (energyAssimilationRate));
				if (desiredTimeForaging < 1) {
					desiredTimeForaging = 1;
				}
				if (desiredTimeForaging > thisForagingInterval.size()) {
					desiredTimeForaging = thisForagingInterval.size();
				}
				if (cropFillingTime > 0 && desiredTimeForaging > cropFillingTime) {
					desiredTimeForaging = cropFillingTime;
				}
			}
			desiredTimeForaging = modifyDesiredTimeForaging(x, thisForagingInterval.iterator().next(), desiredTimeForaging);
			double desiredAmount = desiredTimeForaging * energyAssimilationRate
					/ maxResource.energyContent;
			if (desiredAmount > requirements / maxResource.energyContent) {
				desiredAmount = requirements / maxResource.energyContent;
			}
			Iterator<Integer> thisForagingIntervalIterator = thisForagingInterval.iterator();
			for (int i = 0; i < desiredTimeForaging; i++) {
				realForagingInterval.add(thisForagingIntervalIterator.next());
				thisForagingIntervalIterator.remove();
			}
			if (cropFillingTime > 0) {
				cropContent += Double.valueOf(realForagingInterval.size()) / Double.valueOf(cropFillingTime)
						* ((Double) Config.getParameter("WoodpigeonAdult_cropVolume") - cropContent);
			}
			HashMap<String, Object> parameters = new HashMap<String, Object>();
			parameters.put("resourceId", maxResource.getId());
			parameters.put("resourceAmount", desiredAmount);
			parameters.put("realForagingInterval", realForagingInterval);
			parameters.put("birdId", getId());
			obtainedResource = x.giveResource(parameters);
			addTodaysResource(obtainedResource);
			for (int i : realForagingInterval) {
				feedingTimetable[i].available = false;
				feedingTimetable[i].foragingPlace = x.getId();
			}
			timeLeftForForagingToday -= realForagingInterval.size();
			return realForagingInterval;

		}

		private int modifyDesiredTimeForaging(LandscapeUnit x, int startTime, int desiredTimeForaging) {
			int possibleFeedingTime = ((WoodpigeonLandscapeUnit) x).getPossibleFeedingTime(startTime);
			if (possibleFeedingTime >= desiredTimeForaging) {
				return desiredTimeForaging;
			} else if (possibleFeedingTime == 0) {
				return 1;
			} else {
				return possibleFeedingTime;
			}
		}

		/**
		 * Chooses
		 * <code>LandscapeUnits</code> potentially visited on the way from
		 * current location to
		 * <code>LandscapeUnit</code> x.
		 *
		 * The rules of this choice: 1. Create empty set of
		 * <code>LandscapeUnits</code>. 2. Choose fields within the distance
		 * between location and x from location and from x. 3. Add to the set
		 * <code>LandscapeUnits</code> adjoining location and x. 4. Add to the
		 * set
		 * <code>LandscapeUnits</code> that are within th given distance from
		 * both location and x.
		 *
		 * @param x
		 * @return
		 */
		private LandscapeUnit[] chooseFieldsBetween(LandscapeUnit x) {
			HashSet<LandscapeUnit> fields = new HashSet<LandscapeUnit>();
			double borderDistance = location.calculateCentralDistance(x);
			NavigableMap<Double, ArrayList<LandscapeUnit>> locationMap = location.getLandscapeUnits(borderDistance);
			NavigableMap<Double, ArrayList<LandscapeUnit>> xMap = x.getLandscapeUnits(borderDistance);
			//adding neighbours of location and x to fields:
			if (!locationMap.get(0.0).isEmpty()) {
				fields.addAll(locationMap.get(0.0));
			}
			if (!xMap.get(0.0).isEmpty()) {
				fields.addAll(xMap.get(0.0));
			}
			//adding to fields LU within the given distance from both location and x:
			HashSet<LandscapeUnit> locationSet = new HashSet<LandscapeUnit>();
			HashSet<LandscapeUnit> xSet = new HashSet<LandscapeUnit>();
			for (Double key : locationMap.keySet()) {
				locationSet.addAll(locationMap.get(key));
			}
			for (Double key : xMap.keySet()) {
				xSet.addAll(xMap.get(key));
			}
			locationSet.retainAll(xSet);
			fields.addAll(locationSet);
			return (LandscapeUnit[]) fields.toArray(new LandscapeUnit[0]);
		}

		/**
		 * Choosing set of fields to be visited between current location and
		 * field x chosen from the memory.
		 *
		 * The number of fields depends on the distance between location and x
		 * and mean field size in the landscape.
		 *
		 * @param distance
		 * @param fields
		 * @return
		 */
		private HashSet<LandscapeUnit> chooseFieldsOnTheWay(double distance, LandscapeUnit[] fields) {
			HashSet<LandscapeUnit> set = new HashSet<LandscapeUnit>();
			if (distance <= 0) {
				for (int i = 0; i < fields.length; i++) {
					set.add(fields[i]);
				}
				return set;
			}
			double totalArea = 0;
			for (LandscapeUnit lu : fields) {
				totalArea += lu.getSize();
			}
			int number = (int) Math.round(distance / Math.sqrt(totalArea * 10000.0 / fields.length));
			while (set.size() <= number) {
				set.add(fields[Config.getGenerator().nextInt(fields.length)]);
			}
			return set;
		}

		/**
		 * Chooses
		 * <code>LandscapeUnit</code> from the previously defined set.
		 *
		 * @param set
		 * @param time
		 * @return
		 */
		private WoodpigeonLandscapeUnit checkFlockSize(WoodpigeonLandscapeUnit potentialField, int time) {
			ArrayList<Integer> joinedFlockMinSize = (ArrayList<Integer>) Config.getParameter("WoodpigeonAdult_joinedFlockMinSize");
			int flockSizeRange = 0;
			for (int i : joinedFlockMinSize) {
				if (potentialField.getFlockSize(time) >= i) {
					flockSizeRange = joinedFlockMinSize.indexOf(i);
				} else {
					break;
				}
			}
			if (Config.getGenerator().nextDouble() < ((ArrayList<Double>) Config.getParameter("WoodpigeonAdult_flockJoiningProbability")).get(flockSizeRange)) {
				return potentialField;
			} else {
				return null;
			}
		}

		@Override
		public void assignTerritory(Object ter) {
			myTerritory = (WoodpigeonLandscapeUnit) ter;
			addToContactList(myTerritory.getId());
		}

		/**
		 *
		 * @return
		 */
		public Object getTerritory() {
			if (hasTerritory()) {
				return myTerritory;
			} else {
				return null;
			}
		}

		public void abandonTerritory() {
			Message m = new Message(getId(), myTerritory.getId(), "remove_agent");
			sendMessage(m);
			removeFromContactList(myTerritory.getId());
			myTerritory = null;
		}

		/**
		 *
		 * @return @throws model.core.Agent.AgentException
		 */
		@Override
		public boolean getReadyForForaging() throws AgentException {
			if (getNest() != null) {
				if (!World.Instance().getAgent(getNest()).getGreenhorn()) {
					if (incubationRequired || nestRequirements > 0) {
						return super.getReadyForForaging();
					}
					Set<Message> set = findMessage(getNest());
					if (!set.isEmpty()) {
						for (Message m : set) {
							acceptNestFoodRequirements(m);
						}
						return super.getReadyForForaging();
					}
					throw new AgentException();
				}
			}
			return super.getReadyForForaging();
		}

		/**
		 * Checks if the animal has appropriate bodyweight to start reproduction
		 * behaviour.
		 *
		 * @return true if the bodyweight is sufficient, false otherwise
		 * @since 2012-06-28
		 */
		@Override
		public boolean checkBreedingCondition() {
			if (getEnergyReserves() > 0.0) {
				return true;
			} else {
				return false;
			}
		}

		public boolean isMigrating() {
			if (emigrationDate != null && World.getDate().after(emigrationDate)) {
				if (Config.getGenerator().nextDouble()
						< (Double) Config.getParameter("WoodpigeonAnimal_imigrationProbability")) {
					emigrationDate = null;
					changesInBodyweight = 0;
				} else {
					return true;
				}
			}
			return false;
		}

		@Override
		public AgentState changeState() {
			if (getActivityName().equalsIgnoreCase("WoodpigeonOffSeason")
					&& (World.getDate().get(Calendar.DAY_OF_YEAR) > (Integer) Config.getParameter("SeasonalActivity_endOfSeason")
					|| World.getDate().get(Calendar.DAY_OF_YEAR) < (Integer) Config.getParameter("SeasonalActivity_startOfSeason"))
					&& changesInBodyweight >= (Integer) Config.getParameter("WoodpigeonAdult_daysOfHungerBeforeMigration")) {
				//			&& changesInBodyweight >= Config.getDensityDependentEmigration()) {
				migrate();
			}
			return this;
		}

		private void migrate() {
			emigrationDate = World.getDate();
		}

		private Double getProportionOfFieldsRevisited() {
			ArrayList<Long> fields = getFieldsVisited();
			if (!isMigrating() && fields.size() > 0) {
				double a = fieldsRevisited / new Double(fields.size());
				if (a > 1) {
					int b = 0;
				}
				return a;
			} else {
				return null;
			}
		}

		private ArrayList<Long> getFieldsVisited() {
			ArrayList<Long> fields = new ArrayList<Long>();
			for (int i = 0; i < photoperiod; i++) {
				if (!fields.contains(feedingTimetable[i].foragingPlace) && feedingTimetable[i].foragingPlace > 0) {
					fields.add(feedingTimetable[i].foragingPlace);
				}
			}
			return fields;
		}

		@Override
		public Signature getSignature() {
			Signature newSignature = super.getSignature();
			String feedingPlacesIds = "'{";
			for (int i = 0; i < photoperiod; i++) {
				feedingPlacesIds += String.valueOf(feedingTimetable[i].foragingPlace) + ",";
			}
			if (feedingPlacesIds.equalsIgnoreCase("'{")) {
				feedingPlacesIds += "}'";
			} else {
				feedingPlacesIds = feedingPlacesIds.substring(0, feedingPlacesIds.length() - 1)
						+ "}'";
			}
			newSignature.addPrimitiveField("feedingPlacesIds", "integer[]", feedingPlacesIds);
			newSignature.addPrimitiveField("totalTimeForForagingToday", "integer", Integer.toString(totalTimeForForagingToday));
			newSignature.addPrimitiveField("timeForagedToday", "double precision", Double.toString(new Double(totalTimeForForagingToday - timeLeftForForagingToday)
					* ((Integer) Config.getParameter("Config_timeFactor")).doubleValue() / 60.0));
			Double revisited = getProportionOfFieldsRevisited();
			if (revisited != null) {
				newSignature.addPrimitiveField("revisitedFields", "double precision", Double.toString(revisited));
			} else {
				newSignature.addPrimitiveField("revisitedFields", "double precision", "null");
			}
			newSignature.addPrimitiveField("numberOfFields", "integer", Double.toString(getFieldsVisited().size()));
			String landscapeMemory = "'{";
			for (double D : memory.keySet()) {
				for (Memorable mem : memory.get(D)) {
					landscapeMemory += String.valueOf(mem.lu.getId()) + ", ";
				}
			}
			if (landscapeMemory.equalsIgnoreCase("'{")) {
				landscapeMemory += "}'";
			} else {
				landscapeMemory = landscapeMemory.substring(0, landscapeMemory.length() - 2)
						+ "}'";
			}
			newSignature.addPrimitiveField("memory", "integer[]", landscapeMemory);
//			}
			return newSignature;
		}

		/**
		 * Specifies the behaviour of woodpigeons outside of the breeding season
		 * and conditions necessary to start the breeding activities.
		 *
		 * @author Katarzyna Kułakowska
		 * @since 2012-06-28
		 */
		public class WoodpigeonOffSeason extends OffSeason {

			/**
			 * Constructor for
			 * <code>WoodpigeonOffSeason</code>.
			 *
			 * Chooses the roosting place from all the woods in the landscape
			 * larger then specified.
			 *
			 * @param parameters not required. @parameters
			 * WoodpigeonOffSeason_minSizeOfRoostingWood
			 * @since 2012-06-28
			 */
			public WoodpigeonOffSeason(HashMap<String, Object> parameters) {
				super(parameters);
				Set<LandscapeUnit> LU = World.Instance().getLandscapeUnits();
				ArrayList<LandscapeUnit> woods = new ArrayList<LandscapeUnit>();
				for (LandscapeUnit wood : LU) {
					if (wood.getStateClass().getSimpleName().equalsIgnoreCase("WoodpigeonWood")
							&& wood.getSize() >= (Double) Config.getParameter("WoodpigeonOffSeason_minSizeOfRoostingWood")) {
						woods.add(wood);
					}
				}
				roostingPlace = (WoodpigeonLandscapeUnit) woods.get(Config.getGenerator().nextInt(woods.size()));
			}

			/**
			 * Checks if there is a breeding season.
			 *
			 * The breeding season is different for adults and first-winter
			 * birds.
			 *
			 * @return true if
			 * <code>currentDate</code> is in breeding season,false otherwise
			 * @parameters SeasonalActivity_startOfSeason,
			 * SeasonalActivity_endOfSeason
			 * @since 2012-06-27
			 */
			@Override
			public boolean inSeason() {
				int dayOfYear = World.getDate().get(Calendar.DAY_OF_YEAR);
				if (getDaysSinceLaying() - dayOfYear < 365) {
					if (dayOfYear > (Integer) Config.getParameter("SeasonalActivity_lateStartOfSeason")
							&& dayOfYear < (Integer) Config.getParameter("SeasonalActivity_endOfSeason")) {
						return true;
					} else {
						return false;
					}
				} else {
					return super.inSeason();
				}

			}

			/**
			 * Checks woodpigeon-specific conditions for changing to the next
			 * seasonal activity.
			 *
			 * <p> if the bird is currently migrating, the same seasonal
			 * activity is returned. If not, probability of changing the
			 * seasonal activity depends on input parameter
			 * "Reproduce_readyForReproduction" and current daylength. </p>
			 *
			 * @parameters Reproduce_readyForReproduction
			 *
			 * @since 2012-06-27
			 */
			@Override
			public SeasonalActivity nextSeasonalActivity() {
				if (emigrationDate != null || (Config.getGenerator().nextDouble()
						> (Double) Config.getParameter("Reproduce_readyForReproduction")
						* Config.getRealDayLength(World.getDate()) / Config.getLongestDayRealLength())) {
					return this;
				} else {
					return super.nextSeasonalActivity();
				}
			}
		}

		/**
		 * Seasonal activity specific to males entering the breeding season.
		 *
		 * Enables searching for territories.
		 *
		 * @author Katarzyna Kułakowska
		 * @since 2012-06-28
		 */
		public class WoodpigeonFindTerritory extends FindTerritory {

			/**
			 * Constructor for
			 * <code>WoodpigeonFindTerritory</code>.
			 *
			 * @param parameters not required.
			 * @since 2012-06-28
			 */
			public WoodpigeonFindTerritory(HashMap<String, Object> parameters) {
				super(parameters);
			}

			/**
			 * Specifies the process of searching for territory.
			 *
			 * <p> Chooses a random wood from the whole landscape and checks the
			 * availability of territories. If the territory is granted it is
			 * assigned to
			 * <code>myTerritory</code> and added to
			 * <code>concactList</code>. </p>
			 *
			 * @since 2012-06-28
			 */
			@Override
			public void conduct() {
				Set<LandscapeUnit> LU = World.Instance().getLandscapeUnits();
				ArrayList<LandscapeUnit> woods = new ArrayList<LandscapeUnit>();
				for (LandscapeUnit wood : LU) {
					if (wood.getStateClass().getSimpleName().equalsIgnoreCase("WoodpigeonWood")) {
						woods.add(wood);
					}
				}
				myTerritory = ((WoodpigeonLandscapeUnit) woods.get(Config.getGenerator().nextInt(woods.size()))).getTerritory(getId());
				if (myTerritory != null) {
					addToContactList(myTerritory.getId());
				}
			}
		}

		/**
		 * Seasonal activity specific to females entering the breeding season.
		 *
		 * Enables searching for partners.
		 *
		 * @author Katarzyna Kułakowska
		 * @since 2012-06-28
		 */
		public class WoodpigeonFindPartner extends FindPartner {

			/**
			 * Constructor for
			 * <code>WoodpigeonFindPartner</code>.
			 *
			 * @param parameters (not required)
			 * @since 2012-06-28
			 */
			public WoodpigeonFindPartner(HashMap<String, Object> parameters) {
				super(parameters);
			}

			/**
			 * Specifies the process of searching for a partner.
			 *
			 * <p> Chooses a random wood from the whole landscape and checks the
			 * availability of partners. If there is any available partner, he
			 * is informed via message about the female's choice.
			 *
			 * @since 2012-06-28
			 */
			@Override
			public void conduct() {
				Set<LandscapeUnit> LU = World.Instance().getLandscapeUnits();
				ArrayList<LandscapeUnit> woods = new ArrayList<LandscapeUnit>();
				for (LandscapeUnit wood : LU) {
					if (wood.getStateClass().getSimpleName().equalsIgnoreCase("WoodpigeonWood")) {
						woods.add(wood);
					}
				}
				Long potentialPartner = ((WoodpigeonLandscapeUnit) woods.get(Config.getGenerator().nextInt(woods.size()))).providePartnerId();
				if (potentialPartner != null) {
					informPartner(potentialPartner);
				}
			}
		}

		/**
		 * Seasonal activity specific to females after they have found a
		 * partner.
		 *
		 * Enables checking the conditions necessary for reproduction and laying
		 * eggs.
		 *
		 * @author Katarzyna Kułakowska
		 * @since 2012-06-28
		 */
		public class WoodpigeonReproduce extends Reproduce {

			/**
			 * Keeps the energy in kJ necessary for producing the eggs in the
			 * current reproduction attempt.
			 *
			 * @since 2012-06-28
			 */
			private double energyRequiredForEggs;

			/**
			 * Constructor for
			 * <code>WoodpigeonReproduce</code>.
			 *
			 * Calculates the energy required for producing the eggs specified
			 * in
			 * <code>eggsWeights</code>.
			 *
			 * @param parameters (not required)
			 * @since 2012-06-28 @parameters Reproduce_eggCost
			 */
			public WoodpigeonReproduce(HashMap<String, Object> parameters) {
				super(parameters);
				energyRequiredForEggs = 0.0;
				for (double d : eggsWeights) {
					energyRequiredForEggs += d * (1.0 + ((Double) Config.getParameter("Reproduce_eggCost")));
				}
			}

			/**
			 * Returns the number of eggs for the current reproduction attempt.
			 *
			 * @return number of eggs
			 * @since 2012-06-28 @parameters Reproduce_numberOfEggs
			 *
			 */
			@Override
			public int getNumberOfEggs() {
				return (Integer) Config.getParameter("Reproduce_numberOfEggs");
			}

			/**
			 * Checks the woodpigeon-specific conditions for reproduction and
			 * calls
			 * <code>super.conduct</code> if thy are met.
			 *
			 * The conditions include: sufficient energy reserves in
			 * <code>reproductionBufferSum</code> and successful feeding routine
			 * during the period of pseudo-incubation for the female and her
			 * partner.
			 *
			 * @since 2012-06-28
			 */
			@Override
			public void conduct() {
				if (reproductionBufferSum > energyRequiredForEggs
						&& checkFeedingRoutine()
						&& ((WoodpigeonAnimal) World.Instance().getAgent(getPartner())).checkFeedingRoutine()) {
					while (!reproductionFinished) {
						super.conduct();
					}
				}
				loadReproductionBuffer();
			}

			/**
			 * Adds the location to the nest parameters obtained from the call
			 * to
			 * <code>super.prepareNest</code>.
			 *
			 * @return
			 * <code>HashMap</code> with parameters for
			 * <code>Nest</code> constructor
			 * @since 2012-06-28
			 */
			@Override
			public HashMap<String, Object> prepareNest() {
				HashMap<String, Object> para = super.prepareNest();
				para.put("location", myTerritory);
				return para;
			}
		}

		/**
		 * Seasonal activity specific to males after acquiring the territory in
		 * breeding season.
		 *
		 *
		 * @author Katarzyna Kułakowska
		 * @since 2012-06-28
		 */
		public class WoodpigeonWaitingMale extends WaitingMale {

			/**
			 * Constructor for
			 * <code>WoodpigeonWaitingMale</code>.
			 *
			 * @param parameters (not required)
			 * @since 2012-06-28
			 */
			public WoodpigeonWaitingMale(HashMap<String, Object> parameters) {
				super(parameters);
			}
		}

		/**
		 * Seasonal activity specific to females with partners and nests.
		 *
		 * @author Katarzyna Kułakowska
		 * @since 2012-06-28
		 */
		public class WoodpigeonWaitingFemale extends WaitingFemale {

			/**
			 * Constructor for
			 * <code>WoodpigeonWaitingFemale</code>.
			 *
			 * @param parameters (not required)
			 * @since 2012-06-28 *
			 */
			public WoodpigeonWaitingFemale(HashMap<String, Object> parameters) {
				super(parameters);
			}
		}

		private class TimetableSlot {

			public boolean available;
			public long foragingPlace;

			public TimetableSlot(boolean available) {
				this.available = available;
				foragingPlace = -1;
			}
		}

		@Override
		public SeasonalActivity createSeasonalActivity(HashMap<String, Object> parameters) {
			if (((String) parameters.get("activityName")).equalsIgnoreCase("WaitingMale")) {
				return new WoodpigeonWaitingMale(parameters);
			} else if (((String) parameters.get("activityName")).equalsIgnoreCase("FindTerritory")) {
				return new WoodpigeonFindTerritory(parameters);
			} else if (((String) parameters.get("activityName")).equalsIgnoreCase("WaitingFemale")) {
				return new WoodpigeonWaitingFemale(parameters);
			} else if (((String) parameters.get("activityName")).equalsIgnoreCase("FindPartner")) {
				return new WoodpigeonFindPartner(parameters);
			} else if (((String) parameters.get("activityName")).equalsIgnoreCase("Reproduce")) {
				return new WoodpigeonReproduce(parameters);
			} else if (((String) parameters.get("activityName")).equalsIgnoreCase("OffSeason")) {
				return new WoodpigeonOffSeason(parameters);
			} else {
				return null;
			}
		}
	}

	/**
	 *
	 */
	public class WoodpigeonEgg extends Egg {

		/**
		 *
		 * @param parameters
		 */
		public WoodpigeonEgg(HashMap<String, Object> parameters) {
			super(parameters);
		}

		@Override
		public Signature getSignature() {
			return null;
		}
	}

	/**
	 *
	 */
	public class WoodpigeonNestling extends Nestling {

		/**
		 *
		 * @param parameters
		 */
		public WoodpigeonNestling(HashMap<String, Object> parameters) {
			super(parameters);
		}

		@Override
		public Signature getSignature() {
			return null;
		}

		/**
		 *
		 * @param mum_memory
		 * @param dad_memory
		 */
		public void createMemory(TreeMap<Double, Set<Memorable>> mum_memory,
				TreeMap<Double, Set<Memorable>> dad_memory) {
			memory = new TreeMap<Double, Set<Memorable>>();
			Set<LandscapeUnit> landscapes = new HashSet<LandscapeUnit>();
			if (mum_memory != null && dad_memory == null) {
				memory.putAll(mum_memory);
			} else if (mum_memory == null && dad_memory != null) {
				memory.putAll(dad_memory);
			} else if (mum_memory != null && dad_memory != null) {
				memory.putAll(mum_memory);
				for (Double D : memory.keySet()) {
					for (Memorable mem : memory.get(D)) {
						landscapes.add(mem.lu);
					}
				}
				for (Double D : dad_memory.keySet()) {
					for (Memorable mem : dad_memory.get(D)) {
						if (!landscapes.contains(mem.lu)) {
							if (memory.containsKey(D)) {
								memory.get(D).add(mem);
							} else {
								memory.put(D, dad_memory.get(D));
							}
						}
					}
				}
			}
		}
	}

	/*
	 * Fields:
	 */
	private LandscapeUnit location;
	private LandscapeUnit roostingPlace;
	private LandscapeUnit myTerritory;
	private ArrayList<Boolean> succesfulFeedingRoutine;
	private volatile TreeMap<Double, Set<Memorable>> memory;
	private Calendar emigrationDate;

	/*
	 * Methods:
	 */
	/**
	 *
	 * @param parameters
	 */
	public WoodpigeonAnimal(HashMap<String, Object> parameters) {
		super(parameters);
		if ((Boolean) parameters.get("at_initialization")) {
			location = roostingPlace;
		} else {
			location = (LandscapeUnit) parameters.get("location");
		}
		emigrationDate = null;
	}

	/**
	 *
	 * @return
	 */
	public TreeMap<Double, Set<Memorable>> getMemory() {
		TreeMap<Double, Set<Memorable>> memoryclone = new TreeMap<Double, Set<Memorable>>();
		for (double D : memory.keySet()) {
			for (Memorable mem : memory.get(D)) {
				Memorable memclone = mem.clone();
				if (memoryclone.containsKey(D)) {
					memoryclone.get(D).add(memclone);
				} else {
					Set<Memorable> newset = new HashSet<Memorable>();
					newset.add(memclone);
					memoryclone.put(D, newset);
				}
			}
		}
		return memoryclone;
	}

	@Override
	public boolean hasTerritory() {
		if (myTerritory == null) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 *
	 * @return
	 */
	public boolean checkFeedingRoutine() {
		return !succesfulFeedingRoutine.contains(false);
	}

	/**
	 * @verification Warning of the state name passed in parameters is different
	 * then Adult, Egg or Nestling.
	 */
	@Override
	public AgentState createAgentState(HashMap<String, Object> parameters) {
		if (((String) parameters.get("stateName")).equalsIgnoreCase("Adult")) {
			return new WoodpigeonAdult(parameters);
		} else if (((String) parameters.get("stateName")).equalsIgnoreCase("Nestling")) {
			return new WoodpigeonNestling(parameters);
		} else if (((String) parameters.get("stateName")).equalsIgnoreCase("Egg")) {
			return new WoodpigeonEgg(parameters);
		} else {
// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 1
				try {
					throw new SkyPigException(
							"Improper state name: " + ((String) parameters.get("stateName")));
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
// *****VERIFICATION MODE END*****//
		return null;
	}

	@Override
	public Signature getSignature() {
		Signature newSignature = super.getSignature();
		if (myTerritory != null) {
			newSignature.addPrimitiveField("territory", "bigint", Long.toString(myTerritory.getId()));
		} else {
			newSignature.addPrimitiveField("territory", "bigint", "null");
		}
		if (roostingPlace != null) {
			newSignature.addPrimitiveField("roostingPlace", "bigint", Long.toString(roostingPlace.getId()));
		} else {
			newSignature.addPrimitiveField("roostingPlace", "bigint", "null");
		}
		newSignature.addPrimitiveField("location", "bigint", Long.toString(location.getId()));
		if (emigrationDate != null && World.getDate().after(emigrationDate)) {
			newSignature.addPrimitiveField("migration", "boolean", "true");
		} else {
			newSignature.addPrimitiveField("migration", "boolean", "false");
		}
		return newSignature;

	}

	/**
	 *
	 */
	public class Memorable {

		/*
		 * Fields:
		 */
		private LandscapeUnit lu;
		private Calendar dateMemorized;

		/**
		 *
		 * @param l
		 * @param date
		 */
		public Memorable(LandscapeUnit l, Calendar date) {
			lu = l;
			dateMemorized = date;
		}

		/**
		 *
		 * @return
		 */
		@Override
		public Memorable clone() {
			return new Memorable(lu, (Calendar) dateMemorized.clone());
		}

		/**
		 *
		 * @return
		 */
		public LandscapeUnit getLu() {
			return lu;
		}
	}
}
