//Copyright 2013 Katarzyna Kułakowska, Tomasz Kułakowski.
//
//This file is part of SkyPig.
//
//SkyPig is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//SkyPig is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with SkyPig; if not, write to the Free Software
//Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
package model.woodpigeon;

import java.util.*;
import java.util.logging.Level;
import model.core.*;
import model.woodpigeon.WoodpigeonAnimal.WoodpigeonNestling;

/**
 *
 * @author kasia
 */
public class WoodpigeonNest extends Nest {

	/**
	 * Specifies the characteristics of woodpigeon nest during the incubation
	 * stage.
	 *
	 * @author Katarzyna Kułakowska
	 * @author Tomasz Kułakowski
	 * @since 2012-06-29
	 */
	public class WoodpigeonEggs extends Eggs {

		/**
		 * The ratio of incubation time to the length of the day.
		 *
		 * @since 2012-06-29
		 */
		private double incubationRatio;

		/**
		 * Constructor for
		 * <code>WoodpigeonEggs</code>.
		 *
		 * Sets the
		 * <code>incubationRatio<>/code = 0 and
		 * <code>dateOfHatching</code> =
		 * <code>steDateofHatching</code>
		 *
		 * @param parameters (not required)
		 * @since 2012-06-29
		 */
		public WoodpigeonEggs(HashMap<String, Object> parameters) {
			super(parameters);
			incubationRatio = 0;
			setDateOfHatching();
		}

		/**
		 * Checks if the eggs were predated.
		 *
		 * The probability of predation is specified as Eggs_predationRate from
		 * parameter list multiplied by 1-
		 * <code>incubationRatio</code>.
		 *
		 * @return true if the eggs are predated, false otherwise. @parameters
		 * Eggs_predationRate
		 * @since 2012-06-29
		 */
		@Override
		public boolean checkIfPredated() {
			if (Config.getGenerator().nextDouble()
					< (Double) Config.getParameter("Eggs_backgroundPredationRate")
					|| Config.getGenerator().nextDouble()
			//		< (Double) Config.getParameter("Eggs_predationRate")
			//		* (1 - incubationRatio)) {
					< 1 - Math.pow((Math.pow((1 - (Double) Config.getParameter("Eggs_predationRate")), 
					((Integer) Config.getParameter("Config_timeFactor")).doubleValue() / 60.0)), 
					((1-incubationRatio) * Config.getDayLength(World.getDate())))) {
				return true;
			} else {
				return false;
			}
		}

		/**
		 * Increases the incubation ratio according to the time of incubation.
		 *
		 * <p>If the total incubation time is longer then current day length it
		 * is reduced to current day length. </p>
		 *
		 * @param incubationTime from one parent
		 * @since 2012-06-30
		 */
		@Override
		public void receiveIncubation(double incubationTime) {
			incubationRatio += incubationTime
					/ Config.getDayLength(World.getDate());
			if (incubationRatio > 1) {
				incubationRatio = 1;
			}
		}

		/**
		 * Checks if today is the hatching day.
		 *
		 * @return true if the eggs hatch today, false otherwise.
		 * @since 2012-06-30
		 */
		@Override
		public boolean checkIfEggsHatch() {
			if (getDaysSinceLaying() < ((Integer) Config.getParameter("WoodpigeonEggs_hatchingAge"))) {
				return false;
			} else {
				return true;
			}
		}

		/**
		 * Resets
		 * <code>incubationRatio</code> to 0 and calls
		 * <code>super.reset</code>.
		 *
		 * @since 2012-06-30
		 */
		@Override
		public void reset() {
			incubationRatio = 0;
			super.reset();
		}

		/**
		 * Sets the date of hatching based on the fixed incubation period.
		 *
		 * @parameters WoodpigeonEggs_hatchingAge
		 *
		 * @since 2012-06-29
		 */
		public void setDateOfHatching() {
			Calendar hatchingDate = getDateOfLaying();
			hatchingDate.add(Calendar.DAY_OF_YEAR,
					(Integer) Config.getParameter("WoodpigeonEggs_hatchingAge"));
			super.setDateOfHatching(hatchingDate);
		}

		@Override
		public Signature getSignature() {
			Signature mySignature = super.getSignature();
			mySignature.addPrimitiveField("incubationRatio", "double precision", Double.toString(incubationRatio));
			return mySignature;
		}
	}

	/**
	 *
	 */
	public class WoodpigeonNestlings extends Nestlings {

		/**
		 *
		 * @param parameters
		 */
		public WoodpigeonNestlings(HashMap<String, Object> parameters) {
			super(parameters);
		}

		/**
		 *
		 * @param nestling
		 */
		@Override
		public void prepareNestling(Animal nestling) {
			TreeMap<Double, Set<WoodpigeonAnimal.Memorable>> mum_memory = null;
			TreeMap<Double, Set<WoodpigeonAnimal.Memorable>> dad_memory = null;
			if (getFemaleOwnerId() != null) {
				mum_memory = ((WoodpigeonAnimal) World.Instance().getAgent(getFemaleOwnerId())).getMemory();
			}
			if (getMaleOwnerId() != null) {
				dad_memory = ((WoodpigeonAnimal) World.Instance().getAgent(getMaleOwnerId())).getMemory();
			}
			((WoodpigeonNestling) nestling.getState(getId())).createMemory(mum_memory, dad_memory);
		}

		@Override
		public Signature getSignature() {
			return null;
		}
	}

	/*
	 * Fields:
	 */
	private LandscapeUnit location;

	/*
	 * Methods:
	 */
	/**
	 *
	 * @param parameters
	 */
	public WoodpigeonNest(HashMap<String, Object> parameters) {
		super(parameters);
		location = (LandscapeUnit) parameters.get("location");
	}

	@Override
	public boolean checkIfDestroyed() {
		if ((Boolean) Config.getParameter("WoodpigeonNest_destruction")) {
			for (Calendar date : (ArrayList<GregorianCalendar>) Config.getParameter("WoodpigeonNest_destructionDates")) {
				if (World.calculateDateDifference(date) == 0) {
					if (Config.getGenerator().nextDouble()
							< (Double) Config.getParameter("WoodpigeonNest_destructionProbability")) {
						return true;
					}
				}
			}
		}
		return false;
	}

	/**
	 * @verification TEST 1 Warning if the the state name passed in parameters
	 * is other than Eggs or Nestlings.
	 */
	@Override
	public AgentState createAgentState(HashMap<String, Object> parameters) {
		if (((String) parameters.get("stateName")).equalsIgnoreCase("Eggs")) {
			return new WoodpigeonEggs(parameters);
		} else if (((String) parameters.get("stateName")).equalsIgnoreCase("Nestlings")) {
			return new WoodpigeonNestlings(parameters);
		} else {
// *****VERIFICATION MODE START*****//
			if (World.verificationMode) {
				// ***TEST 1
				try {
					throw new SkyPigException(
							"Improper state name: " + ((String) parameters.get("stateName")));
				} catch (SkyPigException ex) {
					World.log(Level.WARNING, Config.getLogInfo(), ex);
				}
			}
		}
// *****VERIFICATION MODE END*****//
		return null;
	}
}
